/**
	$Id: MasterSlotManager.java 5823 2011-10-24 13:02:46Z nicklas $

	Copyright (C) 2010 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.jobagent.slotmanager;


import java.net.Socket;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.TimerTask;

import net.sf.basedb.clients.jobagent.Agent;
import net.sf.basedb.core.Application;
import net.sf.basedb.core.Job;
import net.sf.basedb.util.jobagent.JobAgentConnection;
import net.sf.basedb.util.jobagent.RequestHandler;

/**
	An extension to the internal slot manager which also accepts
	slot assignment from remote job agents. A custom request
	handler ({@link RemoteSlotRequestHandler}) is installed to take care 
	of the communication. Remotely assigned slots has a time-out (30 seconds) 
	associated with them. This means that if the connection with
	the remote host is lost, the assigned slots are re-cycled.

	<p>
	The requst handler uses a simple protocol with the following actions:
	<ul>
	<li><code>slotmanager://host:port/get-slot?TIME</code>: Get a new slot for the
		TIME estimated execution time. TIME is a name from the {@link net.sf.basedb.core.Job.ExecutionTime}
		enumeration. If succcessful, the request handler answers with <code>OK</code> on the first line
		and then <code>Id:slot-id</code> and <code>Slot:TIME</code> on the following lines.
		It is recommended that the answer is parsed with {@link JobAgentConnection#parseSimpleAnswer(String)}.
	<li><code>slotmanger://slot-id@host:port/release-slot</code>: Release the slot with
		the given id.
	<li><code>slotmanager://slot-id@host:port/ping-slot</code>: Ping the slot with the 
		given id to indicate that the job is still running on the remote job agent. A remote
		slot has a timeout of 30 seconds so the remote job agent has to send a ping at least
		at that interval or the slot will be returned to the pool.
	</ul>
	
	If a request fails, the request handler returns <code>FAILED</code> followed with an
	explanation.

	@author Nicklas
	@since 2.16
	@base.modified $Date: 2011-10-24 15:02:46 +0200 (Mon, 24 Oct 2011) $
*/
public class MasterSlotManager
	extends InternalSlotManager
{

	private static final org.apache.log4j.Logger log = 
		org.apache.log4j.LogManager.getLogger("net.sf.basedb.clients.jobagent.slotmanager.MasterSlotManager");

	
	private Map<String, RemoteSlot> remoteSlots;
	private RemoteSlotRequestHandler remoteRequestHandler;
	private TimerTask timeoutChecker;

	public MasterSlotManager()
	{}
	
	/*
		From the SlotManager interface
		------------------------------
	*/
	@Override
	public void init(Agent agent)
	{
		log.info("Initalizing master slot manager");
		super.init(agent);
		this.remoteSlots = Collections.synchronizedMap(new HashMap<String, RemoteSlot>());
		this.remoteRequestHandler = new RemoteSlotRequestHandler(this);
		agent.registerRequestHandler(remoteRequestHandler, "slotmanager");
		this.timeoutChecker = Application.getScheduler().schedule(new RemoteSlotTimeoutChecker(this), 10000, 10000, false);
	}
	@Override
	public void close()
	{
		log.info("Closing master slot manager");
		getAgent().unregisterRequestHandler("slotmanager");
		timeoutChecker.cancel();
		remoteSlots.clear();
		remoteSlots = null;
		remoteRequestHandler = null;
		super.close();
	}
	// -----------------------------------------------
	
	/**
		Create a slot for a job on a remote job agent with
		the specified estimated execution time. This method will
		call the {@link InternalSlotManager#getSlot(net.sf.basedb.core.Job.ExecutionTime)}
		and then wrap the slot in a {@link RemoteSlot} instance.
		@return A remote slot, or null if not slot is available
	*/
	RemoteSlot getRemoteSlot(Job.ExecutionTime estimated)
	{
		RemoteSlot remoteSlot = null;
		Slot slot = getSlot(estimated);
		if (slot != null)
		{
			remoteSlot = new RemoteSlot(slot);
			String slotId = slot.getId();
			remoteSlots.put(slotId, remoteSlot);
		}
		return remoteSlot;
	}
	
	/**
		Get the remote slot with the given id.
		@param slotId The id of the slot
		@return A remote slot or null if no slot with the given id exists
	*/
	RemoteSlot getRemoteSlot(String slotId)
	{
		return remoteSlots.get(slotId);
	}
	
	/**
		Release the remote slot with the given id.
		@param slotId The id of the slot
		@return A remote slot or null if no slot with the given id exists
	*/
	RemoteSlot releaseRemoteSlot(String slotId)
	{
		RemoteSlot remoteSlot = remoteSlots.remove(slotId);
		if (remoteSlot != null)
		{
			Slot slot = remoteSlot.getSlot();
			releaseSlot(slot);
		}
		return remoteSlot;
	}
	
	/**
		Check the remote slots to see if they have timed out. If so,
		release them.
		@return The number of timed out slots
	*/
	int checkTimeoutOnRemoteSlots()
	{
		long timeout = System.currentTimeMillis() - 30000; // 30 seconds ago
		int num = 0;
		synchronized (remoteSlots)
		{
			Iterator<RemoteSlot> it = remoteSlots.values().iterator();
			while (it.hasNext())
			{
				RemoteSlot remoteSlot = it.next();
				if (remoteSlot.getLastUpdated() < timeout)
				{
					Slot slot = remoteSlot.getSlot();
					log.info("Timeout on slot: " + slot.getId());
					it.remove();
					releaseSlot(slot);
					num++;
				}
			}
		}
		return num;
	}
	
	/**
		Request handler implementation that accepts requests for slots
		from remote hosts.
	*/
	static class RemoteSlotRequestHandler
		implements RequestHandler
		
	{
		private final MasterSlotManager master;
		
		RemoteSlotRequestHandler(MasterSlotManager master)
		{
			this.master = master;
		}
		
		/*
			From the RequestHandler interface
			---------------------------------
		*/
		@Override
		public String handleCmd(Socket socket, String cmd)
		{
			String answer = null;
			try
			{
				URI uri = new URI(cmd);
				String action = uri.getPath();
				String slotId = uri.getUserInfo();
				
				log.debug("Got request '" + action + "' for slot: " + slotId);
				
				if ("/get-slot".equals(action))
				{
					// slotmanager://host:port/get-slot?TIME
					Job.ExecutionTime estimated = Job.ExecutionTime.valueOf(uri.getQuery());
					RemoteSlot remoteSlot = master.getRemoteSlot(estimated);
					if (remoteSlot != null)
					{
						Slot slot = remoteSlot.getSlot();
						slotId = slot.getId();
						answer = "OK\nId:" + slotId + 
							"\nSlot:"+slot.getEstimatedExecutionTime().name();
					}
					else
					{
						answer = "FAILED No slot available";
					}
				}
				else if ("/ping-slot".equals(action))
				{
					// slotmanager://slot-id@host:port/ping-slot
					RemoteSlot remoteSlot = master.getRemoteSlot(slotId);
					if (remoteSlot != null)
					{
						remoteSlot.setLastUpdated();
						answer = "OK";
					}
					else
					{
						answer = "FAILED Unknown slot id";
					}
				}
				else if ("/release-slot".equals(action))
				{
					// slotmanager://slot-id@host:port/release-slot
					RemoteSlot remoteSlot = master.releaseRemoteSlot(slotId);
					if (remoteSlot != null)
					{
						answer = "OK";
					}
					else
					{
						answer = "FAILED Unknown slot id";
					}
				}
			}
			catch (URISyntaxException ex)
			{
				answer = "FAILED " + ex.getMessage();
			}
			return answer;
		}
		// -----------------------------------------------
	}
	
	/**
		Task that is scheduled at regular intervals to check for
		timed out slots.
	*/
	static class RemoteSlotTimeoutChecker
		extends TimerTask
	{
	
		private final MasterSlotManager master;
		
		RemoteSlotTimeoutChecker(MasterSlotManager master)
		{
			this.master = master;
		}

		/*
			From the TimerTask interface
			---------------------------------
		*/
		@Override
		public void run()
		{
			log.debug("Checking for timed out remote slots");
			int timedoutSlots = master.checkTimeoutOnRemoteSlots();
			log.debug("Found " + timedoutSlots + " timed out slot(s)");
		}
		// ----------------------------------
		
	}
	
	/**
		A remote slot is a slot with some information about
		when it was created and last updated. 
	*/
	static class RemoteSlot
	{

		private final Slot slot;
		private final long created;
		private long lastUpdated;
		
		RemoteSlot(Slot slot)
		{
			this.slot = slot;
			this.created = System.currentTimeMillis();
			this.lastUpdated = created;
		}
		
		Slot getSlot()
		{
			return slot;
		}
		
		long getCreated()
		{
			return created;
		}
		
		long getLastUpdated()
		{
			return lastUpdated;
		}
				
		void setLastUpdated()
		{
			this.lastUpdated = System.currentTimeMillis();
		}
		
	}
	
}
