/*
	$Id: DynamicRawDataQuery.java 5886 2011-11-23 13:32:22Z nicklas $
	
	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/
	
	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.
	
	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.HashSet;
import java.util.Set;

import net.sf.basedb.core.query.Dynamic;
import net.sf.basedb.core.query.JoinType;
import net.sf.basedb.core.query.Select;

/**
	Represents a query for raw data in a "dynamic" way. The main difference 
	from a client applications point of view, is that it is possible to 
	specify which columns that should be selected. In a techinical point of
	view the difference is that we are using SQL instead of HQl.
	<p>
	This type of query is needed to bridge the gap between the dynamic
	and regular parts of the database. For example, a query of this type
	can be used by a batcher to insert the original intensities in a raw
	bioassayset.
	
	@author Nicklas
	@version 2.0
	@see RawBioAssay#getDynamicRawData()
	@see net.sf.basedb.util.IntensityCalculatorUtil
	@base.modified $Date: 2011-11-23 14:32:22 +0100 (Wed, 23 Nov 2011) $
*/
public class DynamicRawDataQuery
	extends AbstractSqlQuery
{

	private final RawDataType rawDataType;
	private final RealTable rootTable;
	private final Set<Object> joinedItems;
	private RawBioAssay rawBioAssay;

	/**
		Create a new dynamic query for raw data.		
		@param rawBioAssay The raw bioassay to query on
	*/
	DynamicRawDataQuery(RawBioAssay rawBioAssay)
	{
		this(rawBioAssay.getRawDataType());
		this.rawBioAssay = rawBioAssay;
	}
	
	/**
		Create a new dynamic query for raw data.
		@param rawDataType The raw data type
		@since 3.1
	*/
	DynamicRawDataQuery(RawDataType rawDataType)
	{
		super(rawDataType.getRealTable().getQualifiedTableName());
		this.rawDataType = rawDataType;
		this.rootTable = rawDataType.getRealTable();
		joinedItems = new HashSet<Object>();
		if (!rawDataType.isStoredInDb())
		{
			throw new BaseException("Raw data for raw data type '" + rawDataType + "' is not stored in the database.");
		}
	}
	
	/*
		From the Query interface
		-------------------------------------------
	*/
	/**
		The alias of the item that is returned by this query.
		@see Item#getAlias()
	*/
	public String getRootAlias()
	{
		return HibernateUtil.quote(rootTable.getAlias());
	}
	public boolean isReadonly()
	{
		return false;
	}
	// -------------------------------------------

	/*
		From the AbstractSqlQuery class
		-------------------------------------------
	*/
	/**
		No default selection.
	*/
	Select[] getDefaultSelects()
	{
		return null;
	}
	// -------------------------------------------

	public RawDataType getRawDataType()
	{
		return rawDataType;
	}
	
	RealTable getRootTable()
	{
		return rootTable;
	}
	
	public RawBioAssay getRawBioAssay()
	{
		return rawBioAssay;
	}
	
	/**
		Join information about reporters in the query.
		@param joinType The type of join. Can not be null.
	 	@throws InvalidDataException If joinType parameter is null.
		@see Dynamic#selectReporter(String)
		@see Dynamic#reporter(String)
		@since 3.1
	*/
	public void joinReporters(JoinType joinType)
	{
		if (joinType == null) throw new InvalidUseOfNullException("joinType");
		if (!joinedItems.contains(RealTable.REPORTERS))
		{
			join(new RealJoin(joinType, getRootTable(), RealTable.REPORTERS, "reporter_id"));
			joinedItems.add(RealTable.REPORTERS);
		}
	}

}
