/*
	$Id: PlateType.java 4889 2009-04-06 12:52:39Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.Date;
import java.util.Set;

import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.data.PlateTypeData;

/**
	This class represents plate types, which is a name, description,
	geometry and a list of event types.

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2009-04-06 14:52:39 +0200 (Mon, 06 Apr 2009) $
*/
public class PlateType
	extends CommonItem<PlateTypeData>
	implements Registered
{
	/**
		The type of item represented by this class.
		@see Item#PLATETYPE
		@see #getType()
	*/
	public static final Item TYPE = Item.PLATETYPE;
	
	/**
		Create a new <code>PlateType</code> item.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@param plateGeometry The geometry of the new plate type
		@return The new <code>PlateType</code> item
		@throws PermissionDeniedException If the loggged in user
			doesn't have use permission for the geometry
		@throws InvalidDataException If the geometry is null
		@throws BaseException If there is an error
	*/
	public static PlateType getNew(DbControl dc, PlateGeometry plateGeometry)
		throws BaseException
	{
		PlateType pt = dc.newItem(PlateType.class);
		pt.setName("New plate type");
		pt.setPlateGeometry(plateGeometry);
		pt.getData().setEntryDate(new Date());
		return pt;
	}

	/**
		Get a <code>PlateType</code> item when you know the id.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The id of the item to load
		@return The <code>PlateType</code> item
		@throws ItemNotFoundException If an item with the specified 
			id is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static PlateType getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		PlateType pt = dc.loadItem(PlateType.class, id);
		if (pt == null) throw new ItemNotFoundException("PlateType[id="+id+"]");
		return pt;
	}

	/**
		Get a query that returns plate types.
		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<PlateType> getQuery()
	{
		return new ItemQuery<PlateType>(PlateType.class);
	}

	PlateType(PlateTypeData plateTypeData)
	{
		super(plateTypeData);
	}

	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the Registered interface
		-------------------------------------------
	*/
	public Date getEntryDate()
	{
		return DateUtil.copy(getData().getEntryDate());
	}
	// -------------------------------------------

	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		Check if:
		<ul>
		<li>Any {@link Plate}:s of this type exists
		</ul>
	*/
	public boolean isUsed()
		throws BaseException
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_PLATES_FOR_PLATETYPE", "count(*)");
		/*
			SELECT {1}
			FROM PlateData p
			WHERE p.plateType = :plateType
		*/
		query.setEntity("plateType", this.getData());
		return HibernateUtil.loadData(Long.class, query) > 0 || super.isUsed();
	}
	/**
		Get all:
		<ul>
		<li>{@link Plate}:s created from this plate type
		<ul>
		@since 2.2
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_PLATES_FOR_PLATETYPE", "p.id");
			/*
				SELECT {1}
				FROM PlateData p
				WHERE p.plateType = :plateType
			*/
		query.setEntity("plateType", this.getData());
		addUsingItems(using, Item.PLATE, query);
		return using;
	}
	// -------------------------------------------

	/**
		Get the {@link PlateGeometry} of this plate type.
		@return The <code>PlateGeometry</code> item
		@throws PermissionDeniedException If the logged in user doesn't have
			{@link Permission#READ} permission for the plate geometry
		@throws BaseException If there is another error
	*/
	public PlateGeometry getPlateGeometry()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(PlateGeometry.class, getData().getPlateGeometry()); 
	}
	
	/**
		Set the plate geometry of this plate type.
		@throws PermissionDeniedException If the logged in user
			doesn't have use permission for the geometry
		@throws InvalidDataException If the geometry is null
	*/
	private void setPlateGeometry(PlateGeometry plateGeometry)
		throws PermissionDeniedException, InvalidDataException
	{
		if (plateGeometry == null) throw new InvalidUseOfNullException("plateGeometry");
		plateGeometry.checkPermission(Permission.USE);
		getData().setPlateGeometry(plateGeometry.getData());
	}
	
	/**
		Create a new event type for this plate type.
		@return The new <code>PlateEventType</code> item
		@throws PermissionDeniedException If the logged in user
			doesn't have write permission for this plate type
		@throws BaseException If there is another error
	*/
	public PlateEventType newEventType()
		throws PermissionDeniedException, BaseException
	{
		return PlateEventType.getNew(getDbControl(), this);
	}
	
	/**
		Get a query that returns all event types for this plate type.
		@return An {@link ItemQuery} object
	*/
	public ItemQuery<PlateEventType> getEventTypes()
	{
		return PlateEventType.getQuery(this);
	}
	
	/**
		Create a new plate of this plate type.
		@return The new <code>Plate</code> item
		@throws PermissionDeniedException If the logged in user
			doesn't have use permission for this plate type
		@throws BaseException If there is another error
	*/
	public Plate newPlate()
		throws PermissionDeniedException, BaseException
	{
		return Plate.getNew(getDbControl(), this);
	}
	
	/**
		Get a query that returns all plates of this plate type.
		@return An {@link ItemQuery} object
	*/
	public ItemQuery<Plate> getPlates()
	{
		ItemQuery<Plate> query = Plate.getQuery();
		query.restrictPermanent(
			Restrictions.eq(
				Hql.property("plateType"),
				Hql.entity(this)
			)
		);
		return query;
	}
}
