/*
	$Id: RawDataType.java 5590 2011-03-16 11:48:47Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;
import java.util.List;
import java.util.Map;
import java.util.HashMap;
import java.util.Collections;

import net.sf.basedb.core.data.PlatformData;
import net.sf.basedb.core.data.PlatformVariantData;
import net.sf.basedb.core.data.RawData;

/**
	This class holds information about a raw data type. Unlike many other
	classes this information is not stored in the database, but in the
	<code>raw-data-types.xml</code> XML file. This file is parsed upon
	startup by the {@link RawDataTypes} class.
	<p>
	Raw data types are important for {@link RawBioAssay}:s.

	@author Nicklas
	@version 2.0
	@see RawDataTypes
	@see RawBioAssay
	@base.modified $Date: 2011-03-16 12:48:47 +0100 (Wed, 16 Mar 2011) $
*/
public class RawDataType
	implements Comparable<RawDataType>
{

	private final String id;
	private String name;
	private final String description;
	private final boolean fileOnly;
	private final int channels;
	private final PlatformData platform;
	private final PlatformVariantData variant;
	private String table;
	private RealTable realTable;
	private List<RawDataProperty> properties;
	private Map<String, RawDataProperty> namedProperties;
	private List<IntensityFormula> formulas;
	private Map<String, IntensityFormula> namedFormulas;
	

	/**
		Create a new <code>RawDataType</code>.
		See the getter methods for a description of the parameters
	*/
	RawDataType(String id, String name, String description, int channels, 
		String table, List<RawDataProperty> properties, List<IntensityFormula> formulas)
	{
		this(id, name, description, channels, false, null, null, table, properties, formulas);
	}
	
	/**
		Create a new raw data type representing a file-only platform.
		@param platform The platform
	*/
	RawDataType(PlatformData platform)
	{
		this("platform." + platform.getExternalId(), platform.getName(), 
			platform.getDescription(), platform.getChannels(), true, 
			platform, null, null, null, null);
	}

	/**
		Create a new raw data type representing a file-only platform variant.
		@param variant The platform variant
	*/
	RawDataType(PlatformVariantData variant)
	{
		this("variant." + variant.getExternalId(), variant.getName(), 
				variant.getDescription(), variant.getChannels(), true, 
				variant.getPlatform(), variant, null, null, null);		
	}
	
	private RawDataType(String id, String name, String description, int channels, 
		boolean fileOnly, PlatformData platform, PlatformVariantData variant,
		String table, List<RawDataProperty> properties, List<IntensityFormula> formulas)
	{
		this.id = id;
		this.name = name;
		this.description = description;
		this.channels = channels;
		this.fileOnly = fileOnly;
		this.platform = platform;
		this.variant = variant;
		
		// For database raw data types
		if (!fileOnly)
		{
			if (table == null)
			{
				throw new InvalidDataException("No table specified for raw data type: " + id);
			}
			this.table = table;
			this.realTable = new RealTable(table, "raw");
			
			// Extended properties (=columns) of the table
			if (properties != null)
			{
				this.properties = Collections.unmodifiableList(properties);
				this.namedProperties = new HashMap<String, RawDataProperty>();
				for (RawDataProperty property : properties)
				{
					if (namedProperties.put(property.getName(), property) != null)
					{
						throw new BaseException("RawDataProperty[name="+property.getName()+"] is already defined for raw data type: "+name);
					}
				}
			}
			
			// Intensity formulas
			if (formulas != null)
			{
				this.formulas = Collections.unmodifiableList(formulas);
				this.namedFormulas = new HashMap<String, IntensityFormula>();
				for (IntensityFormula formula : formulas)
				{
					if (namedFormulas.put(formula.getName(), formula) != null)
					{
						throw new BaseException("IntensityFormula[name="+formula.getName()+"] is already defined for raw data type: "+name);
					}
				}
			}
		}
		if (this.formulas == null) this.formulas = Collections.emptyList();
		if (this.properties == null) this.properties = Collections.emptyList();
	}
	
	/**
		Get the id of this raw data type. This value is the same as the
		<code>id</code> attribute of the <code>&lt;raw-data-type&gt;</code>
		tag in the XML file.
		@return The id for this raw data type
	*/
	public String getId()
	{
		return id;
	}

	/**
		Get the platform that this raw data type is associated with. Only 
		raw data types representing file-only platforms are associated
		with that platform. If {@link #isStoredInDb()} returns true this
		method will always return null.
		
		@param dc The DbControl to use for database access
		@return A Platform item or null if this raw data type
			isn't associated with a platform
		@since 2.5
	*/
	public Platform getPlatform(DbControl dc)
	{
		return platform == null ? null : Platform.getById(dc, platform.getId());
	}

	/**
		Get the platform variant that this raw data type is associated with. Only 
		raw data types representing file-only variants are associated
		with that variant. 
		
		@param dc The DbControl to use for database access
		@return A Platform item or null if this raw data type
			isn't associated with a variant
		@since 2.5
	*/
	public PlatformVariant getVariant(DbControl dc)
	{
		return variant == null ? null : PlatformVariant.getById(dc, variant.getId());
	}

	/**
		Get the name of this raw data type. This value is the same as the
		<code>name</code> attribute of the <code>&lt;raw-data-type&gt;</code>
		tag in the XML file. The name is not used by the core, but should be used
		by client applications in the user interface.

		@return The name for this raw data type
	*/
	public String getName()
	{
		return name;
	}

	/**
		Update the name. Used mainly for file-only representations of
		raw data types when the underlying platform is changed.
		@since 2.16.2
	*/
	void setName(String name)
	{
		this.name = name;
	}
	
	/**
		Get a description of this raw data type. This value is the same as the
		<code>description</code> attribute of the <code>&lt;raw-data-type&gt;</code>
		tag in the XML file. The description is not used by the core, but can be used
		by client applications in help texts, etc.

		@return A description for this raw data type
	*/
	public String getDescription()
	{
		return description;
	}

	/**
		The number of channels in this raw data type. Channels are important
		in the core when it comes to the analysis phase. Some analysis methods
		only works for a specific number of channels (mostly two).
		This value is the same as the
		<code>channels</code> attribute of the <code>&lt;raw-data-type&gt;</code>
		tag in the XML file.

		@return The number of channels
	*/
	public int getChannels()
	{
		return channels;
	}

	/**
		If the raw data of this raw data type can be stored in the database
		or not.
		@see #getTableName()
	*/
	public boolean isStoredInDb()
	{
		return !fileOnly;
	}
	
	/**
		Get the name of the database table that is used to store the extra
		properties of this raw data type. This value is the same as the
		<code>table</code> attribute of the <code>&lt;raw-data-type&gt;</code>
		tag in the XML file. The table name is normally not useful for
		any client application.

		@return The name of the database table for this raw data type, or null
			if not stored in database
		@see #isStoredInDb()
	*/
	public String getTableName()
	{
		return table;
	}

	public String getEntityName()
	{
		return "rawdata_"+getId();
	}
	
	public RealTable getRealTable()
	{
		return realTable;
	}
	
	/**
		Get a list of {@link RawDataProperty}:s with definitions for the
		extra properties that has been defined for this raw data type.

		@return A <code>List</code> object containing <code>RawDataProperty</code>
			objects
	*/
	public List<RawDataProperty> getProperties()
	{
		return properties;
	}
	
	public RawDataProperty getProperty(String name)
	{
		return namedProperties.get(name);
	}

	public List<IntensityFormula> getIntensityFormulas()
	{
		return formulas;
	}
	
	public IntensityFormula getIntensityFormula(String name)
	{
		return namedFormulas.get(name);
	}

	public String toString()
	{
		return name;
	}
	
	public RawData getRawDataById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		if (!isStoredInDb())
		{
			throw new BaseException("Raw data for raw data type '" + this + "' is not stored in the database.");
		}
		RawData rd = HibernateUtil.loadData(dc.getStatelessSession(), 
			RawData.class, getEntityName(), id);
		if (rd == null) throw new ItemNotFoundException("RawData[id = " + id + "]");
		
		RawDataUtil.getRawBioAssay(dc, rd); // To check for read permission
		return rd;
	}

	/**
		Compare the id:s of the raw data types.
		Note! Before 2.16.2 the name was used for comparing but this
		led to issues when changing the name of file-only plaforms. See
		ticket #1567.
		@since 2.5
	*/
	@Override
	public int compareTo(RawDataType o) 
	{
		return id.compareTo(o.id);
	}

	@Override
	public boolean equals(Object obj)
	{
		if (this == obj) return true;
		if (obj == null) return false;
		if (obj.getClass() != this.getClass()) return false;
		RawDataType rdt = (RawDataType)obj;
		return id.equals(rdt.id);
	}

	@Override
	public int hashCode()
	{
		return id.hashCode();
	}
	
}
