/*
	$Id: User.java 5829 2011-10-27 06:16:24Z nicklas $

	Copyright (C) 2005 Samuel Andersson, Johan Enell, Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson
	Copyright (C) 2007 Johan Enell, Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.BasicData;
import net.sf.basedb.core.data.GroupData;
import net.sf.basedb.core.data.QuotaIndex;
import net.sf.basedb.core.data.RoleData;
import net.sf.basedb.core.data.UserData;
import net.sf.basedb.core.data.OwnableData;
import net.sf.basedb.core.data.QuotaData;
import net.sf.basedb.core.data.DirectoryData;
import net.sf.basedb.core.hibernate.TypeWrapper;
import net.sf.basedb.util.MD5;
import net.sf.basedb.util.bcrypt.BCrypt;
import net.sf.basedb.core.query.Restriction;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.EntityQuery;
import net.sf.basedb.core.query.ResultList;
import net.sf.basedb.core.query.Selects;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.Collections;


/**
	This class is used to represent a user in BASE.

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2011-10-27 08:16:24 +0200 (Thu, 27 Oct 2011) $
*/
public class User
	extends BasicItem<UserData>
	implements Nameable, Removable, SystemItem, Registered
{
	/**
		The type of item represented by this class.
		@see Item#USER
		@see #getType()
	*/
	public static final Item TYPE = Item.USER;

	/**
		The system id for the <code>User</code> representing the root account.
		The root user has full access to everything in BASE.
	*/
	public static final String ROOT = "net.sf.basedb.core.User.ROOT";

	/**
		This filter will limit a query to only return users
		which are members of at least one of the groups where the logged in user is a
		member unless the logged in user has generic read permission.
	*/
	private static final QueryRuntimeFilter RUNTIME_FILTER = new QueryRuntimeFilterImpl();

	/**
		Create a new <code>User</code> item.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@param login The login for the user (required)
		@param password The password for the user (required)
		@return The new <code>User</code> item
		@throws BaseException If there is an error
	*/
	public static User getNew(DbControl dc, String login, String password)
		throws BaseException
	{
		User u = dc.newItem(User.class);
		u.setName("New user");
		u.setLogin(login);
		u.setPassword(password);
		u.getData().setQuota(HibernateUtil.loadData(dc.getHibernateSession(), QuotaData.class, SystemItems.getId(Quota.DEFAULT)));
		u.getData().setEntryDate(new Date());
		return u;
	}

	/**
		Get a <code>User</code> item when you know the ID.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The ID of the item to load
		@return The <code>User</code> item
		@throws ItemNotFoundException If an item with the specified 
			ID is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static User getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		User u = dc.loadItem(User.class, id);
		if (u == null) throw new ItemNotFoundException("User[id="+id+"]");
		return u;
	}

	/**
		Load the ID:s of all users that are direct members of
		the specified groups.
	 	@param dc DbControl used to access database.
		@param groupIds The ID:s of the groups
		@return The ID:s of all members
		@see Group#getGroupsRecursive(DbControl, Set)
	*/
	public static Set<Integer> getAllMembers(DbControl dc, Set<Integer> groupIds)
	{
		return getAllMembers(dc.getHibernateSession(), groupIds);
	}
	
	static Set<Integer> getAllMembers(org.hibernate.Session session, Collection<Integer> groupIds)
	{
		Set<Integer> userIds = new HashSet<Integer>();
		if (groupIds != null && groupIds.size() > 0)
		{
			// Get the users which are members of the same groups
			org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "GET_USER_IDS_FOR_GROUPS");
			/*
				SELECT ug.userId
				FROM UserGroups ug
				WHERE ug.groupId IN (:groups)
			*/
			query.setParameterList("groups", groupIds, TypeWrapper.INTEGER.getHibernateType());
			userIds.addAll(HibernateUtil.loadList(Integer.class, query, null));
			
			// Get the users that have a quota group among the same groups
			query = HibernateUtil.getPredefinedQuery(session, "GET_USER_IDS_FOR_QUOTAGROUPS");
			/*
				SELECT usr.id
				FROM UserData usr
				WHERE usr.quotaGroup.id IN (:groups)
			*/
			query.setParameterList("groups", groupIds, TypeWrapper.INTEGER.getHibernateType());
			userIds.addAll(HibernateUtil.loadList(Integer.class, query, null));
		}
		return userIds;
	}

	/**
		Add all groups and roles that has been marked as default to
		the user.
		@since 2.4
	*/
	static void addDefultRolesAndGroups(org.hibernate.Session session, UserData user)
	{
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "LOAD_DEFAULT_GROUPS");
		/*
			SELECT grp
			FROM GroupData grp
			WHERE grp.default = true
		*/
		List<GroupData> defaultGroups = HibernateUtil.loadList(GroupData.class, query, null);
		for (GroupData group : defaultGroups)
		{
			group.getUsers().add(user);
		}
		query = HibernateUtil.getPredefinedQuery(session, "LOAD_DEFAULT_ROLES");
		/*
			SELECT rle
			FROM RoleData rle
			WHERE rle.default = true
		*/
		List<RoleData> defaultRoles = HibernateUtil.loadList(RoleData.class, query, null);
		for (RoleData role : defaultRoles)
		{
			role.getUsers().add(user);
		}		
	}
	
	
	/**
		Get a query configured to retrieve users. If the logged in user
		doesn't have generic permission to all users, only users
		that are members in at least one group where the logged in
		user is also a member are returned. The <code>Everyone</code> group
		is not considered since all users automatically are members to that group.
		
		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<User> getQuery()
	{
		return new ItemQuery<User>(User.class, RUNTIME_FILTER);
	}

	/**
	 	Check if a login name already exists in the database
	 	@param dc An open DbControl used to access the database.  
	  	@param login The login name to look for
	 	@return TRUE if an user has the login. FALSE if the login is not in use.
	*/
	public static boolean loginIsUsed(DbControl dc, String login)
	{
		if (login != null) login = login.trim();
		org.hibernate.Session session = dc.getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "GET_USER_FOR_LOGIN");
		/*
			SELECT usr
			FROM UserData usr
			WHERE usr.login = :login
		*/
		query.setString("login", login);
		UserData userData = HibernateUtil.loadData(UserData.class, query);
		return userData != null;
	}
	
	/**
	  	Check if an external ID already exists in the database
	  	@param dc An open DbControl used to access the database.
	  	@param externalId The externalID to look for
	  	@return TRUE if an user has the external ID. FALSE if the externalID is not in use.
	*/	
	public static boolean externalIdIsUsed(DbControl dc, String externalId)
	{
		if (externalId != null) externalId = externalId.trim();
		org.hibernate.Session session = dc.getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "GET_USER_FOR_EXTERNAL_ID");
		/*
			SELECT usr 
			FROM UserData usr
			WHERE usr.externalId = :externalId
		*/
		query.setParameter("externalId", externalId);
		UserData userData = HibernateUtil.loadData(UserData.class, query);
		return userData != null;
	}
	
	/**
		Encrypt the plain-text password. The password is ecnrypted
		by first calculating the MD5 of the password and then
		using bcrypt with a random salt on the MD5.
		
		@param password The plain-text password
	*/
	static String encryptPassword(String password)
	{
		String md5 = MD5.getHashString(password);
		return BCrypt.hashpw(md5, BCrypt.gensalt());
	}
	
	/**
		Check the plain-text password against the crypted password.
		@param password The plain-text password
		@param cryptedPassword The crypted password
	 	@return TRUE if the password is correct, FALSE if not
	*/
	static boolean checkPassword(String password, String cryptedPassword)
	{
		String md5 = MD5.getHashString(password);
		return BCrypt.checkpw(md5, cryptedPassword);
	}
	
	User(UserData userData)
	{
		super(userData);
	}

	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the Nameable interface
		-------------------------------------------
	*/
	public String getName()
	{
		return getData().getName();
	}
	public void setName(String name)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setName(getData(), name);
	}
	public String getDescription()
	{
		return getData().getDescription();
	}
	public void setDescription(String description)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.RESTRICTED_WRITE);
		NameableUtil.setDescription(getData(), description);
	}
	// -------------------------------------------
	/*
		From the Removable interface
		-------------------------------------------
	*/
	public boolean isRemoved()
	{
		return getData().isRemoved();
	}
	public void setRemoved(boolean removed)
		throws PermissionDeniedException
	{
		checkPermission(removed ? Permission.DELETE : Permission.WRITE);
		getData().setRemoved(removed);
	}
	// -------------------------------------------
	/*
		From the SystemItem interface
		-------------------------------------------
	*/
	public String getSystemId()
	{
		return getData().getSystemId();
	}
	public boolean isSystemItem()
	{
		return getSystemId() != null;
	}
	// -------------------------------------------
	/*
		From the Registered interface
		-------------------------------------------
	*/
	public Date getEntryDate()
	{
		return DateUtil.copy(getData().getEntryDate());
	}
	// -------------------------------------------
	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		Checks if:
		<ul>
		<li>any {@link Ownable} item is owned by this user
		</ul>
		A user can also be referenced from groups, roles, projects,
		item keys, sessions and settings, but those references are 
		automatically deleted if the user is deleted and aren't inclued 
		in this check.
	*/
	public boolean isUsed()
		throws BaseException
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_OWNABLE_ITEMS_FOR_USER");
		/*
			SELECT o
			FROM net.sf.basedb.core.data.OwnableData o
			WHERE o.owner = :user
		*/
		query.setEntity("user", this.getData());
		return HibernateUtil.loadData(OwnableData.class, query) != null || super.isUsed();
	}
	/**
		Get all:
		<ul>
		<li>{@link Ownable} items owned by this user
		<ul>
		@since 2.2
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		org.hibernate.Session session = getDbControl().getHibernateSession();
	
		// Ownable items
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_OWNABLE_ITEMS_FOR_USER");
			/*
				SELECT o
				FROM net.sf.basedb.core.data.OwnableData o
				WHERE o.owner = :user
			*/
		query.setEntity("user", this.getData());
		addUsingItems(using, query);
		return using;
	}
	
	
	/**
		If the logged in user is the same as this user RESTRICTED_WRITE permissions is granted,
		unless it is a multiuser account.
		If this is a system user, delete and create permissions are revoked.
		Finally READ permission is granted to if the user is a member of at least
		one group where the logged in user is also a member.
	*/
	void initPermissions(int granted, int denied)
		throws BaseException
	{
		if (isSystemItem()) 
		{
			denied |= Permission.deny(Permission.DELETE, Permission.CREATE);
		}
		if (getSessionControl().getLoggedInUserId() == getId())
		{
			granted |= Permission.grant(isMultiuserAccount() ? Permission.READ : Permission.RESTRICTED_WRITE);
		}
		if (getSessionControl().isFriendOf(this))
		{
			granted |= Permission.grant(Permission.READ);
		}
		super.initPermissions(granted, denied);
	}
	/**
		Make sure plug-ins always has at least read access to the logged in user.
		@since 2.9
	*/
	PluginPermission getPluginPermissions()
	{
		PluginPermission pp = super.getPluginPermissions();
		if (getId() == getSessionControl().getLoggedInUserId())
		{
			pp = new PluginPermission(pp.granted | Permission.grant(Permission.READ), pp.denied);
		}
		return pp;
	}
	// -------------------------------------------

	/**
		Set the password. 
	 	@param password Password to set. A null password is not allowed.
		@throws PermissionDeniedException If the logged in user doesn't have
			restricted_write permission
		@throws InvalidDataException If the new password is null
	*/
	public void setPassword(String password)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.RESTRICTED_WRITE);
		if (password == null) throw new InvalidUseOfNullException("password");
		getData().getPassword().setCryptedPassword(encryptPassword(password));
	}

	/**
		Get the expire date of the account. When the expiration date have been passed
		the user can't login. A null value indicates that the account will 
		never expire.
	 	@return A {@link java.util.Date} object or null. 
	*/
	public Date getExpirationDate()
	{
		return DateUtil.copy(getData().getExpirationDate());
	}
	/**
		Sets the expiration date of the account. A null value indicates that 
		the account will never expire.
	 	
	 	@param expirationDate Date when the account should expire.
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
	*/
	public void setExpirationDate(Date expirationDate)
		throws PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		getData().setExpirationDate(DateUtil.setNullableDate(expirationDate, "expirationDate"));
	}
	
	/**
		Check if this account has been disabled. It is not possible to login
		if the account is disabled.
		@return TRUE if the account is disabled, FALSE otherwise
	*/
	public boolean isDisabled()
	{
		return getData().isDisabled();
	}
	/**
		Disables or enables the account.
		@param disabled TRUE to disabled the account, FALSE to enabled it
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@see #isDisabled()
	*/
	public void setDisabled(boolean disabled)
		throws PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		getData().setDisabled(disabled);
	}

	/**
		Check if the user account is a multiuser account or not. Multiuser accounts
		don't have write permissions for contact information and settings.
		@return TRUE if the user account is a multiuser account, FALSE otherwise
	*/
	public boolean isMultiuserAccount()
	{
		return getData().isMultiuserAccount();
	}

	/**
		Sets if the user account is multiuser account. 
	 	@param multiuserAccount Set to TRUE if it is, FALSE otherwise. 
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
	*/
	public void setMultiuserAccount(boolean multiuserAccount)
		throws PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		getData().setMultiuserAccount(multiuserAccount);
	}

	/**
		The maximum length of the external ID that can be stored in the database.
		@see #setExternalId(String)
	*/
	public static final int MAX_EXTERNAL_ID_LENGTH = UserData.MAX_EXTERNAL_ID_LENGTH;
	/**
		Get the external id for the user account. The external id is intended to
		be used by external applications which need to synchronize data between
		the BASE database and some external database. It is not used by the core.
	 	@return A String object or null if not set. 
	*/
	public String getExternalId()
	{
		return getData().getExternalId();
	}
	/**
		Set the external id for the user account.
	 	@param externalId External id to be set, null is allowed. 
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws InvalidDataException If the new value is longer than
			{@link #MAX_EXTERNAL_ID_LENGTH}
		@throws BaseException If there is another error
	*/
	public void setExternalId(String externalId)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		checkPermission(Permission.WRITE);
		getData().setExternalId(StringUtil.setNullableString(externalId, "externalId", MAX_EXTERNAL_ID_LENGTH));
	}

	/**
		The maximum length of the login that can be stored in the database.
		@see #setLogin(String)
	*/
	public static final int MAX_LOGIN_LENGTH = UserData.MAX_LOGIN_LENGTH;
	/**
		Get the login for the user account.
		@return A String object.
	*/
	public String getLogin()
	{
		return getData().getLogin();
	}
	/**
		Set the login for the user account.
	 	@param login Login name to set. Null is not allowed. 
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws InvalidDataException If the new value is null or longer than
			{@link #MAX_LOGIN_LENGTH}
	*/
	public void setLogin(String login)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		getData().setLogin(StringUtil.setNotNullString(login, "login", MAX_LOGIN_LENGTH));
	}

	/**
		The maximum length of the organisation that can be stored in the database.
		@see #setOrganisation(String)
	*/
	public static final int MAX_ORGANISATION_LENGTH = UserData.MAX_ORGANISATION_LENGTH;
	/**
		Get the organisation this user works for, or null if unknown.
		@return A String object or null if not set.
	*/
	public String getOrganisation()
	{
		return getData().getOrganisation();
	}
	/**
		Set the organisation this user works for, or null if unknown.
	 	@param organisation Some kind of name or description of 
	 		the organisation to associate with the user. Not longer then {@value #MAX_ORGANISATION_LENGTH} 
		@throws PermissionDeniedException If the logged in user doesn't have
			restricted write permission
		@throws InvalidDataException If the new value is longer than
			{@link #MAX_ORGANISATION_LENGTH}
	*/
	public void setOrganisation(String organisation)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.RESTRICTED_WRITE);
		getData().setOrganisation(StringUtil.setNullableString(organisation, "organisation", MAX_ORGANISATION_LENGTH));
	}

	/**
		The maximum length of the address that can be stored in the database.
		@see #setAddress(String)
	*/
	public static final int MAX_ADDRESS_LENGTH = UserData.MAX_ADDRESS_LENGTH;
	/**
		Get the address for the user, or null if unknown.
	 	@return A String object, or null if not set. 
	*/
	public String getAddress()
	{
		return getData().getAddress();
	}
	/**
		Set the address for the user, or null if unknown.
	 	@param address Address to be set. 
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws InvalidDataException If the new value is longer than
			{@link #MAX_ADDRESS_LENGTH}
	*/
	public void setAddress(String address)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.RESTRICTED_WRITE);
		getData().setAddress(StringUtil.setNullableString(address, "address", MAX_ADDRESS_LENGTH));
	}

	/**
		The maximum length of the phone that can be stored in the database.
		@see #setPhone(String)
	*/
	public static final int MAX_PHONE_LENGTH = UserData.MAX_PHONE_LENGTH;
	/**
		Get the phone number to the user, or null if unknown.
	 	@return A String object or null if not set.
	*/
	public String getPhone()
	{
		return getData().getPhone();
	}
	/**
		Set the phone number to the user, or null if unknown.
	 	@param phone Phone number to set. 
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws InvalidDataException If the new value is longer than
			{@link #MAX_PHONE_LENGTH}
	*/
	public void setPhone(String phone)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.RESTRICTED_WRITE);
		getData().setPhone(StringUtil.setNullableString(phone, "phone", MAX_PHONE_LENGTH));
	}

	/**
		The maximum length of the fax that can be stored in the database.
		@see #setFax(String)
	*/
	public static final int MAX_FAX_LENGTH = UserData.MAX_FAX_LENGTH;
	/**
		Get the fax number to the user, or null if unknown.
	 	@return A String object or null if not set. 
	*/
	public String getFax()
	{
		return getData().getFax();
	}
	/**
		Set the fax number to the user, or null if unknown.
	 	@param fax Fax number to be set. 
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws InvalidDataException If the new value is longer than
			{@link #MAX_FAX_LENGTH}
	*/
	public void setFax(String fax)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.RESTRICTED_WRITE);
		getData().setFax(StringUtil.setNullableString(fax, "fax", MAX_FAX_LENGTH));
	}

	/**
		The maximum length of the email address that can be stored in the database.
		@see #setEmail(String)
	*/
	public static final int MAX_EMAIL_LENGTH = UserData.MAX_EMAIL_LENGTH;
	/**
		Get the email address to the user.
	 	@return A java.lang.String object, or null if unknown.
	*/
	public String getEmail()
	{
		return getData().getEmail();
	}
	/**
		Set the email address to the user, or null if unknown.
	 	@param email Email to be set, null is allowed. 
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws InvalidDataException If the new value is longer than
			{@link #MAX_EMAIL_LENGTH}
	*/
	public void setEmail(String email)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.RESTRICTED_WRITE);
		getData().setEmail(StringUtil.setNullableString(email, "email", MAX_EMAIL_LENGTH));
	}

	/**
		If this flag has been set, messges to the user is also sent as an
		email. This requires that the server admin has configured an outgoing
		smtp server and that this user has specified a valid email address.
		@return TRUE to send messages as email, FALSE if not
		@since 2.16
	*/
	public boolean getSendMessagesAsEmail()
	{
		return getData().getSendMessagesAsEmail();
	}
	/**
		Set a flag indicating that all messages should also be sent as
		email. See {@link #getSendMessagesAsEmail()} for information
		about restrictions.
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@since 2.16
	*/
	public void setSendMessagesAsEmail(boolean sendMessagesAsEmail)
	{
		checkPermission(Permission.RESTRICTED_WRITE);
		getData().setSendMessagesAsEmail(sendMessagesAsEmail);
	}
	
	/**
		The maximum length of the url that can be stored in the database.
		@see #setUrl(String)
	*/
	public static final int MAX_URL_LENGTH = UserData.MAX_URL_LENGTH;
	/**
		Get the URL to the user's homepage.
		@return A String object, or null if unknown.
	*/
	public String getUrl()
	{
		return getData().getUrl();
	}
	/**
		Set the URL to the user's homepage, or null if unknown.
	 	@param url URL to set. The String can be at maximum {@value #MAX_URL_LENGTH}.
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws InvalidDataException If the new value is longer than
			{@link #MAX_URL_LENGTH}
	*/
	public void setUrl(String url)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.RESTRICTED_WRITE);
		getData().setUrl(StringUtil.setNullableString(url, "url", MAX_URL_LENGTH));
	}
	
	/**
	 	Gets an extended property with a certain name
	 	@param name The name of the property
	 	@return An Object or 
	 		null if no extended property exists with the name
	 	@since 2.4
	*/
	public Object getExtended(String name)
	{
		return getData().getExtended(name);
	}
	/**
		Sets an extended property
		@param name The name to call the extended property by
		@param value The value the extended property should have
		@throws PermissionDeniedException If the logged in user doesn't
		 	have @link {Permission#WRITE} permission for the user.
		@since 2.4
	*/
	public void setExtended(String name, Object value)
	{
		checkPermission(Permission.RESTRICTED_WRITE);
		getData().setExtended(name, value);
	}

	/**
		Get the {@link Quota} that applies to the user.
		@return A <code>Quota</code> item
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#READ} permission to the item
	 	@throws BaseException If there is some other kind of error. 
	*/
	public Quota getQuota() 
		throws PermissionDeniedException, BaseException 
	{
		return getDbControl().getItem(Quota.class, getData().getQuota());
	}
	
	/**
		Set the quota for the user.
		@param quota The new <code>Quota</code>
		@throws InvalidDataException If the quota is null
		@throws PermissionDeniedException If the logged in user doesn't
			have {@link Permission#WRITE} permission for the user
			or {@link Permission#USE} permission for the quota
	*/
	public void setQuota(Quota quota)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		if (quota == null) throw new InvalidUseOfNullException("quota");
		quota.checkPermission(Permission.USE);
		getData().setQuota(quota.getData());
	}
	
	/**
		Get the {@link Group} whose {@link Quota} also applies
		to the user.
		@return A <code>Group</code> item, or null if no group has been specified
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#READ} permission to the item
	 	@throws BaseException If there is some other kind of error. 
	*/
	public Group getQuotaGroup()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(Group.class, getData().getQuotaGroup());
	}

	/**
		Set the group whose quota should be checked for disk consuming
		items.
		@param quotaGroup The new <code>Group</code>
		@throws PermissionDeniedException If the logged in user doesn't
			have {@link Permission#WRITE} permission for the user
			or {@link Permission#USE} permission for the group
	*/
	public void setQuotaGroup(Group quotaGroup)
		throws PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		if (quotaGroup != null) quotaGroup.checkPermission(Permission.USE);
		getData().setQuotaGroup(quotaGroup == null ? null : quotaGroup.getData());
	}

	/**
		Get the used number of bytes for the specified quota type and location.
		@param quotaType The {@link QuotaType}
		@param location The location
		@return The number of bytes that have been used
	 	@throws BaseException If one of the required parameter is null or
	 		if there is some other kind of error.
	*/
	public long getDiskUsage(QuotaType quotaType, Location location)
		throws BaseException
	{
		if (quotaType == null) throw new InvalidUseOfNullException("quotaType");
		if (location == null) throw new InvalidUseOfNullException("location");
		return getDbControl().getDiskUsage(getData(), quotaType.getData(), location);
	}
	
	/**
		Check if the user has used more quota than what has been assigned. This
		method will check both the user's primary quota and, if it exists, the
		group quota. The most restrictive quota is used.
		
		@param quotaType The quota type to check
		@param location The location
		@return TRUE if the quota has been exceeded, FALSE otherwise
		@since 2.4
	*/
	public boolean hasExceededQuota(QuotaType quotaType, Location location)
	{
		QuotaData userQuota = getData().getQuota();
		GroupData quotaGroup = getData().getQuotaGroup();
		QuotaData groupQuota = quotaGroup == null ? null : quotaGroup.getQuota();
		QuotaIndex index = new QuotaIndex(quotaType.getData(), location.getValue());
		
		DbControl dc = getDbControl();
		Long assigned = userQuota.getQuotaValues().get(index);
		
		if (groupQuota != null)
		{
			Long groupAssigned = groupQuota.getQuotaValues().get(index);
			if (groupAssigned != null)
			{
				if (assigned == null)
				{
					assigned = groupAssigned;
				}
				else if (groupAssigned != Quota.UNLIMITED && groupAssigned < assigned)
				{
					assigned = groupAssigned;
				}
			}
		}
		long used = dc.getDiskUsage(getData(), quotaType.getData(), location);
		boolean hasExceeded = false;
		if (assigned == null)
		{
			hasExceeded = QuotaType.TOTAL.equals(quotaType.getSystemId());
		}
		else if (assigned != Quota.UNLIMITED && assigned <= used)
		{
			hasExceeded = true;
		}
		return hasExceeded;
	}

	/**
		Get the home {@link Directory} for this user.
		@return A <code>Directory</code> item, or null if no home directory
			has been specified
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#READ} permission to the directory
	 	@throws BaseException If there is some other kind of error. 
	*/
	public Directory getHomeDirectory()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(Directory.class, getData().getHomeDirectory());
	}

	/**
		Set the home directory for the user.
		@param homeDirectory The new home <code>Directory</code>
		@throws PermissionDeniedException If the logged in user doesn't
			have {@link Permission#WRITE} permission for the user
			or {@link Permission#USE} permission for the directory
		@throws InvalidDataException If the home directory is not a subdirectory
			to /root/home
	*/
	public void setHomeDirectory(Directory homeDirectory)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		if (homeDirectory != null) 
		{
			homeDirectory.checkPermission(Permission.USE);
			DirectoryData parent = homeDirectory.getData().getParent();
			if (parent == null || parent.getId() != SystemItems.getId(Directory.HOME))
			{
				throw new InvalidDataException("homeDirectory is not a subdirectory to /root/home");	
			}
		}
		getData().setHomeDirectory(homeDirectory == null ? null : homeDirectory.getData());
	}
	
	/**
		Add this user to the groups and roles that has been marked as
		default for new users.
		@throws BaseException If there is a problem
		@since 2.4
	*/
	public void addToDefaultRolesAndGroups()
		throws BaseException
	{
		addDefultRolesAndGroups(getDbControl().getHibernateSession(), getData());
	}
	
	/**
		Get a query that returns the roles where this user is a
		member. The query excludes roles that the logged in user doesn't have 
		permission to read.
		@return a restricted {@link net.sf.basedb.core.ItemQuery}.
		@see Role#getQuery()
	*/
	public ItemQuery<Role> getRoles()
	{
		ItemQuery<Role> query = Role.getQuery();
		query.joinPermanent(Hql.innerJoin("users", Item.USER.getAlias()));
		query.restrictPermanent(
			Restrictions.eq(
				Hql.alias(Item.USER.getAlias()), 
				Hql.entity(this)
			)
		);
		return query;
	}

	/**
		Get a query that returns the groups where this user is a
		member. The query excludes groups that the logged in user doesn't have 
		permission to read.
		@return A restricted {@link net.sf.basedb.core.ItemQuery}.
		@see Group#getQuery()
	*/
	public ItemQuery<Group> getGroups()
	{
		ItemQuery<Group> query = Group.getQuery();
		query.joinPermanent(Hql.innerJoin("users", Item.USER.getAlias()));
		query.restrictPermanent(
			Restrictions.eq(
				Hql.alias(Item.USER.getAlias()), 
				Hql.entity(this)
			)
		);
		return query;
	}

	/**
		Get a query that returns the projects where this user is a
		member. The query excludes projects that the logged in user doesn't have 
		permission to read. The query doesn't include projects where this user is
		the owner.
		@return A restricted {@link net.sf.basedb.core.ItemQuery}
		@see Project#getQuery()
	*/
	public ItemQuery<Project> getProjects()
	{
		ItemQuery<Project> query = Project.getQuery();
		query.joinPermanent(Hql.innerJoin("users", Item.USER.getAlias()));
		query.restrictPermanent(
			Restrictions.eq(
				Hql.index(Item.USER.getAlias(), null), 
				Hql.entity(this)
			)
		);
		return query;
	}

	/**
		Create a new default setting for this user.
		@param name The name of the setting
		@param value The value of the setting
		@return The new <code>UserDefaultSetting</code> object
		@throws PermissionDeniedException If the logged in user doesn't have
			restricted write permission
		@throws InvalidDataException If the name or value is null or 
			longer than what is allowed
		@throws BaseException If there is another error
		@see UserDefaultSetting#getNew(DbControl, User, String, String)
	*/
	public UserDefaultSetting newDefaultSetting(String name, String value)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		checkPermission(Permission.RESTRICTED_WRITE);
		return UserDefaultSetting.getNew(getDbControl(), this, name, value);
	}

	/**
		Get a query that returns all default settings for this user.
		@return An {@link ItemQuery} object
		@see UserDefaultSetting#getQuery(User)
	*/
	public ItemQuery<UserDefaultSetting> getDefaultSettings()
	{
		return UserDefaultSetting.getQuery(this);
	}

	/**
		Create a new client application setting for this user.
		@param client The client application
		@param name The name of the setting
		@param value The value of the setting
		@return The new <code>UserDefaultSetting</code> object
		@throws PermissionDeniedException If the logged in user doesn't have
			restricted write permission
		@throws InvalidDataException If the name or value is null or 
			longer than what is allowed
		@throws BaseException If there is another error
		@see UserClientSetting#getNew(DbControl, User, Client, String, String)
	*/
	public UserClientSetting newClientSetting(Client client, String name, String value)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		checkPermission(Permission.RESTRICTED_WRITE);
		return UserClientSetting.getNew(getDbControl(), this, client, name, value);
	}

	/**
		Get a query that returns the settings for the
		specified client.

		@param client The client application
		@return An {@link ItemQuery} object
		@see UserClientSetting#getQuery(User, Client)
	*/
	public ItemQuery<UserClientSetting> getClientSettings(Client client)
	{
		return UserClientSetting.getQuery(this, client);
	}
	
	/**
		Get a query that returns all messages for the user.
		@return An {@link ItemQuery} object
		@see Message#getQuery(User)
	*/
	public ItemQuery<Message> getMessages()
	{
		return Message.getQuery(this);
	}

	/**
		@see #getItems(Set, int, int, Permission, Restriction)
	*/
	public ResultList<Ownable> getItems(Item itemType, int firstItem, int maxItems, Permission permission)
	{
		return getItems(itemType == null ? null : Collections.singleton(itemType), 
			firstItem, maxItems, permission, null);
	}
	
	/**
		Load the items owned by this user. If this user is not the same as the
		currently logged in user this method will only load items that are
		shared to the logged in user with the given permission.
	
		@param itemTypes Limit the list to only return items of the specified types,
			use null or an empty set to return items of all types
		@param firstItem The index of the first item to return (0-based)
		@param maxItems The maximum number of items to return, or 0 to return all items
		@param permission The permission the logged in user should have on
			the items
		@param restriction An optional restriction that is applied to the
			queries to further reduce the list
		@return A list containing ownable items
		@since 2.15
	*/
	public ResultList<Ownable> getItems(Set<Item> itemTypes, int firstItem, int maxItems, 
		Permission permission, Restriction restriction)
	{
		DbControl dc = getDbControl();
		SessionControl sc = dc.getSessionControl();
		org.hibernate.Session session = dc.getHibernateSession();
	
		// Check if this is the logged in user -- add a sharing filter otherwise
		if (this.getId() != sc.getLoggedInUserId())
		{
			org.hibernate.Filter sharedToFilter = null;
			if (sc.getActiveProjectId() != 0)
			{
				sharedToFilter = HibernateUtil.enableFilter(session, "sharedToOrInProject");
				Set<Integer> projectKeys = sc.getProjectKeys(permission);
				if (projectKeys == null || projectKeys.size() == 0)
				{
					projectKeys = QueryRuntimeFilterFactory.ZERO_SET;
				}
				sharedToFilter.setParameterList("projectKeys", projectKeys);
			}
			else
			{
				sharedToFilter = HibernateUtil.enableFilter(session, "sharedTo");
			}
			Set<Integer> itemKeys = sc.getItemKeys(permission);
			if (itemKeys == null || itemKeys.size() == 0) 
			{
				itemKeys = QueryRuntimeFilterFactory.ZERO_SET;
			}
			sharedToFilter.setParameter("owner", sc.getLoggedInUserId());
			sharedToFilter.setParameterList("itemKeys", itemKeys);
		}

		List<OwnableData> ownedBy = null;
		if (itemTypes != null && itemTypes.size() > 0)
		{
			ownedBy = new ArrayList<OwnableData>();
			for (Item item : itemTypes)
			{
				if (OwnableData.class.isAssignableFrom(item.getDataClass()))
				{
					// Query to load all ownable items of a specific subclass
					AbstractHqlQuery hql = 
						new AbstractHqlQuery(item.getDataClass().getName(), "item", false){};
					hql.select(Selects.expression(Hql.alias("item"), null, true));
					hql.restrict(Hql.restriction("item.owner = :user", null));
					hql.setParameter("user", getId(), Type.INT);
					if (restriction != null) hql.restrict(restriction);
					/*
						SELECT item
						FROM <class> item
						WHERE item.owner = :user
						[AND restriction]
					*/
					ownedBy.addAll(HibernateUtil.loadList(OwnableData.class, hql.getMainHqlQuery(dc), sc));
				}
			}
		}
		else
		{
			AbstractHqlQuery hql = 
				new AbstractHqlQuery("net.sf.basedb.core.data.OwnableData", "item", false){};
			hql.select(Selects.expression(Hql.alias("item"), null, true));
			hql.restrict(Hql.restriction("item.owner = :user", null));
			hql.setParameter("user", getId(), Type.INT);
			if (restriction != null) hql.restrict(restriction);
			/*
				SELECT item
				FROM net.sf.basedb.core.data.OwnableData item
				WHERE item.owner = :user
				[AND restriction]
			*/
			ownedBy = HibernateUtil.loadList(OwnableData.class, hql.getMainHqlQuery(dc), sc);
		}
		// Disable filters
		HibernateUtil.disableFilter(session, "sharedTo");
		HibernateUtil.disableFilter(session, "sharedToOrInProject");
		
		// Create items and apply limits
		List<Ownable> items = new ArrayList<Ownable>(ownedBy.size());
		int index = 0;
		int totalItems = 0;
		boolean useLimit = firstItem >= 0 && maxItems > 0;
		int lastItem = firstItem + maxItems;
		for (OwnableData od : ownedBy)
		{
			BasicData data = (BasicData)od;
			try
			{
				BasicItem item = dc.getItem(BasicItem.class, data);
				if (!useLimit || (index >= firstItem && index < lastItem))
				{
					items.add((Ownable)item);
				}
				index++;
				totalItems++;
			}
			catch (PermissionDeniedException ex)
			{}
		}
		return new UnmodifiableResultList<Ownable>(items, totalItems);
	}
	
	/**
		A runtime filter implementation that limits a query to only
		return users which are members of at least one group where the logged in 
		user is also a member unless the logged in user has generic read permission.
	*/
	private static class QueryRuntimeFilterImpl
		implements QueryRuntimeFilter
	{
		public void enableFilters(QueryRuntimeFilterManager manager, EntityQuery query, DbControl dc)
		{
			SessionControl sc = dc.getSessionControl();
			if (sc.isDenied(query.getItemPermission(), Item.USER))
			{
				manager.enableFilter("denyAll");
			}
			if (!sc.hasPermission(query.getItemPermission(), Item.USER))
			{
				if (sc.isLoggedIn())
				{
					// Load users that are friends of the logged in user
					org.hibernate.Filter filter = manager.enableFilter("memberOf");
					if (filter != null)
					{
						Set<Integer> friends = query.getItemPermission() == Permission.READ ? 
							sc.getFriends() : null;
						if (friends == null || friends.size() == 0) friends = Collections.singleton(0);
						filter.setParameterList("items", friends);
						filter.setParameter("owner", sc.getLoggedInUserId());
					}
				}
				else 
				{
					// Do not load any users if not logged in
					manager.enableFilter("denyAll");
				}
			}
		}
	}
}

