/*
	$Id: RawBioAssayData.java 5818 2011-10-21 11:15:41Z nicklas $
	
	Copyright (C) 2005 Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/
	
	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.
	
	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.data;

import java.util.Date;
import java.util.HashSet;
import java.util.Set;
import java.util.Map;
import java.util.HashMap;

/**
	This class holds information about a raw bioassay, which is the
	raw data extracted from a scan.
	
	@author Nicklas
	@version 2.0
	@see net.sf.basedb.core.RawBioAssay
	@see <a href="../../../../../../html/developer/api/data_api.bioassays.html">Developer documentation: Bioassays and raw data</a>
	@base.modified $Date: 2011-10-21 13:15:41 +0200 (Fri, 21 Oct 2011) $
	@hibernate.class table="`RawBioAssays`" lazy="true"
*/
public class RawBioAssayData 
	extends AnnotatedData
	implements DiskConsumableData, FileStoreEnabledData, RegisteredData
{

	public RawBioAssayData()
	{}

	/*
		From the DiskConsumableData interface
		-------------------------------------------
	*/
	private DiskUsageData diskUsage;
	public DiskUsageData getDiskUsage()
	{
		if(diskUsage == null) diskUsage = new DiskUsageData();
		return diskUsage;
	}
	void setDiskUsage(DiskUsageData diskUsage)
	{
		this.diskUsage = diskUsage;
	}
	// -------------------------------------------
	/*
		From the FileStoreEnabledData interface
		-------------------------------------------
	*/
	private FileSetData fileSet;
	public FileSetData getFileSet()
	{
	   return fileSet;
	}
	public void setFileSet(FileSetData fileSet)
	{
	   this.fileSet = fileSet;
	}
	// -------------------------------------------
	/*
		From the RegisteredData interface
		-------------------------------------------
	*/
	private Date entryDate;
	public Date getEntryDate()
	{
		return entryDate;
	}
	public void setEntryDate(Date entryDate)
	{
		this.entryDate = entryDate;
	}
	// -------------------------------------------	

	private PlatformData platform;
	/**
		Get the platform this raw bioassay uses.
		@since 2.5
		@hibernate.many-to-one outer-join="false" 
		@hibernate.column name="`platform_id`" not-null="true"
	*/
	public PlatformData getPlatform()
	{
		return platform;
	}
	public void setPlatform(PlatformData platform)
	{
		this.platform = platform;
	}
	
	private PlatformVariantData variant;
	/**
		Get the platform variant this raw bioassay uses, or null.
		@since 2.5
		@hibernate.many-to-one outer-join="false" 
		@hibernate.column name="`variant_id`"
	*/
	public PlatformVariantData getVariant()
	{
		return variant;
	}
	public void setVariant(PlatformVariantData variant)
	{
		this.variant = variant;
	}
	
	private JobData job;
	/**
		Get the job that was used to import raw data to this raw bioassay
		@hibernate.many-to-one column="`job_id`" not-null="false" outer-join="false"
		@since 2.7
	*/
	public JobData getJob()
	{
		return job;
	}
	/**
		@since 2.7
	*/
	public void setJob(JobData job)
	{
		this.job = job;
	}
	
	private SoftwareData software;
	/**
		Get the software used to analyse the scanned image and produce the
		raw data.
		@hibernate.many-to-one column="`software_id`" not-null="false" outer-join="false"
	*/
	public SoftwareData getSoftware()
	{
		return software;
	}
	public void setSoftware(SoftwareData software)
	{
		this.software = software;
	}

	private ProtocolData protocol;
	/**
		Get the protocol used for the image analysis.
		@hibernate.many-to-one column="`protocol_id`" not-null="false" outer-join="false"
	*/
	public ProtocolData getProtocol()
	{
		return protocol;
	}
	public void setProtocol(ProtocolData protocol)
	{
		this.protocol = protocol;
	}

	private ArrayDesignData arrayDesign;
	/**
		Get the array design used for the hybridization this raw bio assay comes from.
		@hibernate.many-to-one column="`arraydesign_id`" not-null="false" outer-join="false"
	*/
	public ArrayDesignData getArrayDesign()
	{
		return arrayDesign;
	}
	public void setArrayDesign(ArrayDesignData arrayDesign)
	{
		this.arrayDesign = arrayDesign;
	}
	
	private SpotImagesData spotImages;
	/**
		Get the settings used for generating spot images for this raw bio assay.
		@hibernate.one-to-one property-ref="rawBioAssay" cascade="delete"
	*/
	public SpotImagesData getSpotImages()
	{
		return spotImages;
	}
	void setSpotImages(SpotImagesData spotImages)
	{
		this.spotImages = spotImages;
	}
	
	private DerivedBioAssayData bioAssay;
	/**
		Get the derived bioassay that this raw bioassay has been created from.
		@hibernate.many-to-one column="`bioassay_id`" not-null="false" outer-join="false"
	*/
	public DerivedBioAssayData getParentBioAssay()
	{
		return bioAssay;
	}
	public void setParentBioAssay(DerivedBioAssayData bioAssay)
	{
		this.bioAssay = bioAssay;
	}

	private ExtractData extract;
	/**
		Get the derived bioassay that this raw bioassay has been created from.
		@hibernate.many-to-one column="`extract_id`" not-null="false" outer-join="false"
	*/
	public ExtractData getParentExtract()
	{
		return extract;
	}
	public void setParentExtract(ExtractData extract)
	{
		this.extract = extract;
	}

	
	private String rawDataType;
	/**
		Get the name of the raw data type used by this raw bio assay.
		@hibernate.property column="`rawdatatype`" type="string" length="255" not-null="true"
	*/
	public String getRawDataType()
	{
		return rawDataType;
	}
	public void setRawDataType(String rawDataType)
	{
		this.rawDataType = rawDataType;
	}
	
	private boolean hasData;
	/**
		If the raw bioassay has data or not.
		@hibernate.property column="`has_data`" type="boolean" not-null="true"
	*/
	public boolean getHasData()
	{
		return hasData;
	}
	public void setHasData(boolean hasData)
	{
		this.hasData = hasData;
	}

	private int spots;
	/**
		The number of data spots in this raw bio assay, that
		are stored in the database.
		@hibernate.property column="`spots`" type="int" not-null="true"
		@since 2.5
	*/
	public int getSpots()
	{
		return spots;
	}
	public void setSpots(int spots)
	{
		this.spots = spots;
	}

	private int fileSpots;
	/**
		The number of spots on this raw bioassay that are stored inside
		files.
		@hibernate.property column="`file_spots`" type="int" not-null="true"
		@since 2.5
	*/
	public int getNumFileSpots()
	{
		return fileSpots;
	}
	public void setNumFileSpots(int fileSpots)
	{
		this.fileSpots = fileSpots;
	}
	
	private long bytes;
	/**
		The number of bytes the raw data occupies on disk. 
		@hibernate.property column="`bytes`" type="long" not-null="true"
	*/
	public long getBytes()
	{
		return bytes;
	}
	public void setBytes(long bytes)
	{
		this.bytes = bytes;
	}
	
	public static final int MAX_HEADER_NAME_LENGTH = 255;
	public static final int MAX_HEADER_VALUE_LENGTH = 65535;
	private Map<String, String> headers;
	/**
		File headers from the raw data file.
		@hibernate.map table="`RawBioAssayHeaders`" lazy="true" cascade="all"
		@hibernate.collection-key column="`rawbioassay_id`"
		@hibernate.collection-index column="`name`" type="string" length="255" not-null="true"
		@hibernate.collection-element column="`value`" type="text" not-null="true"
	*/
	public Map<String, String> getHeaders()
	{
		if (headers == null) headers = new HashMap<String, String>();
		return headers;
	}
	void setHeaders(Map<String, String> headers)
	{
		this.headers = headers;
	}

	private Set<ExperimentData> experiments;
	/**
		This is the inverse end. We keep the inverse="true" attribute because we don't
		want to remove a raw bioassay that is used by an experiment.
		@see ExperimentData#getRawBioAssays()
		@hibernate.set table="`ExperimentRawBioAssays`" lazy="true" inverse="true"
		@hibernate.collection-key column="`rawbioassay_id`"
		@hibernate.collection-many-to-many column="`experiment_id`" class="net.sf.basedb.core.data.ExperimentData"
	*/
	Set<ExperimentData> getExperiments()
	{
		return experiments;
	}
	void setExperiments(Set<ExperimentData> experiments)
	{
		this.experiments = experiments;
	}
	private Set<TransformationData> transformations;
	/**
		This is the inverse end. We keep the inverse="true" attribute because we don't
		want to remove a raw bioassay that is used by a transformation.
		@see TransformationData#getRawSources()
		@hibernate.set table="`TransformationRawBioAssays`" lazy="true" inverse="true"
		@hibernate.collection-key column="`rawbioassay_id`"
		@hibernate.collection-many-to-many column="`transformation_id`" class="net.sf.basedb.core.data.TransformationData"
	*/
	Set<TransformationData> getTransformations()
	{
		return transformations;
	}
	void setTransformations(Set<TransformationData> transformations)
	{
		this.transformations = transformations;
	}


	private Set<BioAssayData> bioAssays;
	/**
		This is the inverse end. We keep the inverse="true" attribute because we don't
		want to remove a raw bioassay that is used by a bioassay.
		@see BioAssayData#getRawParents()
		@hibernate.set table="`BioAssayRawParents`" lazy="true" inverse="true"
		@hibernate.collection-key column="`parent_id`"
		@hibernate.collection-many-to-many column="`bioassay_id`" class="net.sf.basedb.core.data.BioAssayData"
	*/
	Set<BioAssayData> getBioAssays()
	{
		if (bioAssays == null) bioAssays = new HashSet<BioAssayData>(); 
		return bioAssays;
	}
	void setBioAssays(Set<BioAssayData> bioAssays)
	{
		this.bioAssays = bioAssays;
	}
}
