/*
	$Id: ReporterData.java 5818 2011-10-21 11:15:41Z nicklas $

	Copyright (C) 2005 Samuel Andersson, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.data;

import java.util.Date;
import java.util.Map;
import java.util.HashMap;
import java.util.Set;

/**
	This class represents the idea of clone, oligo or some other
	reporter for a gene. Each reporter must have a unique
	id, the reporter ID. The only required property, except the
	reporter ID is the name.
	<p>
	The server administrator may also add columns in the database
	for other properties of the reporters. The new columns are
	mapped to BASE in the <code>extended-properties.xml</code> file.
	For more information about this see the {@link net.sf.basedb.core.ExtendedProperties}
	class.

	@author Samuel
	@version 2.0
	@see ExtendableData
	@see net.sf.basedb.core.Reporter
	@see net.sf.basedb.core.ReporterBatcher
	@see net.sf.basedb.core.ExtendedProperties
	@see <a href="../../../../../../html/developer/api/data_api.reporters.html">Developer documentation: Reporters</a>
	@hibernate.class table="`Reporters`" lazy="true"
*/
public class ReporterData
	extends BasicData
	implements NameableData, ExtendableData, BatchableData, RegisteredData
{

	public ReporterData()
	{}

	/**
		Create a reporter object with a given ID.
		@param id The reporter ID
		@since 2.8
	*/
	public ReporterData(int id)
	{
		setId(id);
	}
	
	/*
		From the RegisteredData interface
		-------------------------------------------
	*/
	private Date entryDate;
	public Date getEntryDate()
	{
		return entryDate;
	}
	void setEntryDate(Date entryDate)
	{
		this.entryDate = entryDate;
	}
	// -------------------------------------------
	
	/**
		The maximum length of the external ID that can be stored in the database.
		@see #setExternalId(String)
	*/
	public static final int MAX_EXTERNAL_ID_LENGTH = 255;
	private String externalId;
	/**
		Get the external id for this <code>ReporterData</code> item.
		@return The reporter's external ID.
		@hibernate.property column="`external_id`" type="string" length="255" not-null="true" unique="true"
	*/
	public String getExternalId()
	{
		return externalId;
	}
	public void setExternalId(String externalId)
	{
		this.externalId = externalId;
	}
	
	/**
		The maximum length of the gene symbol that can be stored in the
		database. 
	*/
	public static final int MAX_SYMBOL_LENGTH = 255;
	private String symbol;
	/**
		Get the gene symbol of the reporter.
		@return The gene symbol of the reporter
		@hibernate.property column="`symbol`" type="string" length="255"
	*/
	public String getSymbol()
	{
		return symbol;
	}
	public void setSymbol(String symbol)
	{
		this.symbol = symbol != null ? symbol.trim() : null;
	}
	
	/**
		The last update date.
	*/
	private Date lastUpdate;
	/**
		Get the date and time the information about this reporter
		last was updated.
		@return A <code>Date</code> object
		@hibernate.property column="`last_update`" type="timestamp" not-null="true"
	*/
	public Date getLastUpdate()
	{
		return lastUpdate;
	}
	void setLastUpdate(Date lastUpdate)
	{
		this.lastUpdate = lastUpdate;
	}
	
	/**
		The maximum length of the last source that can be stored in the
		database. 
	*/
	public static final int MAX_LAST_SOURCE_LENGTH = 255;
	/**
	 	The source of the last update
	 */
	private String lastSource;
	/**
	 	Get the source that was used when this reporter
	 	was last updated.
	 	@return A <code>String</code> object
	 	@hibernate.property column="`last_source`" type="string" not-null="false" length="255"
	 */
	public String getLastSource()
	{
		return lastSource;
	}
	void setLastSource(String source)
	{
		this.lastSource = source;
	}
	
	/*
		From the NameableData interface
		-------------------------------------------
	*/	
	private String name;
	public String getName()
	{
		return name;
	}
	public void setName(String name)
	{
		this.name = name;
	}
	private String description;
	public String getDescription()
	{
		return description;
	}
	public void setDescription(String description)
	{
		this.description = description;
	}
	// -------------------------------------------
	
	/*
		From the ExtendableData interface
		------------------------------------------
	*/
	private Map<String, Object> extendedProperties;
	public Object getExtended(String name)
	{
		return extendedProperties == null ? null : extendedProperties.get(name);
	}
	public void setExtended(String name, Object value)
	{
		if (extendedProperties == null) extendedProperties = new HashMap<String, Object>();
		extendedProperties.put(name, value);
	}
	// -------------------------------------------
	
	private ReporterTypeData reporterType;
	/**
		Get the {@link ReporterTypeData} of this the reporter. Package private since
		we cannot expose the data object to client applications.
		@return The <code>ReporterTypeData</code> item
		@hibernate.many-to-one column="`reportertype_id`" not-null="false" outer-join="false"
		@see net.sf.basedb.core.Reporter#getReporterType(net.sf.basedb.core.DbControl, ReporterData)
	*/
	ReporterTypeData getReporterType()
	{
		return reporterType;
	}
	void setReporterType(ReporterTypeData reporterType)
	{
		this.reporterType = reporterType;
	}
	
	private Set<ReporterListScoreData> reporterListScores;
	/**
		The lists this reporter is used in. This is the inverse end.
		@see ReporterListData#getReporterListScores()
		@hibernate.set lazy="true" inverse="true"
		@hibernate.collection-key column="`reporter_id`"
		@hibernate.collection-one-to-many class="net.sf.basedb.core.data.ReporterListScoreData"
	*/
	Set<ReporterListScoreData> getReporterListScores()
	{
		return reporterListScores;
	}
	void setReporterListScores(Set<ReporterListScoreData> reporterListScores)
	{
		this.reporterListScores = reporterListScores;
	}
	
	
}
