/**
	$Id: SavePointWork.java 5881 2011-11-21 10:54:26Z nicklas $

	Copyright (C) 2011 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.hibernate;

import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Savepoint;


/**
	A wrapper work that protects a parent work with a savepoint. 
	Before executing the parent work, a {@link Savepoint} is
	created on the {@link Connection}. If the parent work
	results throws an {@link SQLException} the connection is
	rolled back to the savepoint. Other exceptions and a normal
	completion just releases the savepoint.

	@author nicklas
	@since 3.1
	@base.modified $Date: 2011-11-21 11:54:26 +0100 (Mon, 21 Nov 2011) $
*/
public class SavePointWork<R>
	implements JdbcWork<R>
{

	private final JdbcWork<R> work;
	
	/**
		Create a new work item.
		@param work The parent work
	*/
	public SavePointWork(JdbcWork<R> work)
	{
		this.work = work;
	}
	
	/**
		From the Work interface
		-----------------------
	*/
	@Override
	public void execute(Connection connection) 
		throws SQLException
	{
		Savepoint savepoint = null;
		try
		{
			savepoint = connection.setSavepoint();
			work.execute(connection);
		}
		catch (SQLException ex)
		{
			if (savepoint != null) 
			{
				connection.rollback(savepoint);
				savepoint = null;
			}
			throw ex;
		}
		finally
		{
			if (savepoint != null) 
			{
				connection.releaseSavepoint(savepoint);
			}
		}
	}
	// --------------------------
	
	/*
		From the JdbcWork interface
		---------------------------
	*/
	/**
		@return The result from the parent work
	*/
	@Override
	public R getResult()
	{
		return work.getResult();
	}
	// ----------------------------

}
