/**
	$Id: AbstractExporterPlugin.java 4523 2008-09-15 09:18:48Z nicklas $

	Copyright (C) 2006 Nicklas Nordborg, Martin Svensson
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.plugin;

import java.io.IOException;
import java.util.Arrays;

import net.sf.basedb.core.BooleanParameterType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Directory;
import net.sf.basedb.core.File;
import net.sf.basedb.core.ItemAlreadyExistsException;
import net.sf.basedb.core.Path;
import net.sf.basedb.core.PathParameterType;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.User;
import net.sf.basedb.util.FileExportOutputStream;

/**
	An abstract base class for all exporter plugins that wish to implement
	the {@link ImmediateDownloadExporter} interface. This class implements
	that interface, but you have to implement the {@link #performExport(ExportOutputStream, ProgressReporter)}
	instead which has the same requirements as the {@link ImmediateDownloadExporter#doExport(ExportOutputStream, ProgressReporter)} 
	method.
	<p>
	This class also implements the {@link #run(Request, Response, ProgressReporter)} method
	but if you want to use it you must use the parameters for the file path and
	overwrite option as they are defined in this class. Use the {@link #getSaveAsParameter(String, String, String, boolean)}
	and {@link #getOverwriteParameter(String, String)} to generate the parameter
	objects and then the {@link #getSaveAsPath(DbControl)} and {@link #getOverwrite()}
	to get values entered by the user.
	<p>
	This class doesn't override any methods of the {@link InteractivePlugin}
	interface. It is your responsibility to ask for the correct parameters
	and save them to the job. See the {@link net.sf.basedb.plugins.HelpExporter} 
	implementation forexample code.

	@author nicklas
	@version 2.2
	@base.modified $Date: 2008-09-15 11:18:48 +0200 (Mon, 15 Sep 2008) $
*/
public abstract class AbstractExporterPlugin
	extends AbstractPlugin
	implements ImmediateDownloadExporter
{

	protected static final PathParameterType optionalSaveAsType = 
		new PathParameterType(Path.Type.FILE, null, false);

	protected static final PathParameterType requiredSaveAsType = 
		new PathParameterType(Path.Type.FILE, null, true);
	
	protected static final BooleanParameterType overwriteType = 
		new BooleanParameterType(null, false);
	
	private File saveTo;
	
	/**
		Create a new <code>AbstractExporterPlugin</code>.
	*/
	protected AbstractExporterPlugin()
	{}

	/*
		From the Plugin interface
		-------------------------------------------
	*/
	/**
		@return MainType.EXPORT
	*/
	public MainType getMainType()
	{
		return MainType.EXPORT;
	}
	/**
		Checks the {@link #SAVE_AS} parameter for the path to save
		the exported data. If the file doesn't exists it is created. 
		If the file exists the {@link #OVERWRITE} parameter is checked
		if overwriting is ok. Files marked for deletion will always be
		overwritten. If no path has been specified an error response is sent.
		<p>
		Then, the subclass is called to generate the export. The following methods
		are called:
		<ol>
		<li>{@link #begin(DbControl)}
		<li>{@link #performExport(ExportOutputStream, ProgressReporter)}
		<li>{@link #end(boolean)} (after the DbControl has been committed)
		<li>{@link #getSuccessMessage()} (only if no error ocurred)
		</ol>
	*/
	public void run(Request request, Response response, ProgressReporter progress)
	{
		String command = request.getCommand();
		if (command.equals(Request.COMMAND_EXECUTE))
		{
			DbControl dc = null;
			try
			{
				dc = sc.newDbControl();
				
				// Get or create the file where the exported data should be saved
				Path path = getSaveAsPath(dc);
				if (path == null)
				{
					response.setError("No path has been specified for saving the data.", null);
					return;
				}
				boolean overwrite = getOverwrite();
				saveTo = File.getByPath(dc, path, true);
				if (saveTo.isInDatabase() && !overwrite && !saveTo.isRemoved())
				{	
					// Not allowed to overwrite the existing file
					throw new ItemAlreadyExistsException("File[path="+path+"]");			
				}
				else if (!saveTo.isInDatabase())
				{
					dc.saveItem(saveTo);
				}
				
				// Create a stream for exporting
				ExportOutputStream out = new FileExportOutputStream(saveTo, true);
				begin(dc);
				performExport(out, progress);
				dc.commit();
				end(true);
				response.setDone(getSuccessMessage());
			}
			catch (Throwable t)
			{
				end(false);
				response.setError(t.getMessage(), Arrays.asList(t));
			}
			finally
			{
				saveTo = null;
				if (dc != null) dc.close();
			}
		}
		else
		{
			response.setError("Unknown command: "+command, null);
		}
	}
	// -------------------------------------------
	/*
		From the ImmediateDownloadExporter interface
		---------------------------------------------
	*/
	/**
		Implements immediate downloading. Calls the subclass methods in the following
		order:
		<ol>
		<li>{@link #begin(DbControl)}
		<li>{@link #performExport(ExportOutputStream, ProgressReporter)}
		<li>{@link #end(boolean)} (after the DbControl has been committed)
		</ol>
	*/
	public final void doExport(ExportOutputStream out, ProgressReporter progress)
		throws IOException
	{
		DbControl dc = sc.newDbControl();
		boolean success = false;
		try
		{
			begin(dc);
			performExport(out, progress);
			dc.commit();
			success = true;
		}
		finally
		{
			if (dc != null) dc.close();
			end(success);
		}
	}
	// ---------------------------------------------

	/**
		Called just before the export starts. A subclass may override 
		this method if it needs to initialise some resources.
		@param dc The DbControl that is used to write to the database.
			A subclass may use this read from the database as well
		@see #end(boolean)
	*/
	protected void begin(DbControl dc)
	{}
	
	/**
		Get the file object that the current export is writing to. This method
		may return a file object that already exists in the database, a new
		file that is saved when the transaction is committed or null if the
		export is a 'download immediately' export.
		@return A file object or null
		@since 2.7
	*/
	protected File getSaveTo()
	{
		return saveTo;
	}
	
	/**
		Perform the export. This method replaces the 
		{@link ImmediateDownloadExporter#doExport(ExportOutputStream, ProgressReporter)}
		method.
		@param out Output stream to write the exported data to
		@param progress An optional progress reporter
	*/
	protected abstract void performExport(ExportOutputStream out, ProgressReporter progress)
		throws IOException;
	
	/**
		Called just after everything has been committed to the database,
		or immediately after an error has ocurred. A subclass should clean
		up any resources aquired in the {@link #begin(DbControl)} method here.
		@param success TRUE if the file was parsed successfully, FALSE otherwise
		@see #begin(DbControl)
	*/
	protected void end(boolean success)
	{}

	/**
		Called if the export was successful to let the subclass generate a simple
		message that is sent back to the core and user interface. An example
		message might by: <code>178 reporters exported successfully</code>.
		The default implementation always return null.
	*/
	protected String getSuccessMessage()
	{
		return null;
	}
	
	/**
		Parameter name for the {@link #getSaveAsParameter(String, String, String, boolean)} 
		parameter.
	*/
	protected static final String SAVE_AS = "saveAs";

	/**
		Get a plugin parameter that asks for the path to a file where the exported
		data should be saved. If no path is specified by the user the data will 
		be downloaded immediately.
		
		@param label The parameter label, or null to use the default (Save as)
		@param description The parameter description, or null to use the default 
			(The path to the file where the exported data should be placed. 
			Leave empty to download immediately.)
		@param defaultValue A default filename, or null to not use any default
		@param required If the parameter is required or optional
	*/
	protected PluginParameter<String> getSaveAsParameter(String label, String description, 
		String defaultValue, boolean required)
	{
		if (label == null) label = "Save as";
		if (description == null) 
		{
			description = "The path to the file where the exported data should be placed. ";
			if (!required) description += " Leave empty to download immediately.";
		}
		PluginParameter<String> saveAsParameter = new PluginParameter<String>
			(
				SAVE_AS, label, description, defaultValue, 
				required ? requiredSaveAsType : optionalSaveAsType
			);
		return saveAsParameter;
	}

	/**
	 	Get the configured path where the exported data should be saved.
		@param dc A DbControl which is needed to find the currently logged in user
			if the path starts with '<b>~/</b>' = Home directory of logged in user
		@return A path or null if no path has been specified
	*/
	protected Path getSaveAsPath(DbControl dc)
	{
		Path path = null;
		String sPath = (String)job.getValue(SAVE_AS);
		if (sPath != null)
		{
			if (sPath.startsWith("~/"))
			{
				sPath = sPath.substring(1);
				User user = User.getById(dc, sc.getLoggedInUserId());
				Directory home = user.getHomeDirectory();
				if (home != null) sPath = "~" + home.getName() + sPath;
			}
			path = new Path(sPath, Path.Type.FILE);			
		}
		return path;
	}

	/**
		Parameter name for the {@link #getOverwriteParameter(String, String)} parameter.
	*/
	protected static final String OVERWRITE = "overwrite";
	
	/**
		Get a plugin parameter that asks if existing files can be overwritten
		or not.
		
		@param label The parameter label, or null to use the default (Overwrite)
		@param description The parameter description, or null to use the default 
			(If it is allowed to overwrite an existing file or not.)
	*/
	protected PluginParameter<Boolean> getOverwriteParameter(String label, String description)
	{
		if (label == null) label = "Overwrite";
		if (description == null) 
		{
			description = "If it is allowed to overwrite an existing file or not.";
		}
		PluginParameter<Boolean> overwriteParameter = new PluginParameter<Boolean>
		(
			OVERWRITE, label, description, overwriteType
		);
		return overwriteParameter;
	}
	
	/**
		Get the configured value for the overwrite parameter.
		@return A boolean, if no values has been specified false is returned
		@see #getOverwriteParameter(String, String)
	*/
	public boolean getOverwrite()
	{
		Boolean value = (Boolean)job.getValue(OVERWRITE);
		return value == null ? false : value;
	}
	
	/**
	  	See if file path already exists and if it's allowed to overwrite it.	  	
	  	@param strPath The file path to be checked.
	  	@param overwrite TRUE If it is allowed to overwrite the file, FALSE otherwise.
	  	@return FALSE If the file path exists and isn't allowed to be overwritten or TRUE otherwise.
	*/
	protected boolean pathCanBeUsed(String strPath, boolean overwrite)
	{		
		DbControl dc = null;
		try
		{
			dc = sc.newDbControl();		
			Path path = new Path(strPath, Path.Type.FILE);
			File file = File.getByPath(dc, path, true);
			
			return (overwrite || !file.isInDatabase() || file.isRemoved());
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
}
