/**
	$Id $

	Copyright (C) 2010 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.plugin;

import java.util.List;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.InvalidDataException;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.ParameterType;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.PluginConfiguration;
import net.sf.basedb.core.PluginDefinition;

/**
	A helper class that wraps a parameter values object with
	a namespace. All write operations will be made to the specified
	namespace. Eg, the 'name' parameter is replaced with the namespace 
	+ dot + the original name. For example, if namespace="export" and
	name="toDir", the resulting name is "export.toDir". Read operations will 
	first check the namespace if the parameter exists, otherwise the parameter 
	in the global namespace is used. All other methods simply forward to the 
	parent parameter values object.
	<p>
	The main usage for this class is when a master plug-in needs to use
	other child plug-ins. By wrapping the parameter values with an instance 
	of this class it makes it possible to store all parameters together
	without having to worry about name clashes among the children.
	
	@author nicklas
	@since 2.15
	@base.modified $Date $
	@see NamespacePluginParameter
	@see NamespaceRequestWrapper
*/
public class NamespaceParameterValuesWrapper
	implements ParameterValues
{

	private final String namespace;
	private final ParameterValues parent;
	
	/**
		Create a new namespace wrapper for the parent parameter
		values.
		@param parent The parent parameter values, null is not allowed
		@param namespace The namespace, null is not allowed
	*/
	public NamespaceParameterValuesWrapper(ParameterValues parent, String namespace)
	{
		if (parent == null) throw new NullPointerException("parent");
		if (namespace == null) throw new NullPointerException("namespace");
		this.parent = parent;
		this.namespace = namespace;
	}
	
	@Override
	public int getId()
	{
		return parent.getId();
	}
	@Override
	public Job getJob()
	{
		return parent.getJob();
	}
	@Override
	public PluginConfiguration getPluginConfiguration()
	{
		return parent.getPluginConfiguration();
	}
	@Override
	public PluginDefinition getPluginDefinition()
	{
		return parent.getPluginDefinition();
	}

	@Override
	public Object getValue(String name)
		throws PermissionDeniedException, BaseException
	{
		Object value = parent.getValue(namespace + "." + name);
		if (value == null) value = parent.getValue(name);
		return value;
	}
	
	@Override
	public List<?> getValues(String name)
		throws PermissionDeniedException, BaseException
	{
		List<?> values = parent.getValues(namespace + "." + name);
		if (values == null) values = parent.getValues(name);
		return values;
	}

	@Override
	public <T> void setValue(String name, ParameterType<T> type, T value)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		parent.setValue(namespace + "." + name, type, value);
	}

	@Override
	public <T> void setValues(String name, ParameterType<T> type, List<T> values)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		parent.setValues(namespace + "." + name, type, values);
	}

}
