/*
	$Id $

	Copyright (C) 2010 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util;

import java.io.InputStream;
import java.io.OutputStream;

/**
	Runnable implementation that copies data from an input stream
	to an output stream when executed. This is useful when executing
	external processes since reading and writing to their stdin and
	stdout is best done with separate thread.

	@author nicklas
	@since 2.15
	@base.modified $Date $
*/
public class FileCopyRunnable
	implements Runnable
{

	private final InputStream in;
	private final OutputStream out;
	private Throwable error;
	
	/**
		Create a new file copy object. The stream are automatically closed 
		in case of an error
		
		@param in The input stream to read data from (required)
		@param out The output stream to write data to (optional; if not
			given the input stream is read and then discarded)
	*/
	public FileCopyRunnable(InputStream in, OutputStream out)
	{
		if (in == null) throw new NullPointerException("in");
		this.in = in;
		this.out = out;
	}
	
	@Override
	public void run()
	{
		try
		{
			if (out == null)
			{
				FileUtil.read(in);
			}
			else
			{
				FileUtil.copy(in, out);
			}
		}
		catch (Throwable t)
		{
			error = t;
		}
		finally
		{
			FileUtil.close(in);
			FileUtil.close(out);
		}
	}

	/**
		A boolean flag indicating if the copying was sucessful or not.
		This method should only be called after the thread has ended.
		@see #getError()
	*/
	public boolean hasError()
	{
		return error != null;
	}
	
	/**
		Get the exception that was thrown when executing the file copy
		operation. This method should only be called after the thread 
		has ended.
		@return An error or null if the operation was sucessful
		@see #hasError()
	*/
	public Throwable getError()
	{
		return error;
	}
	
}
