/**
	$Id: AnnotationModel.java 5384 2010-08-13 08:50:27Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.bfs;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import net.sf.basedb.util.Values;

/**
	A model container for BFS annotation files. This class is primarily
	intended to be used when parsing small annotation files. It is not
	recommended to use this class with large files since it may use too
	much memory.
	<p>
	Data may be retreived by either column name or index and by row id
	or index. Lookup by index is a lot faster than by name or id.

	@author Nicklas
	@version 2.15
	@base.modified $Date: 2010-08-13 10:50:27 +0200 (Fri, 13 Aug 2010) $
*/
public class AnnotationModel
	implements EventHandler
{
	
	private String[] headers;
	private List<String[]> data;
	
	private Map<String, Integer> columnIndex;
	private Map<Integer, Integer> rowIndex;

	/**
		Create a new annotation model.
	*/
	public AnnotationModel()
	{
		this.data = new ArrayList<String[]>();
		this.rowIndex = new HashMap<Integer, Integer>();
	}
	
	/*
		From the EventHandler interface
		--------------------------------
	*/
	@Override
	public <T> void handleEvent(EventType<T> eventType, T eventData, BfsParser parser)
	{
		if (eventType == AnnotationParser.HEADER_EVENT)
		{
			setHeaders((String[])eventData);
		}
		else if (eventType == AnnotationParser.DATA_EVENT)
		{
			addRow((String[])eventData);
		}
	}
	// ------------------------------------

	
	/**
		Set the column headers. It is expected that each
		array element is unique.
		@param headers An array with the column headers
	*/
	public void setHeaders(String... headers)
	{
		this.headers = headers;
		this.columnIndex = new HashMap<String, Integer>(headers.length);
		for (int i = 0; i < headers.length; ++i)
		{
			columnIndex.put(headers[i], i);
		}
	}
	
	/**
		Get all column headers.
	*/
	public String[] getHeaders()
	{
		return headers;
	}
	
	/**
		Add a row of data to the model. The number of array elements
		in the array should be the same as the number of columns. The
		first array element must be the ID of the row. This must be
		a unique positive integer.
	*/
	public void addRow(String... row)
	{
		int id = Values.getInt(row[0]);
		data.add(row);
		rowIndex.put(id, data.size()-1);
	}
	
	/**
		Get the number of column.
	*/
	public int getColumnCount()
	{
		return headers.length;
	}
	
	/**
		Get the number of rows.
	*/
	public int getRowCount()
	{
		return data.size();
	}
	
	/**
		Get the column index for a given header.
		@param header The column header
		@return The column index, or -1 if no column is found
	*/
	public int getColumnIndex(String header)
	{
		Integer index = columnIndex.get(header);
		return index == null ? -1 : index;
	}
	
	/**
		Get the column header for a given column index.
		@param colIndex The index of the column, a value between
			0 and {@link #getColumnCount()}-1
		@return The column header
	*/
	public String getHeader(int colIndex)
	{
		return headers[colIndex];
	}
	
	/**
		Get the row index for a given row id.
		@param rowId The row id
		@return The row index, or -1 if no row with the given id exists
	*/
	public int getRowIndex(int rowId)
	{
		Integer index = rowIndex.get(rowId);
		return index == null ? -1 : index;
	}

	/**
		Get the row id for a given row.
		@param rowIndex The index of the row, a value between 0 and 
			{@link #getRowCount()}-1
		@return The row id
	*/
	public int getRowId(int rowIndex)
	{
		String[] row = getRowByIndex(rowIndex);
		return Values.getInt(row[0]);
	}
	
	/**
		Get the data element at a given row and column index.
		@param rowIndex The row index, a value between 0 and 
			{@link #getRowCount()}-1
		@param colIndex The column index, a value between
			0 and {@link #getColumnCount()}-1
		@return The data at the given location
	*/
	public String getData(int rowIndex, int colIndex)
	{
		String[] row = getRowByIndex(rowIndex);
		return row[colIndex];
	}
	
	/**
		Get the data element at a given row and column. 
		Do not use this method repeatedly with the same
		row id or column header. It is faster to use {@link
		#getData(int, int)}.
		@param rowId The row id
		@param header The column header
		@return The data at the given location, or null if the
			location doesn't exists
	*/
	public String getData(int rowId, String header)
	{
		String[] row = getRowById(rowId);
		return row[getColumnIndex(header)];
	}
	
	public String[] getRowById(int rowId)
	{
		int rowIndex = getRowIndex(rowId);
		return getRowByIndex(rowIndex);
	}
	
	public String[] getRowByIndex(int rowIndex)
	{
		return data.get(rowIndex);
	}


}
