/**
	$Id: ProxyActionFactory.java 5946 2012-02-03 13:18:47Z nicklas $

	Copyright (C) Authors contributing to this file.

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.extensions.debug;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.util.HashMap;
import java.util.Map;

import net.sf.basedb.util.Values;
import net.sf.basedb.util.extensions.Action;
import net.sf.basedb.util.extensions.ActionFactory;
import net.sf.basedb.util.extensions.ExtensionPoint;
import net.sf.basedb.util.extensions.InvokationContext;
import net.sf.basedb.util.extensions.xml.PathSetter;
import net.sf.basedb.util.extensions.xml.VariableSetter;

/**
	Generic action factory class that can generate instances of
	actions if the action class is an interface. Use this factory
	for development and debugging only. The performance and
	memory requirements of this implementation is unsure for a 
	production environment.
	<p>
	
	This class works ONLY if the {@link ExtensionPoint#getActionClass()}
	is an interface, eg. {@link Class#isInterface()} return true. If so,
	we use Java reflection to generate a {@link Proxy} that implements
	the desired action class interface.
	<p>
	
	Parameters in the XML file will be passed on as return values for
	method calls on the proxy, following the usual syntax rules. Eg.
	The parameter value <code>&lt;icon&gt;/images/icon.png&lt;/icon&gt; will
	be returned as a result of calling <code>getIcon()</code> on the
	proxy. Null is returned for parameters that have not been specified 
	in the XML file.
	<p>
	
	The factory can convert the string values to other data types for the 
	following classes: int, long, float, double and boolean. It can handle 
	both primitive and wrapper classes. Example:
	<code>&lt;enabled&gt;1&lt;/enabled&gt; will return <code>true</code> if
	<code>isEnabled()</code> is called on the proxy.
	<p>
	
	<b>Parameters supported by this factory</b><br>
	This factory supports all parameters. Some parameters may trigger special
	actions:
	
	<ul>
	<li>disabled: If <code>true</code> the {@link #prepareContext(InvokationContext)} 
		method returns false
	</ul>

	@author nicklas
	@version 2.7
	@base.modified $Date: 2012-02-03 14:18:47 +0100 (Fri, 03 Feb 2012) $
*/
public class ProxyActionFactory
	implements ActionFactory<Action>, InvocationHandler
{

	private boolean disabled;
	private Action proxy;
	private Action[] actions;
	private Map<String, String> parameters;
	private Map<String, Object> valueCache;
	private volatile boolean initialised;
	
	/**
		Creates a new proxy factory.
	*/
	public ProxyActionFactory()
	{}

	/*
		From the ActionFactory interface
		--------------------------------
	*/
	@Override
	public boolean prepareContext(InvokationContext<? super Action> context)
	{
		if (disabled) return false;
		if (!initialised)
		{
			initProxy(context.getExtensionPoint().getActionClass());
		}
		return true;
	}

	
	@Override
	public Action[] getActions(InvokationContext context)
	{
		return disabled ? null : actions;
	}
	// -----------------------------------
	
	/*
		From the InvokationHandler interface
		------------------------------------
	*/
	@Override
	public Object invoke(Object proxy, Method method, Object[] args)
		throws Throwable
	{
		String parameterName = getParameterName(method.getName());
		return getParameterValue(parameterName, method.getReturnType());
	}
	// -------------------------------------
	
	/**
		Set generic parameters.
		@param name The name of the parameter
		@param value The value of the parameter
	*/
	@PathSetter
	@VariableSetter
	public void setParameter(String name, String value)
	{
		if (parameters == null) 
		{
			parameters = new HashMap<String, String>();
		}
		parameters.put(name, value);
	}
	
	/**
		Sets the disabled/enabled status of this factory.
		@param disabled A string that is parsed to a boolean
			by {@link Values#getBoolean(String)}
	*/
	public void setDisabled(String disabled)
	{
		this.disabled = Values.getBoolean(disabled);
	}
	
	/**
		Initialise the proxy class.
		@param actionClass The interface that the proxy must implement
	*/
	private synchronized void initProxy(Class<? extends Action> actionClass)
	{
		if (initialised) return;
		try
		{
			if (!actionClass.isInterface())
			{
				throw new IllegalArgumentException("Action class " + actionClass + " must be an interface");
			}
			ClassLoader loader = actionClass.getClassLoader();
			proxy = actionClass.cast(Proxy.newProxyInstance(loader, new Class[] { actionClass }, this));
			actions = new Action[] { proxy };
		}
		finally
		{
			initialised = true;
		}
	}

	/**
		Convert a method name to a parameter name. Methods that
		start with <code>get</code> or <code>is</code> will have that part
		removed. The first character of the remaining name is converted to 
		lower case. All other method names are left as they are.
		@param methodName The name of the method
		@return The name of the parameter
	*/
	private String getParameterName(String methodName)
	{
		int index = 0;
		if (methodName.startsWith("get"))
		{
			index = 3;
		}
		else if (methodName.startsWith("is"))
		{
			index = 2;
		}
		if (index > 0)
		{
			String firstChar = methodName.substring(index, index+1).toLowerCase();
			methodName = firstChar + methodName.substring(index+1);
		}
		return methodName;
	}
	
	/**
		Get the value of a parameter.
		@param parameterName The name of the parameter
		@return The value of the parameter
	*/
	private Object getParameterValue(String parameterName, Class<?> returnType)
	{
		if (valueCache == null)
		{
			valueCache = new HashMap<String, Object>();
		}
		// The value to return
		Object value = null;

		// Check the cache first
		if (valueCache.containsKey(parameterName))
		{
			value = valueCache.get(parameterName);
		}
		else
		{
			// Get the string value from the XML file
			String sValue = parameters.get(parameterName);
			// Convert to proper object type
			value = StringConverter.convertString(returnType, sValue);
			// Save the converted value in the cache
			valueCache.put(parameterName, value);
		}
		return value;
	}
}
