/**
	$Id: RegisterExtensionsProcessor.java 5691 2011-08-11 10:45:58Z nicklas $

	Copyright (C) 2011 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.extensions.manager.processor;

import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import net.sf.basedb.util.FileUtil;
import net.sf.basedb.util.extensions.Extension;
import net.sf.basedb.util.extensions.ExtensionPoint;
import net.sf.basedb.util.extensions.Registry;
import net.sf.basedb.util.extensions.manager.ExtensionKey;
import net.sf.basedb.util.extensions.manager.ExtensionPointKey;
import net.sf.basedb.util.extensions.manager.ExtensionsFile;
import net.sf.basedb.util.extensions.manager.ExtensionsFileProcessor;
import net.sf.basedb.util.extensions.manager.ExtensionsManager;
import net.sf.basedb.util.extensions.manager.FactoryParametersKey;
import net.sf.basedb.util.extensions.manager.ProcessResults;
import net.sf.basedb.util.extensions.manager.ExtensionsFile.WriteableExtensionsFile;
import net.sf.basedb.util.extensions.manager.filter.ValidAndNewOrModifiedFilter;
import net.sf.basedb.util.extensions.xml.XmlLoader;

/**
	Extension file processor implementation that will load the
	extension definitions from each file and register them with
	the registry that is managed by the manager. This processor
	is a two-step processer. The first step will load extension
	definitions from the files. The second step will register the
	loaded extension with the managers registry.
	<p>
	The first step is executed via the manager: 
	{@link ExtensionsManager#processFiles(ExtensionsFileProcessor)}.
	The second step is either automatically executed when all files
	have been processed or executed by by calling 
	{@link #finalizeRegistration(ExtensionsManager, boolean)}.
	<p>
	Note! It is recommended that this processor is paired with a 
	{@link ValidAndNewOrModifiedFilter} since it is usually no
	point in trying to register invalid extensions.


	@author Nicklas
	@since 3.0
	@base.modified $Date: 2011-08-11 12:45:58 +0200 (Thu, 11 Aug 2011) $
*/
public class RegisterExtensionsProcessor
	implements ExtensionsFileProcessor
{

	
	private static final org.apache.log4j.Logger log = 
		org.apache.log4j.LogManager.getLogger("net.sf.basedb.util.extensions.manager.RegisterExtensionsProcessor");
	
	private final XmlLoader loader;
	private final ProcessResults results;
	private boolean delayRegistration;
	private boolean forceUpdate;
	private List<FileData> allData;

	private int numFiles;
	private int numError;
	private int numRegistered;
	private int numUnregistered;

	/**
		Create a new processor. 
		
		@param loader The XML loader to use when parsing the metadata
	*/
	public RegisterExtensionsProcessor(XmlLoader loader, ProcessResults results)
	{
		this.loader = loader;
		this.results = results;
	}
	
	/*
		From the ExtensionsFileProcessor interface
		------------------------------------------
	*/
	@Override
	public void begin(ExtensionsManager manager, int numFiles)
	{
		this.allData = new ArrayList<FileData>(numFiles);
		this.numFiles = 0;
		this.numError = 0;
		this.numRegistered = 0;
		this.numUnregistered = 0;
	}

	@Override
	public void processFile(ExtensionsManager manager, WriteableExtensionsFile wFile)
	{
		InputStream in = null;
		ExtensionsFile xtFile = wFile.getExtensionsFile();
		
		try
		{
			log.info("Loading extensions from file: " + xtFile);
			FileData data = new FileData();
			data.writeableFile = wFile;
			allData.add(data);
			in = xtFile.getXmlStream();
			loader.loadXmlFile(in, xtFile.getName(), xtFile.getClassLoader(), true);
			data.extensionPoints = new ArrayList<ExtensionPoint<?>>(loader.getExtensionPoints());
			data.extensions = new ArrayList<Extension<?>>(loader.getExtensions());
			numFiles++;
			log.info("Loaded " + data.extensionPoints.size() + "/" + data.extensions.size() + 
					" extensions from file: " + xtFile);
		}
		catch (Exception ex)
		{
			wFile.setError(true);
			numError++;
			if (results != null)
			{
				results.addErrorMessage(xtFile, "Failed to load extension(s): " + ex.getMessage());
			}
			log.error("Failed to load extensions from file: " + xtFile, ex);
		}
		finally
		{
			FileUtil.close(in);
		}
	}
	
	@Override
	public void done(ExtensionsManager manager)
	{
		if (!delayRegistration) finalizeRegistration(manager, true);
	}
	
	@Override
	public void done(ExtensionsManager manager, Throwable t)
	{
		if (allData != null) allData.clear();
		allData = null;
	}
	// ------------------------------------------

	/**
		Set a flag to indicate if the actual registration of the found
		extensions and extension points should be delayed or not.
		The default is to not delay registration, which means that
		it happend when the {@link #done(ExtensionsManager)} method is
		called by the manager. If this flag is set the 
		{@link #finalizeRegistration(ExtensionsManager, boolean)}
		method must be explicitely called.
	*/
	public void setDelayRegistration(boolean delayRegistration)
	{
		this.delayRegistration = delayRegistration;
	}
	
	/**
		Is the actual registration delayed or not?
	*/
	public boolean isRegistrationDelayed()
	{
		return delayRegistration;
	}
	
	/**
		Set a flag to indicate if already registered extensions should
		be re-registered or not.
		@param forceUpdate TRUE to always re-register extensions, 
			FALSE to only register new or updated extensions
	*/
	public void setForceUpdate(boolean forceUpdate)
	{
		this.forceUpdate = forceUpdate;
	}
	
	/**
		Get the loader the processor is using for parsing xml files with
		extension definitions.
	*/
	public XmlLoader getXmlLoader()
	{
		return loader;
	}
	
	public ProcessResults getProcessResults()
	{
		return results;
	}
	
	public void finalizeRegistration(ExtensionsManager manager, boolean unregisterMissing)
	{
		if (allData == null) return;

		Registry registry = manager.getRegistry();

		// First loop -- unregister extensions and extension points that are missing
		if (unregisterMissing)
		{
			log.info("Unregistering all missing extension points and extensions");
			for (FileData data : allData)
			{
				WriteableExtensionsFile wFile = data.writeableFile;
				ExtensionsFile xtFile = wFile.getExtensionsFile();
	
				if (!forceUpdate && (!xtFile.wasModified() || xtFile.isNew()))
				{
					log.debug("Skipping new or unmodified file: " + xtFile);
					continue;
				}
	
				// Unregister all extensions if the file has an error
				if (xtFile.hasError())
				{
					data.extensions = Collections.emptyList();
					data.extensionPoints = Collections.emptyList();
				}
				
				try
				{
					wFile.open();
					int num = 0;
					// Extension points
					for (ExtensionPoint ep : xtFile.getObjectsOfClass(ExtensionPoint.class))
					{
						if (unregisterExtensionPoint(data, ep, registry))
						{
							num++;
						}
					}
					if (num > 0 && results != null)
					{
						numUnregistered += num;
						results.addMessage(xtFile, num + " extension point(s) unregistered.");
					}
					log.info("Unregistered " + num + " extension point(s) from file: " + xtFile);
					
					// Extensions
					num = 0;
					for (Extension ext : xtFile.getObjectsOfClass(Extension.class))
					{
						if (unregisterExtension(data, ext, registry)) 
						{
							num++;
						}
					}
					if (num > 0 && results != null)
					{
						numUnregistered += num;
						results.addMessage(xtFile, num + " extension(s) unregistered.");
					}
					log.info("Unregistered " + num + " extension(s) from file: " + xtFile);
				}
				catch (Throwable ex)
				{
					log.error("Could not unregister extensions from file: " + xtFile, ex);
				}
				finally
				{
					wFile.close(); // Important! release the write lock
				}
			}
		}
		
		// Second loop -- register extension points
		log.info("Registering all loaded extension points");
		for (FileData data : allData)
		{
			WriteableExtensionsFile wFile = data.writeableFile;
			ExtensionsFile xtFile = wFile.getExtensionsFile();

			// NOTE! Do not register files with an error
			if (xtFile.hasError())
			{
				log.debug("Skipping file with error: " + xtFile);
				continue; // with the next file
			}
			
			try
			{
				wFile.open();
				int num = registerExtensionPoints(data, registry, forceUpdate || xtFile.wasModified());
				if (num > 0 && results != null)
				{
					results.addMessage(xtFile, num + " extension point(s) registered.");
				}
				log.info("Registered " + num + " extension point(s) from file: " + xtFile);
				numRegistered += num;
			}
			catch (Throwable ex)
			{
				wFile.setError(true);
				if (results != null)
				{
					results.addErrorMessage(xtFile, 
						"Could not register extension point(s): " + ex.getMessage());
				}
				log.error("Could not register extension points from file: " + xtFile, ex);
			}
			finally
			{
				wFile.close(); // Important! release the write lock
			}
		}
		
		// Third loop -- register extensions
		log.info("Registering all loaded extensions");
		for (FileData data : allData)
		{
			WriteableExtensionsFile wFile = data.writeableFile;
			ExtensionsFile xtFile = wFile.getExtensionsFile();

			// NOTE! Do not register files with an error
			if (xtFile.hasError())
			{
				log.debug("Skipping file with error: " + xtFile);
				continue; // with the next file
			}
			
			try
			{
				wFile.open();
				int num = registerExtensions(data, registry, forceUpdate || xtFile.wasModified());
				if (num > 0 && results != null)
				{
					results.addMessage(xtFile, num + " extension(s) registered.");
				}
				log.info("Registered " + num + " extension(s) from file: " + xtFile);
				numRegistered += num;
			}
			catch (Throwable ex)
			{
				wFile.setError(true);
				if (results != null)
				{
					results.addErrorMessage(xtFile, 
						"Could not register extension(s): " + ex.getMessage());
				}
				log.error("Could not register extension points from file: " + xtFile, ex);
			}
			finally
			{
				wFile.close(); // Important! release the write lock
			}
		}
		
		allData.clear();
		allData = null;
	}

	/**
		Get the number of files that was successfully processed.
	*/
	public int getNumFiles()
	{
		return numFiles;
	}
	
	/**
		Get the number of files that was had an error.
	*/
	public int getNumError()
	{
		return numError;
	}
	
	/**
		Get the number of extensions + extension points that
		was registered.
	*/
	public int getNumRegistered()
	{
		return numRegistered;
	}
	
	/**
		Get the number of extensions + extension points that
		was unregistered.
	*/
	public int getNumUnregistered()
	{
		return numUnregistered;
	}

	
	private int registerExtensionPoints(FileData data, Registry registry, boolean forceUpdate)
	{
		int numRegistered = 0;
		for (ExtensionPoint<?> ep : data.extensionPoints)
		{
			if (forceUpdate || !registry.extensionPointIsRegistered(ep.getId()))
			{
				log.debug("Registering extension point: " + ep);
				ExtensionPointKey key = new ExtensionPointKey(ep);
				data.writeableFile.registerObject(key, ep);
				registerFactory(data, ep.getErrorHandlerFactory());
				registerFactory(data, ep.getRendererFactory());
				registry.registerExtensionPoint(ep);
				numRegistered++;
			}
		}
		return numRegistered;
	}
	
	/**
		Unregister the given extension point if it is not included in the new
		file data information. In all cases, factory metadata must be removed,
		since we leak memory otherwise (new factory instances are always created
		when scanning a modified file).
		@return TRUE if the extension point is not going to be registered again
	*/
	private boolean unregisterExtensionPoint(FileData data, ExtensionPoint ep, Registry registry)
	{
		log.debug("Unregistering extension point: " + ep);
		unregisterFactory(data, ep.getRendererFactory());
		unregisterFactory(data, ep.getErrorHandlerFactory());
		ExtensionPointKey key = new ExtensionPointKey(ep);
		data.writeableFile.unregisterObject(key);
		if (!data.extensionPoints.contains(ep))
		{
			registry.unregisterExtensionPoint(ep.getId());
			return true;
		}
		return false;
	}
	
	private int registerExtensions(FileData data, Registry registry, boolean forceUpdate)
	{
		int numRegistered = 0;
		for (Extension<?> ext : data.extensions)
		{
			if (forceUpdate || !registry.extensionIsRegistered(ext.getId()))
			{
				log.debug("Registering extension: " + ext);
				ExtensionKey key = new ExtensionKey(ext);
				data.writeableFile.registerObject(key, ext);
				registerFactory(data, ext.getActionFactory());
				registerFactory(data, ext.getRendererFactory());
				registry.registerExtension(ext);
				numRegistered++;
			}
		}
		return numRegistered;
	}

	/**
		Unregister the given extension if it is not included in the new
		file data information. In all cases, factory metadata must be removed,
		since we leak memory otherwise (new factory instances are always created
		when scanning a modified file).
		@return TRUE if the extension point is not going to be registered again
	*/
	private boolean unregisterExtension(FileData data, Extension ext, Registry registry)
	{
		log.debug("Unregistering extension: " + ext);
		unregisterFactory(data, ext.getRendererFactory());
		unregisterFactory(data, ext.getActionFactory());
		ExtensionKey key = new ExtensionKey(ext);
		data.writeableFile.unregisterObject(key);
		if (!data.extensions.contains(ext))
		{
			registry.unregisterExtension(ext.getId());
			return true;
		}
		return false;
	}
	
	/**
		Register factory parameters for the given factory instance. 
		The parameters are retreived from the {@link XmlLoader#getFactoryParameters(Object)}
		method.
		
		@param data The extension file the factory was defined in
		@param factory The factory instance (if null, no registration is done)
	*/
	private void registerFactory(FileData data, Object factory)
	{
		if (factory == null) return;
		data.writeableFile.registerMetadata(new FactoryParametersKey(factory), loader.getFactoryParameters(factory));
	}

	private void unregisterFactory(FileData data, Object factory)
	{
		if (factory == null) return;
		data.writeableFile.unregisterMetadata(new FactoryParametersKey(factory));
	}
	
	/**
		Keep track of the extension points and extensions found in a file.
	*/
	static class FileData
	{
		WriteableExtensionsFile writeableFile;
		List<Extension<?>> extensions;
		List<ExtensionPoint<?>> extensionPoints;
	}

}
