/**
	$Id: ExperimentLoader.java 5966 2012-02-16 08:27:56Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.loader;

import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Experiment;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ItemResultIterator;
import net.sf.basedb.core.RawBioAssay;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.node.ChildNodeDirection;
import net.sf.basedb.util.overview.node.NameableNameGenerator;
import net.sf.basedb.util.overview.Node;
import net.sf.basedb.util.overview.node.NodeFactory;

/**
	Node loader implementation for experiments. An experiment is the
	endpoint of the forward-direction. The reverse direction goes
	to the {@link RawBioAssay}:s included in the experiment.
	
	@author Nicklas
	@version 2.10
	@base.modified $Date: 2012-02-16 09:27:56 +0100 (Thu, 16 Feb 2012) $
*/
public class ExperimentLoader
	extends BasicItemNodeLoader<Experiment>
{
	public ExperimentLoader()
	{
		super(Item.EXPERIMENT, ALLOW_ROOT_NODE, 
				new NameableNameGenerator<Experiment>("experiment", "Experiment"));
	}

	/*
		From the NodeLoader interface
		------------------------------
	*/
	/**
		Create property-loading experiment nodes from a given rawbioassay node.
		The returned node is a folder-type node containing item nodes for 
		each experiment were the raw bioassay is used.
		@return The folder node for the experiments
	*/
	@Override
	public Node createForwardNode(DbControl dc, OverviewContext context, Node rawBioAssayNode)
	{
		NodeFactory<Experiment> nf = getNodeFactory(dc, context);
		Node folderNode = null;
		RawBioAssay rba = (RawBioAssay)rawBioAssayNode.getItem(dc);
		ItemQuery<Experiment> query = context.initQuery(rba.getExperiments(), "name");
		ItemResultIterator<Experiment> it = query.iterate(dc);
		while (it.hasNext())
		{
			Experiment experiment = it.next();
			if (folderNode == null)
			{
				folderNode = new Node("experiments", "Experiments", rawBioAssayNode, ChildNodeDirection.PROPERTY);
			}
			createItemNode(nf, experiment, experiment, false, folderNode, ChildNodeDirection.FORWARD);
		}
		postValidateFolder(nf, folderNode, rawBioAssayNode, false);
		return folderNode;
	}
	// ------------------------------
	
	/*
		From the AbstractNodeLoader class
		----------------------------------
	*/
	/**
		Load property nodes for an experiment:
		<ul>
		<li>Experimental factors: {@link ExperimentalFactorLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		<li>Annotations: {@link AnnotationLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		</ul>
	*/
	@Override
	protected void loadPropertyChildNodes(DbControl dc, OverviewContext context, Node experimentNode)
	{
		getNodeLoader(context, "EXPERIMENTAL.FACTOR").createPropertyNode(dc, context, experimentNode);
		getNodeLoader(context, Item.ANNOTATION).createPropertyNode(dc, context, experimentNode);
		getNodeLoader(context, Item.ANYTOANY).createPropertyNode(dc, context, experimentNode);
	}
	
	/**
		Loads the raw bioassay nodes that are part of this experiment
		@see RawBioAssayLoader#createReverseNode(DbControl, OverviewContext, Node)
	*/
	@Override
	protected void loadReverseChildNodes(DbControl dc, OverviewContext context, Node experimentNode)
	{
		getNodeLoader(context, Item.RAWBIOASSAY).createReverseNode(dc, context, experimentNode);
	}
	// --------------------------------
}
