/**
	$Id: BasicNodeValidator.java 5818 2011-10-21 11:15:41Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.validator;

import net.sf.basedb.core.DbControl;
import net.sf.basedb.util.overview.Fix;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.Validator;
import net.sf.basedb.util.overview.Node;

/**
	A simple node validation implementation that takes care of missing
	and denied access to items. This implementation will create
	child nodes for items that we are denied access to but not to
	missing items. Missing items are reported as failures to the
	parent node. Denied access are reported as failures to the node
	representing the item.
	
	@author Nicklas
	@version 2.10
	@base.modified $Date: 2011-10-21 13:15:41 +0200 (Fri, 21 Oct 2011) $
*/
public abstract class BasicNodeValidator<I>
	implements NodeValidator<I>
{

	private Validator ruleMissing;
	private Validator ruleDenied;
	
	/**
		Creates a new validator.
		@param ruleMissing The validation rule that is broken if an item is missing,
			or null to not report missing items as a failure
		@param ruleDenied The validation rule that is broken if we are denied access
			to an item, or null to not report this as a failure
	 */
	public BasicNodeValidator(Validator ruleMissing, Validator ruleDenied)
	{
		this.ruleDenied = ruleDenied;
		this.ruleMissing = ruleMissing;
	}
	
	/*
		From the NodeValidator interface
		--------------------------------
	*/
	/**
		Submits the {@link #ruleMissing} validator to the context as a failure
		unless it is null.
		@return Always false
	*/
	@Override
	public boolean preMissingItem(DbControl dc, OverviewContext context, Node parentNode)
	{
		if (ruleMissing != null)
		{
			context.createFailure(ruleMissing, parentNode, null, getMissingItemFix(dc, parentNode));
		}
		return false;
	}
	
	/**
		@return Always true
	*/
	@Override
	public boolean preDeniedItem(DbControl dc, OverviewContext context, Node parentNode)
	{
		return true;
	}
	
	/**
		Subclasses should override this method if the need more validation for an
		existing and accessible item.
		@return Always true
	*/
	@Override
	public boolean preValidate(DbControl dc, OverviewContext context, I item, Node parentNode)
	{
		return true;
	}
	
	/**
		Do nothing.
	*/
	@Override
	public void postMissingItem(DbControl dc, OverviewContext context, Node node, Node parentNode)
	{}
	
	/**
		Submits the {@link #ruleDenied} validator to the context as a failure
		unless it is null.
	*/
	@Override
	public void postDeniedItem(DbControl dc, OverviewContext context, Node node, Node parentNode)
	{
		if (ruleDenied != null)
		{
			context.createFailure(ruleDenied, node, null, getDeniedItemFix(dc, node, parentNode));
		}
	}

	/**
		Do nothing. Subclasses should override this method if they need more validation for an
		existing and accessible item.
	*/
	@Override
	public void postValidate(DbControl dc, OverviewContext context, Node node, Node parentNode)
	{}
	
	/**
		Do nothing. Subclasses should override this method if they need to validate all items
		inside a folder-type node.
	*/
	@Override
	public void postValidateFolder(DbControl dc, OverviewContext context, Node folderNode, Node parentNode)
	{}
	// -----------------------------------

	/**
		Subclasses should override this method if they have a suggessted fix
		for a missing item problem. The default implementation does nothing,
		but the following template may be usable by most subclasses:
		<pre class="code">
		new Fix("Add item to parent", parentNode.getItem())
		</pre>
	*/
	protected abstract Fix getMissingItemFix(DbControl dc, Node parentNode);
	
	/**
		Subclasses should override this method if they have a suggessted fix
		for an access denied problem. The default implementation does nothing.
	*/
	protected Fix getDeniedItemFix(DbControl dc, Node node, Node parentNode)
	{
		return null;
	}
}
