/**
	$Id: SampleValidator.java 5807 2011-10-14 13:08:23Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.validator;

import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.BioSource;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Extract;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemSubtype;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.Sample;
import net.sf.basedb.util.overview.Fix;
import net.sf.basedb.util.overview.Node.Type;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.Validator;
import net.sf.basedb.util.overview.Node;
import net.sf.basedb.util.overview.node.ChildNodeDirection;

/**
	Validator implementation for sample. Validation rules:
	<ul>
	<li>Missing item: {@link Validator#MISSING_SAMPLE}
	<li>Access denied: {@link Validator#DENIED_SAMPLE}
	<li>Subtype of parent sample: {@link Validator#INCORRECT_SAMPLE_TYPE}
	<li>Subtype of parent biosource: {@link Validator#INCORRECT_BIOSOURCE_TYPE}
	</ul>

	@author Nicklas
	@version 2.10
	@base.modified $Date: 2011-10-14 15:08:23 +0200 (Fri, 14 Oct 2011) $
*/
public class SampleValidator
	extends NameableNodeValidator<Sample>
{
	
	public SampleValidator()
	{
		super(Validator.MISSING_SAMPLE, Validator.DENIED_SAMPLE);
	}
	
	/* 
		From BasicValidator class
		-------------------------
	*/
	@Override
	public void postValidate(DbControl dc, OverviewContext context, Node node, Node parentNode) 
	{
		super.postValidate(dc, context, node, parentNode);

		if (parentNode != null)
		{
			Node parentItemNode = parentNode.getNodeType() == Type.ITEM ? parentNode : parentNode.getParent();
			Item parentType = parentItemNode.getItemType();

			if (parentType == Item.BIOSOURCE)
			{
				checkParentBioSourceSubtype(dc, context, node, parentItemNode);
			}
			else if (parentType == Item.SAMPLE)
			{
				if (node.getChildNodeDirection() == ChildNodeDirection.FORWARD)
				{
					checkParentSampleSubtype(dc, context, node, parentItemNode);
				}
				else
				{
					checkParentSampleSubtype(dc, context, parentItemNode, node);
				}
			}
			else if (parentType == Item.EXTRACT)
			{
				ExtractValidator.checkParentSampleSubtype(dc, context, parentItemNode, node);
			}
		}
	}

	
	/**
		The suggested fix is to add a parent sample to the extract
		or one or more samples to the pooled sample.
	*/
	@Override
	protected Fix getMissingItemFix(DbControl dc, Node parentNode)
	{
		Fix fix = null;
		Item parentType = parentNode.getItemType();
		BasicItem parentItem = parentNode.getItem();
		if (parentType == Item.EXTRACT)
		{
			fix =  new Fix("Add parent sample to '" + ((Extract)parentItem).getName() + "'", parentItem);
		}
		else if (parentType == Item.SAMPLE)
		{
			fix = new Fix("Add parents to '" + ((Sample)parentItem).getName() + "'", parentItem);
		}
		return fix;
	}
	// ----------------------------

	/**
		Check the subtype of the biosurce that is the parent to a sample. Given that
		the sample has a non-null subtype, which has a related BIOSOURCE subtype, this
		method checks if the parent biosource has the same subtype. If there is no related
		subtype for the sample, any subtype for the parent is accepted.
		@since 3.0
	*/
	public static void checkParentBioSourceSubtype(DbControl dc, OverviewContext context, Node sampleNode, Node bioSourceNode)
	{
		Sample sample = (Sample)sampleNode.getItem(dc);
		BioSource bioSource = (BioSource)bioSourceNode.getItem(dc);
		
		try
		{
			ItemSubtype expectedSubtype = ItemSubtype.getRelatedSubtype(dc, sample, Item.BIOSOURCE, Item.SAMPLE);
			ItemSubtype subtype = bioSource.getItemSubtype();
			
			if (expectedSubtype != null && !expectedSubtype.equals(subtype))
			{
				ItemSubtype mySubtype = sample.getItemSubtype();
				Fix subtypeFix = subtype == null ? null : 
					new Fix("Change related biosource subtype of '" + mySubtype.getName() + 
						"' to '" + subtype.getName() + "'", mySubtype);

				if (expectedSubtype.getMainItemType() == Item.BIOSOURCE)
				{
					// Expected another biosource subtype
					context.createFailure(Validator.INCORRECT_BIOSOURCE_TYPE, sampleNode, 
						"Expected a biosource with subtype: " + expectedSubtype.getName(),
						new Fix("Change subtype of '" + bioSource.getName() + "' to '" + expectedSubtype.getName() + "'", bioSource),
						new Fix("Change subtype or parent biosource of '" + sampleNode.getTitle() + "'", sample),
						subtypeFix
					);
				}
				else
				{
					// Expected a sample as parent
					context.createFailure(Validator.INCORRECT_BIOSOURCE_TYPE, sampleNode, 
							"Expected a sample with subtype: " + expectedSubtype.getName(),
							new Fix("Change subtype or parent biosource of '" + sampleNode.getTitle() + "'", sample),
							subtypeFix
						);
				}
			}
		}
		catch (PermissionDeniedException ex)
		{}
	}
	
	/**
		Check the subtype of the sample that is the parent to another sample. Given that
		the child sample has a non-null subtype, which has a related SAMPLE subtype, this
		method checks if the parent sample has the same subtype. If there is no related
		subtype for the child sample, any subtype for the parent is accepted.
		@since 3.0
	*/
	public static void checkParentSampleSubtype(DbControl dc, OverviewContext context, Node childNode, Node parentNode)
	{
		Sample child = (Sample)childNode.getItem(dc);
		Sample parent = (Sample)parentNode.getItem(dc);
		
		try
		{
			ItemSubtype expectedSubtype = ItemSubtype.getRelatedSubtype(dc, child, Item.SAMPLE, Item.BIOSOURCE);
			ItemSubtype subtype = parent.getItemSubtype();
			
			if (expectedSubtype != null && !expectedSubtype.equals(subtype))
			{
				ItemSubtype mySubtype = child.getItemSubtype();
				Fix subtypeFix = subtype == null || subtype.equals(mySubtype) ? null : 
					new Fix("Change related sample subtype of '" + mySubtype.getName() + 
						"' to '" + subtype.getName() + "'", mySubtype);

				if (expectedSubtype.getMainItemType() == Item.SAMPLE)
				{
					context.createFailure(Validator.INCORRECT_SAMPLE_TYPE, childNode, 
						"Expected a sample with subtype: " + expectedSubtype.getName(),
						new Fix("Change subtype of '" + parent.getName() + "' to '" + expectedSubtype.getName() + "'", parent),
						new Fix("Change subtype or parent sample of '" + childNode.getTitle() + "'", child),
						subtypeFix
					);
				}
				else
				{
					context.createFailure(Validator.INCORRECT_SAMPLE_TYPE, childNode, 
						"Expected a biosource with subtype: " + expectedSubtype.getName(),
						new Fix("Change subtype or parent of '" + childNode.getTitle() + "'", child),
						subtypeFix
					);
				}
			}
		}
		catch (PermissionDeniedException ex)
		{}
	}


}
