/**
	$Id: QueryOptions.java 4512 2008-09-11 20:05:50Z jari $

	Copyright (C) 2007 Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

package net.sf.basedb.info;

/**
	This class holds information to configure 
	itemqueries in webservices with. 	
	It is which items to include in the query
	and which restriction to put on the query.
	<p>
		The includes in this class correspond 
		directly to the includes in 
		{@link net.sf.basedb.core.Include}.
		The include options will be set to TRUE if nothing
		else is specified. 
	</p>
	
  	@author Martin
  	@version 2.5
 */
public class QueryOptions
	implements Transferable
{
	/**
	 	Include items that belongs to the logged in user.
	 	Included by default.
	 */
	private boolean includeMine = true;
	
	/**
	 	Include items, shared to the logged in user.
	 	Included by default.
	 */
	private boolean includeShared = true;
	
	/**
	 	Include items from the active project.
	 	Included by default.
	 */
	private boolean includeInProject = true;
	
	/**
		Include items that belong to other users.
		Included by default.
	 */
	private boolean includeOthers = true;
	
	
	private boolean includeNotRemoved = true;
	
	private QueryParameter[] parameters;
	
	/**
	 	A restriction to set on
	 	the query.
	*/
	private String restriction = null;
	
	/**
	 	The prefix that represent a root alias in 
	 	a query. This string can be used in {{@link #restriction} 
	 	and will be replaced when the restriction is 
	 	inserted into a HQL query.
	 */
	private String prefix = null;
		
	/**
		An empty constructor.
	*/
	public QueryOptions()
	{}
	
	/**
	 	Constructor where the parameters is set from start.
	  	@param includeMine TRUE if logged in user's items should be included.
	  	@param includeShared TRUE if items shared to logged in user should be included.
	  	@param includeInProject TRUE if items in the active project should be included.
	  	@param includeOthers TRUE if other users' items should be included.
	  	@param restriction A {@link net.sf.basedb.core.query.Restriction} as a string. 
	  	@param prefix Prefix used in the restriction.
	 */
	public QueryOptions
		(
			Boolean includeMine,
			Boolean includeShared,
			Boolean includeInProject,
			Boolean includeOthers,
			String restriction,
			String prefix
		)
	{
		if (includeMine != null) this.includeMine = includeMine;
		if (includeShared != null) this.includeShared = includeShared;
		if (includeInProject != null) this.includeInProject = includeInProject;
		if (includeOthers != null) this.includeOthers = includeOthers;
		this.restriction = restriction;
		this.prefix = prefix;
	}
	
	/**
	 	Sets if the logged in user's items should be included or not.
	  	@param includeMine TRUE to include logged in user's items, otherwise FALSE
	 */
	public void setIncludeMine(boolean includeMine)
	{
		this.includeMine = includeMine;
	}
	
	/**
	 	Gets if items of logged in user should be included 
	  	@return TRUE to include, otherwise FALSE
	 */
	public boolean getIncludeMine()
	{
		return includeMine;
	}
	
	/**
	 	Sets if items shared to logged in user should be included.
	  	@param includeShared TRUE to include shared items, otherwise FALSE.
	 */
	public void setIncludeShared(boolean includeShared)
	{
		this.includeShared = includeShared;
	}
	
	/**
	 	Gets if items shared to the logged in user should 
	 	be included or not
	  	@return TRUE if shared items should be included, FALSE otherwise.
	 */
	public boolean getIncludeShared()
	{
		return includeShared;
	}
	
	/**
	 	Sets if items in currently active project should be included.	 	
	  	@param includeInProject TRUE to include, FALSE otherwise.
	 */
	public void setIncludeInProject(boolean includeInProject)
	{
		this.includeInProject = includeInProject;
	}
	
	/**
	 	Gets if items in the project that is currently active should be included.
	  	@return TRUE to include the items, FALSE otherwise.
	 */
	public boolean getIncludeInProject()
	{
		return includeInProject;
	}
	
	/**
	 	Sets if other users' items should be included.
	  	@param includeOthers TRUE if the item should be included, FALSE otherwise.
	 */
	public void setIncludeOthers(boolean includeOthers)
	{
		this.includeOthers = includeOthers;
	}
	
	/**
	 	Gets if other users' item should be included in the query
	  	@return TRUE if the items should be included, FALSE otherwise.
	 */
	public boolean getIncludeOthers()
	{
		return includeOthers;
	}
	
	/**
	 	Sets if not removed items should be included
	  	@param include TRUE(default) to include. 
	 */
	public void setIncludeNotRemoved(boolean include)
	{
		this.includeNotRemoved = include;
	}
	
	/**
	 	Include not removed items
	  	@return TRUE to include, FALSE otherwise.
	 */
	public boolean getIncludeNotRemoved()
	{
		return this.includeNotRemoved;
	}
	/**
	 	Sets restriction for these query options.
	 	<p>
	 		The string must be in right format so it can be inserted 
	 		directly into a hql query. The restriction string
	 		can contain a prefix, but it has to be set separatly 
	 		{@link #setRestrictionPrefix(String)}. 
	 	</p>
	 	<p>
	 		Example how this property can be used together with the prefix:<br>
	 		<i>$name='item1'</i> <br>
	 		When this string is used in a query, ie that select users, it will look like this:
	 		<i>usr.name='item1'</i>.<br>
	 	</p>
	 	
	 	@param restriction A restriction as a string to use with HQL queries. 
	 */
	public void setRestriction(String restriction)
	{
		this.restriction = restriction;
	}
	
	/**
	 	Gets the restriction for this QueryOptions
	 	@return A String with restriction
	 */
	public String getRestriction()
	{
		return restriction;
	}
	
	/**
	 	Sets the prefix to use in restrictions or expression
	  	@param prefix A prefix where the root-alias should be put.
	 */
	public void setRestrictionPrefix(String prefix)
	{
		this.prefix = prefix;
	}
	
	/**
	 	Gets the prefix used with the restriction/expression
	 	in this object.
	  	@return A String that should be replaced with the 
	  		root alias when put in a query.
	 */
	public String getRestrictionPrefix()
	{
		return prefix;
	}
	
	/**
	 	Sets both the restriction and the prefix with it.
	  	@param restriction A string with a valid hql restriction.
	  	@param prefix A string that is used where root-alias should be put in
	  		the restriction/expression.
	 */
	public void setRestrictionWithPrefix(String restriction, String prefix)
	{
		this.restriction = restriction;
		this.prefix = prefix;
	}
	
	/**
		Get the parameters used in this query.
		@return An array of query parameters, may be null or empty if no
			parameters are used
	*/
	public QueryParameter[] getParameters()
	{
		return parameters;
	}
	
	/**
		Set the parmeters to use in this query.
		@param parameters An array of query parameters, may be null or empty
			if not parameters are used
	*/
	public void setParameters(QueryParameter[] parameters)
	{
		this.parameters = parameters;
	}
	
}