/**
	$Id: BfsExporterPlugin.java 5610 2011-04-15 13:01:15Z nicklas $

	Copyright (C) 2010 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

package net.sf.basedb.plugins;

import net.sf.basedb.core.AnnotationType;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.BooleanParameterType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Directory;
import net.sf.basedb.core.ExtendedProperties;
import net.sf.basedb.core.ExtendedProperty;
import net.sf.basedb.core.ExtraValue;
import net.sf.basedb.core.ExtraValueType;
import net.sf.basedb.core.File;
import net.sf.basedb.core.Formula;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemParameterType;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.Path;
import net.sf.basedb.core.PathParameterType;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.RawDataProperty;
import net.sf.basedb.core.RequestInformation;
import net.sf.basedb.core.StringParameterType;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.ExtraValue.CoordinateType;
import net.sf.basedb.core.plugin.AbstractPlugin;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.plugin.InteractivePlugin;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.plugin.Response;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Orders;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.signal.SignalHandler;
import net.sf.basedb.core.signal.SignalTarget;
import net.sf.basedb.core.signal.ThreadSignalHandler;
import net.sf.basedb.util.Enumeration;
import net.sf.basedb.util.FileUtil;
import net.sf.basedb.util.bfs.AnnotationWriter;
import net.sf.basedb.util.bfs.DataWriterFactory;
import net.sf.basedb.util.bfs.FilenameGenerator;
import net.sf.basedb.util.bfs.GenericBaseDataWriterFactory;
import net.sf.basedb.util.bfs.GenericDiskDataWriterFactory;
import net.sf.basedb.util.bfs.MetadataWriter;
import net.sf.basedb.util.bfs.SequenceFilenameGenerator;
import net.sf.basedb.util.bfs.SuggestedFilenameGenerator;
import net.sf.basedb.util.export.spotdata.AssayField;
import net.sf.basedb.util.export.spotdata.BfsExporter;
import net.sf.basedb.util.export.spotdata.DynamicField;
import net.sf.basedb.util.export.spotdata.ExportableField;
import net.sf.basedb.util.export.spotdata.ExportableFieldFactory;
import net.sf.basedb.util.export.spotdata.MatrixBfsExporter;
import net.sf.basedb.util.export.spotdata.SerialBfsExporter;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

/**
	 Exports bioassay set data to BFS format.

	 @author Nicklas
	 @since 2.15
	 @base.modified $Date: 2011-04-15 15:01:15 +0200 (Fri, 15 Apr 2011) $
*/
public class BfsExporterPlugin
	extends AbstractPlugin
	implements InteractivePlugin, SignalTarget
{

	private static final Set<GuiContext> guiContexts =
		Collections.singleton(GuiContext.item(Item.BIOASSAYSET));

	private ThreadSignalHandler signalHandler;
	private RequestInformation configureJob;

	public BfsExporterPlugin()
	{}
	
	/*
		From the Plugin interface
		-------------------------------------------
	*/
	/**
		@return MainType.EXPORT
	*/
	@Override
	public MainType getMainType()
	{
		return MainType.EXPORT;
	}
	@Override
	public boolean supportsConfigurations()
	{
		return false;
	}
	@Override
	public boolean requiresConfiguration()
	{
		return false;
	}
	// --------------------------------------------
	/*
		From the InteractivePlugin interface
		-------------------------------------------
	*/
	@Override
	public Set<GuiContext> getGuiContexts()
	{
		return guiContexts;
	}
	@Override
	public String isInContext(GuiContext context, Object item)
	{
		String message = null;	
		if (item == null)
		{
			message = "The object is null";
		}
		else if (!(item instanceof BioAssaySet))
		{
			message = "The object is not a BioAssaySet: " + item;
		}
		else
		{
			BioAssaySet bas = (BioAssaySet)item;
			if (bas.getNumSpots() <= 0)
			{
				message = "This plug-in requires spot data in the database";
			}
		}
		return message;
	}

	@Override
	public RequestInformation getRequestInformation(GuiContext context, String command)
		throws BaseException
	{
		RequestInformation requestInformation = null;
		if (command.equals(Request.COMMAND_CONFIGURE_JOB))
		{
			requestInformation = getConfigureJobParameters(context);
		}
		return requestInformation;
	}
	
	@Override
	public void configure(GuiContext context, Request request, Response response)
	{
		try
		{
			String command = request.getCommand();
			if (command.equals(Request.COMMAND_CONFIGURE_JOB))
			{
				RequestInformation ri = getConfigureJobParameters (context);
				List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size()+" invalid parameter(s) were found " +
														"in the request", errors);
					return;
				}
				
				// Source and file options
				storeValue(job, request, ri.getParameter("source"));
				storeValue(job, request, ri.getParameter("saveTo"));
				storeValue(job, request, ri.getParameter("toZipFile"));
				
				// Export options
				storeValue(job, request, ri.getParameter(PARAMETER_FILEFORMAT));
				storeValue(job, request, ri.getParameter(PARAMETER_FILENAMEGENERATOR));
				storeValues(job, request, ri.getParameter("spotFields"));
				storeValues(job, request, ri.getParameter("reporterFields"));
				storeValues(job, request, ri.getParameter("assayFields"));

				BioAssaySet bioAssaySet = (BioAssaySet)job.getValue("source");
				response.setSuggestedJobName("Export bioassay set '" +bioAssaySet.getName() + "' to " + 
						job.getValue(PARAMETER_FILEFORMAT) + " BFS file: " + job.getValue("saveTo"));
				
				response.setDone("Job configuration complete", Job.ExecutionTime.SHORT);
			}
		}
		catch (Throwable ex)
		{
			response.setError(ex.getMessage(), Arrays.asList(ex));
		}
	}
	public void run(Request request, Response response, ProgressReporter progress)
	{
		String command = request.getCommand();
		if (command.equals(Request.COMMAND_EXECUTE))
		{
			DbControl dc = null;
			try
			{
				dc = sc.newDbControl();
				begin(dc);
				performExport(progress);
				dc.commit();
				end(true);
				response.setDone(getSuccessMessage());
			}
			catch (Throwable t)
			{
				end(false);
				response.setError(t.getMessage(), Arrays.asList(t));
			}
			finally
			{
				if (dc != null) dc.close();
			}
		}
		else
		{
			response.setError("Unknown command: "+command, null);
		}		
	}
	// --------------------------------------------
	/*
		From the SignalTarget interface
		-------------------------------------------
	*/
	@Override
	public SignalHandler getSignalHandler()
	{
		signalHandler = new ThreadSignalHandler();
		return signalHandler;
	}
	// -------------------------------------------

	/*
		From AbstractExporterPlugin
		------------------------------------------.
	*/
	private DbControl dc;
	private BioAssaySet bas;
	private Map<String, ExtendedExportableField> exportableFields;


	protected void begin(DbControl dc)
	{
		this.dc = dc;
		bas = (BioAssaySet)job.getValue("source");
	}	

	@SuppressWarnings("unchecked")
	protected void performExport(ProgressReporter progress)
		throws IOException
	{
		if (signalHandler != null) signalHandler.setWorkerThread(null);

		// Get job configuration parameters
		BioAssaySet source = (BioAssaySet)job.getValue("source");
		source = BioAssaySet.getById(dc, source.getId());
		String path = (String)job.getValue("saveTo");		
		boolean toZipFile = Boolean.TRUE.equals(job.getValue("toZipFile"));
		
		Directory saveTo = null;
		File zipFile = null;
		java.io.File tmpDir = null;
		try
		{
			if (toZipFile)
			{
				if (!path.endsWith(".zip")) path += ".zip";
				zipFile = File.getByPath(dc, new Path(path, Path.Type.FILE), true);
				if (!zipFile.isInDatabase()) dc.saveItem(zipFile);
				zipFile.setMimeType("application/zip");
				tmpDir = FileUtil.createTempDirectory("bfs-export", null, null);
			}
			else
			{
				saveTo = Directory.getNew(dc, new Path(path, Path.Type.DIRECTORY));
				if (!saveTo.isInDatabase()) dc.saveItem(saveTo);
			}
	
			String format = (String)job.getValue(PARAMETER_FILEFORMAT);
			String filenameGen = (String)job.getValue(PARAMETER_FILENAMEGENERATOR);
			List<String> reporterFields = (List<String>)job.getValues("reporterFields");
			List<String> spotFields = (List<String>)job.getValues("spotFields");
			List<String> assayFields = (List<String>)job.getValues("assayFields");
			if (reporterFields == null) reporterFields = Collections.emptyList();
			if (spotFields == null) spotFields = Collections.emptyList();
			if (assayFields == null) assayFields = Collections.emptyList();
			Map<String, ExtendedExportableField> all = getExportableFields(dc, source);
		
			// Prepare for export
			if (progress != null) progress.display(0, "Exporting ...");
			BfsExporter exporter = null;
			if (SerialBfsExporter.SUBTYPE.equals(format))
			{
				exporter = new SerialBfsExporter();
			}
			else if (MatrixBfsExporter.SUBTYPE.equals(format))
			{
				exporter = new MatrixBfsExporter();
			}
			else
			{
				throw new BaseException("Unknown format. "+format);
			}
			exporter.setDbControl(dc);
			exporter.setProgressReporter(progress);
			exporter.setSource(source);
			
			// Create files & file factory
			if (toZipFile)
			{
				java.io.File metadata = createFile(tmpDir, "metadata.txt");
				exporter.setMetadataWriter(MetadataWriter.create(metadata));
			}
			else
			{
				File metadata = createFile(dc, saveTo, "metadata.txt");
				exporter.setMetadataWriter(MetadataWriter.create(metadata));
			}
			
			if (reporterFields != null && reporterFields.size() > 0)
			{
				AnnotationWriter reporterWriter = null;
				if (toZipFile)
				{
					java.io.File reporter = createFile(tmpDir, "rdata.txt");
					reporterWriter = AnnotationWriter.create(reporter);
				}
				else
				{
					File reporter = createFile(dc, saveTo, "rdata.txt");
					reporterWriter = AnnotationWriter.create(reporter);
				}
				exporter.setReporterAnnotationsWriter(reporterWriter);
				for (String field : reporterFields)
				{
					ExtendedExportableField ef = all.get(field);
					exporter.addReporterField((DynamicField)ef.getField());
				}
			}
	
			if (assayFields != null && assayFields.size() > 0)
			{
				AnnotationWriter assayWriter = null;
				if (toZipFile)
				{
					java.io.File assay = createFile(tmpDir, "pdata.txt");
					assayWriter = AnnotationWriter.create(assay);
				}
				else
				{
					File assay = createFile(dc, saveTo, "pdata.txt");
					assayWriter = AnnotationWriter.create(assay);
				}
				exporter.setAssayAnnotationsWriter(assayWriter);			
				for (String field : assayFields)
				{
					ExtendedExportableField ef = all.get(field);
					exporter.addAssayField((AssayField)ef.getField());
				}
			}
			
			if (spotFields != null && spotFields.size() > 0)
			{
				FilenameGenerator filenameGenerator = "sequence".equals(filenameGen) ?
					new SequenceFilenameGenerator("sdata-", ".txt", 1) : new SuggestedFilenameGenerator();
				DataWriterFactory dataWriterFactory = toZipFile ?
					new GenericDiskDataWriterFactory(tmpDir, filenameGenerator) :
					new GenericBaseDataWriterFactory(dc, saveTo, filenameGenerator);
				exporter.setDataWriterFactory(dataWriterFactory);
				for (String field : spotFields)
				{
					ExtendedExportableField ef = all.get(field);
					exporter.addSpotField((DynamicField)ef.getField());
				}
			}
			exporter.doExport();
			
			if (toZipFile)
			{
				ZipOutputStream zipStream = new ZipOutputStream(zipFile.getUploadStream(false));
				for (java.io.File file : tmpDir.listFiles())
				{
					ZipEntry entry = new ZipEntry(file.getName());
					entry.setTime(file.lastModified());
					entry.setSize(file.length());
					zipStream.putNextEntry(entry);
					FileUtil.copy(FileUtil.getInputStream(file), zipStream);
					zipStream.flush();
					zipStream.closeEntry();
					file.delete();
				}
				zipStream.close();
				tmpDir.delete();
			}
		}
		finally
		{
			if (tmpDir != null) FileUtil.deleteTempDirectory(tmpDir);
		}
	}
	
	protected void end(boolean success)
	{
		this.dc = null;
	}
	
	protected String getSuccessMessage()
	{
		return "Bioassay set '" + bas.getName() + "' exported successfully";
	}
	// -------------------------------------------

	private RequestInformation getConfigureJobParameters(GuiContext context)
	{
		if (configureJob != null) return configureJob;

		DbControl dc = null;
		try
		{
			dc = sc.newDbControl();
			int currentBioAssaySetId = sc.getCurrentContext(Item.BIOASSAYSET).getId();			
			BioAssaySet source = BioAssaySet.getById(dc, currentBioAssaySetId);
			
			// Generate a default path 
			String defaultPath = "~/" + Path.makeSafeFilename(
					source.getName() + "-" + source.getId(), "").toLowerCase();
			
			
			List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();
			
			// Source bioassay set parameter
			parameters.add(new PluginParameter<BioAssaySet>(
					"source",
					"Bioassay set",
					"The bioassay set to export",
					new ItemParameterType<BioAssaySet>(BioAssaySet.class, source, true, 1, null) 
				));

			// File format parameter
			parameters.add(getFileFormatParameter());
			// File name generation strategy parameter
			parameters.add(getFileNameGeneratorParameter());

			// Save to directory options
			PluginParameter<String> saveToParameter = new PluginParameter<String>
			(
				"saveTo", "Save to", "The path to the directory/zip file where the files should be saved", 
				defaultPath, new PathParameterType(Path.Type.DIRECTORY, null, true));
			parameters.add(saveToParameter);
			parameters.add(new PluginParameter<Boolean>(
					"toZipFile", "Zip file", "Select this option to save the data " +
					"in a zip file. If not selected, the data is saved in a directory.",
					new BooleanParameterType(null, false)
			));
			
			// Spot and reporter fields
			Enumeration<String, String> spotFields = new Enumeration<String, String>();
			Enumeration<String, String> reporterFields = new Enumeration<String, String>();
			Enumeration<String, String> assayFields = new Enumeration<String, String>();
			
			for (ExtendedExportableField f : getExportableFields(dc, source).values())
			{
				if (f.getField() instanceof AssayField)
				{
					assayFields.add(f.getId(), f.getGuiTitle());
				}
				else if (f.isReporterField())
				{
					reporterFields.add(f.getId(), f.getGuiTitle());
				}
				else
				{
					spotFields.add(f.getId(), f.getGuiTitle());
				}
			}

			parameters.add(new PluginParameter<String>(
					"spotFields",
					"Spot fields",
					"Select the spot data that should be exported.",
					new StringParameterType(null, null, false, 0, 0, 0, spotFields)
				));

			parameters.add(new PluginParameter<String>(
					"reporterFields",
					"Reporter fields",
					"Select the reporter annotations that should be exported.",
					new StringParameterType(null, null, false, 0, 0, 0, reporterFields)
				));
			
			parameters.add(new PluginParameter<String>(
					"assayFields",
					"Assay fields",
					"Select the assay annotations/properties that should be exported.",
					new StringParameterType(null, null, false, 0, 0, 0, assayFields)
				));
			
			configureJob = new RequestInformation(
					Request.COMMAND_CONFIGURE_JOB,
					"Select export options.",
					"Select the save path and other options such as " +
					"which fields to export.",
					parameters
				);
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return configureJob;
	}
	
	public Map<String, ExtendedExportableField> getExportableFields(DbControl dc, BioAssaySet source)
	{
		if (exportableFields != null) return exportableFields;
		exportableFields = new LinkedHashMap<String, ExtendedExportableField>();
		
		List<ExtendedExportableField> all = new ArrayList<ExtendedExportableField>();
		all.addAll(getSpotFields(source));
		all.addAll(getReporterFields());
		all.addAll(getFormulaFields(dc, source));
		all.addAll(getRawdataFields(source));
		all.addAll(getExtraValueFields(dc, source));
		all.addAll(getAssayFields(dc, source));
		
		for (ExtendedExportableField ef : all)
		{
			exportableFields.put(ef.getId(), ef);
		}
		return exportableFields;
	}

	private List<ExtendedExportableField> getFormulaFields(DbControl dc, BioAssaySet source)
	{
		List<ExtendedExportableField> items = new ArrayList<ExtendedExportableField>();
		ItemQuery<Formula> queryFormula = Formula.getQuery(Formula.Type.COLUMN_EXPRESSION, 
				source.getRawDataType(), source.getIntensityTransform());
		queryFormula.restrict(Restrictions.neq(Hql.property("valueType"), null));
		queryFormula.include(Include.MINE, Include.SHARED, Include.IN_PROJECT, Include.OTHERS);
		queryFormula.order(Orders.asc(Hql.property("name")));
		for (Formula f : queryFormula.list(dc))
		{
			DynamicField fld = ExportableFieldFactory.formula(f, f.getName(), null);
			items.add(new ExtendedExportableField("frm." + f.getId(), "[Formula] "+f.getName(), false, fld));
		}
		return items;
	}
	
	private List<ExtendedExportableField> getRawdataFields(BioAssaySet source)
	{
		Formula.AverageMethod NONE = Formula.AverageMethod.NONE;
		List<ExtendedExportableField> items = new ArrayList<ExtendedExportableField>();
		items.add(rawField("position", "Position", Type.INT, NONE));
		items.add(rawField("block", "Array block", Type.INT, NONE));
		items.add(rawField("metaGridX", "Meta grid X", Type.INT, NONE));
		items.add(rawField("metaGridY", "Meta grid Y", Type.INT, NONE));
		items.add(rawField("row", "Row in block", Type.INT, NONE));
		items.add(rawField("column", "Column in block", Type.INT, NONE));
		items.add(rawField("x", "X", Type.FLOAT, NONE));
		items.add(rawField("y", "Y", Type.FLOAT, NONE));
		for (RawDataProperty rdp : source.getRawDataType().getProperties())
		{
			items.add(rawField(rdp.getName(), rdp.getTitle(), rdp.getType(), rdp.getAverageMethod()));
		}
		return items;
	}

	private ExtendedExportableField rawField(String property, String title, Type type, Formula.AverageMethod averageMethod)
	{
		DynamicField df = ExportableFieldFactory.rawData(property, title, type, averageMethod, null);
		return new ExtendedExportableField("raw." + property, "[Raw] " + title, false, df);
	}

	public static Collection<ExtendedExportableField> getReporterFields()
	{
		Formula.AverageMethod NONE = Formula.AverageMethod.NONE;
		Formula.AverageMethod MAX = Formula.AverageMethod.MAX;
		List<ExtendedExportableField> items = new ArrayList<ExtendedExportableField>();
		items.add(reporterField("id", "Internal id", Type.INT, NONE));
		items.add(new ExtendedExportableField("rep.position", true, 
				ExportableFieldFactory.jep("pos()", "Position", Type.INT, NONE, null)));
		items.add(reporterField("name", "Name", Type.STRING, MAX));
		items.add(reporterField("externalId", "External id", Type.STRING, MAX));
		items.add(reporterField("symbol", "Gene symbol", Type.STRING, MAX));
		items.add(reporterField("description", "Description", Type.TEXT, MAX));
		items.add(reporterField("lastUpdate", "Last updated", Type.TIMESTAMP, MAX));
		
		for (ExtendedProperty ep : ExtendedProperties.getProperties("ReporterData"))
		{
			items.add(reporterField(ep.getName(), ep.getTitle(), ep.getType(), MAX));
		}
		return items;
	}

	
	private static ExtendedExportableField reporterField(String property, String guiTitle, Type type, Formula.AverageMethod averageMethod)
	{
		DynamicField df = ExportableFieldFactory.reporter(property, guiTitle, type, averageMethod, null);
		return new ExtendedExportableField("rep." + property, guiTitle, true, df);
	}
	
	private Collection<ExtendedExportableField> getExtraValueFields(DbControl dc, BioAssaySet source)
	{
		List<ExtendedExportableField> items = new ArrayList<ExtendedExportableField>();
		ItemQuery<ExtraValue> query = source.getExtraValues();
		query.order(Orders.asc(Hql.property("extraValueType.name")));
		for (ExtraValue ev : query.list(dc))
		{
			ExtraValueType evt = ev.getExtraValueType();
			DynamicField df = ExportableFieldFactory.extraValue(evt, evt.getName(), null);
			boolean positionValue = ev.getCoordinateType() == CoordinateType.POSITION;
			items.add(new ExtendedExportableField("xtra." + evt.getId(), "[Extra] " + evt.getName(), positionValue, df));
		}
		return items;
	}
	
	private Collection<ExtendedExportableField> getSpotFields(BioAssaySet source)
	{
		List<ExtendedExportableField> items = new ArrayList<ExtendedExportableField>();
		int channels = source.getRawDataType().getChannels();
		for (int channel = 1; channel <= channels; ++channel)
		{
			DynamicField df = ExportableFieldFactory.channel(channel, "Ch " + channel, null);
			items.add(new ExtendedExportableField("ch" + channel, false, df));
		}
		return items;
	}
	
	private List<ExtendedExportableField> getAssayFields(DbControl dc, BioAssaySet source)
	{
		List<ExtendedExportableField> items = new ArrayList<ExtendedExportableField>();
		items.add(assayField("id", "Internal id", Type.INT));
		items.add(assayField("name", "Name", Type.STRING));
		items.add(assayField("description", "Description", Type.TEXT));
		
		ItemQuery<AnnotationType> query = source.getExperiment().getExperimentalFactors();
		query.include(Include.ALL);
		query.order(Orders.asc(Hql.property("name")));
		
		for (AnnotationType at : query.list(dc))
		{
			AssayField df = ExportableFieldFactory.annotation(at, null);
			items.add(new ExtendedExportableField("annotation." + at.getId(), "[A] " + at.getName(), false, df));
		}
		return items;
	}
	
	private ExtendedExportableField assayField(String property, String guiTitle, Type type)
	{
		AssayField af = ExportableFieldFactory.assayProperty(property, guiTitle, type, null);
		return new ExtendedExportableField("assay." + property, guiTitle, false, af);
	}
	
	private java.io.File createFile(java.io.File dir, String name)
		throws IOException
	{
		java.io.File f = new java.io.File(dir, name);
		f.createNewFile();
		return f;
	}
 
	
	private File createFile(DbControl dc, Directory dir, String name)
	{
		File f = File.getFile(dc, dir, name, true);
		f.setCharacterSet("UTF-8");
		f.setMimeType("text/plain");
		if (!f.isInDatabase()) dc.saveItem(f);
		return f;
	}
 	
	public final static String PARAMETER_FILEFORMAT = "fileformat";
	
	/**
		Create a plug-in parameter that asks for the BFS subtype to use in
		the export.
	*/
	public static PluginParameter<String> getFileFormatParameter()
	{
		Enumeration<String, String> formats = new Enumeration<String, String>();
		formats.add(SerialBfsExporter.SUBTYPE, "Serial BFS");
		formats.add(MatrixBfsExporter.SUBTYPE, "Matrix BFS");
		return new PluginParameter<String>(
				PARAMETER_FILEFORMAT,
				"File format",
				"Please specify if you want to export a serial or matrix BFS. " +
				"The serial format has one data file for each assay, and the matrix format " +
				"has one data file for each selected spot field. In both cases, " +
				"reporter fields and assay fields are exported in two separate files.",
				new StringParameterType(255, SerialBfsExporter.SUBTYPE, true, 1, 0, 0, formats) 
			);
	}
	
	public final static String PARAMETER_FILENAMEGENERATOR = "filenamegenerator";

	/**
		Create a plug-in parameter that asks for the file name generation strategy.
	*/
	public static PluginParameter<String> getFileNameGeneratorParameter()
	{
		Enumeration<String, String> filenameGen = new Enumeration<String, String>();
		filenameGen.add("names", "Assay/formula names");
		filenameGen.add("sequence", "Sequence");
		return new PluginParameter<String>(
				PARAMETER_FILENAMEGENERATOR,
				"File name strategy",
				"Please select a data file name generation strategy.\n\n" +
				"Assay formula/names = file names are created based on the " +
				"names of assays or selected spot fields\n" +
				"Sequence = file names follows a strict sequnce, sdata-1.txt, " +
				"sdata-2.txt, etc.",
				new StringParameterType(255, "names", true, 1, 0, 0, filenameGen) 
			);
	}
	
 	public static class ExtendedExportableField
 	{

 		private final String id;
 		private final ExportableField parent;
 		private final String guiTitle;
 		private final boolean reporterField;
 		
 		public ExtendedExportableField(String id, boolean reporterField, ExportableField parent)
 		{
 			this(id, parent.getTitle(), reporterField, parent);
 		}
 		
 		public ExtendedExportableField(String id, String guiTitle, boolean reporterField, ExportableField parent)
 		{
 			this.id = id;
 			this.guiTitle = guiTitle;
 			this.reporterField = reporterField;
 			this.parent = parent;
 		}
		
		
		public String getId()
		{
			return id;
		}
		
		public String getGuiTitle()
		{
			return guiTitle;
		}
		
		public boolean isReporterField()
		{
			return reporterField;
		}
		
		public ExportableField getField()
		{
			return parent;
		}
 	}
 	
}
