/**
	$Id$

	Copyright (C) 2011 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Set;

import net.sf.basedb.core.AnnotationType;
import net.sf.basedb.core.AnyToAny;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BooleanParameterType;
import net.sf.basedb.core.DataFileType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.DerivedBioAssay;
import net.sf.basedb.core.File;
import net.sf.basedb.core.FileParameterType;
import net.sf.basedb.core.FileSet;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemParameterType;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ItemSubtype;
import net.sf.basedb.core.ItemSubtypeFileType;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.PluginConfiguration;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.RequestInformation;
import net.sf.basedb.core.StringParameterType;
import net.sf.basedb.core.plugin.AbstractAnalysisPlugin;
import net.sf.basedb.core.plugin.AnnotationSetterPlugin;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.plugin.InteractivePlugin;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.plugin.Response;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Orders;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.formatter.NameableFormatter;

/**
	Plug-in that allows a user to manually register an external analysis
	procedure for derived bioassays and to upload files that was generated by it.
	
	@author Nicklas
	@since 3.0
*/
public class ManualDerivedBioAssayCreator 
	extends AbstractAnalysisPlugin 
	implements InteractivePlugin, AnnotationSetterPlugin
{

	private RequestInformation configureJob;
	private RequestInformation configurePlugin;
	private List<PluginParameter<?>> toolParameters;
	
	
	public ManualDerivedBioAssayCreator()
	{}
	
	/*
		From the Plugin interface
		-------------------------------------------
	*/
	@Override
	public boolean supportsConfigurations()
	{
		return true;
	}
	@Override
	public boolean requiresConfiguration()
	{
		return true;
	}

	@SuppressWarnings("unchecked")
	@Override
	public void run(Request request, Response response, ProgressReporter progress) 
	{
		DbControl dc = null;
		try
		{
			dc = sc.newDbControl();
			
			// Get parameters
			DerivedBioAssay source = getSourceDerivedBioAssay(dc);
			Job currentJob = getCurrentJob(dc);
			String childName = (String)job.getValue(CHILD_NAME);
			String childDescription = (String)job.getValue(CHILD_DESCRIPTION);
			ItemSubtype childSubtype = (ItemSubtype)configuration.getValue("childSubtype");
			boolean copyAnnotations = Boolean.TRUE.equals(job.getValue(COPY_ANNOTATIONS));
			List<DataFileType> fileTypes = (List<DataFileType>)configuration.getValues("fileTypes");
			boolean allowMoreFiles = Boolean.TRUE.equals(configuration.getValue("allowMoreFiles"));
			
			// Create child items
			DerivedBioAssay child = DerivedBioAssay.getNew(dc, source, currentJob);
			dc.saveItem(child);
			child.setName(childName);
			child.setDescription(childDescription);
			child.setItemSubtype(childSubtype);
			
			// Copy annotations
			if (copyAnnotations)
			{
				PluginConfiguration cfg = getCurrentConfiguration(dc);
				if (cfg.isAnnotated()) child.getAnnotationSet().copyFrom(cfg, false);
			}
			
			// Set files
			if (fileTypes != null && fileTypes.size() > 0)
			{
				FileSet fs = child.getFileSet();
				for (DataFileType ft : fileTypes)
				{
					List<File> files = (List<File>)job.getValues("file." + ft.getExternalId());
					if (files != null && files.size() > 0)
					{
						for (File f : files)
						{
							fs.addMember(f, ft);
						}
					}
				}
				fs.validate(dc);
			}
			if (allowMoreFiles)
			{
				List<File> moreFiles = (List<File>)job.getValues("morefiles");
				if (moreFiles != null)
				{
					for (File f : moreFiles)
					{
						AnyToAny ata = AnyToAny.getNew(dc, child, f, f.getName(), false);
						dc.saveItem(ata);
					}
				}
			}
			
			dc.commit();
			response.setDone("Bioassay set " + childName + " created successfully");
		}
		catch (Throwable ex)
		{
			response.setError(ex.getMessage(), Arrays.asList(ex));
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	// ---------------------------------------------
	/*
		From the InteractivePlugin interface
		-------------------------------------------
	*/
	@Override
	public Set<GuiContext> getGuiContexts() 
	{
		return Collections.singleton(CONTEXT_DERIVEDBIOASSAY);
	}

	/**
		Adds check of the item subtype.
	 */
	@SuppressWarnings("unchecked")
	@Override
	public String isInContext(GuiContext context, Object item) 
	{
		String msg = super.isInContext(context, item);
		if (msg == null)
		{
			DerivedBioAssay dba = (DerivedBioAssay)item;
			List<ItemSubtype> subtypes = (List<ItemSubtype>)configuration.getValues("subtypes");
			if (subtypes != null && subtypes.size() > 0)
			{
				ItemSubtype subtype = dba.getItemSubtype();
				if (!subtypes.contains(subtype))
				{
					msg = "The subtype of '" + dba.getName() +  "' (";
					msg += subtype == null ? "null" : subtype.getName();
					msg += ") is not among the supported subtypes: " + Values.getString(subtypes, ", ", true, new NameableFormatter());
				}
			}
		}
		return msg;
	}

	@SuppressWarnings("unchecked")
	@Override
	public void configure(GuiContext context, Request request, Response response) 
	{
		String command = request.getCommand();
		if (command.equals(Request.COMMAND_CONFIGURE_PLUGIN))
		{
			RequestInformation ri = getConfigurePluginParameters(context);
			List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
			if (errors != null)
			{
				response.setError(errors.size() + " invalid parameter(s) were found in the request", errors);
				return;
			}
			
			storeValues(configuration, request, ri.getParameter("subtypes"));
			storeValue(configuration, request, ri.getParameter("childSubtype"));
			storeValues(configuration, request, ri.getParameter("fileTypes"));
			storeValue(configuration, request, ri.getParameter("allowMoreFiles"));
			storeValues(configuration, request, ri.getParameter("parameters"));
			response.setDone("Configuration complete");
		}
		else if (command.equals(Request.COMMAND_CONFIGURE_JOB))
		{
			RequestInformation ri = getConfigureJobParameters(context);
			List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
			if (errors != null)
			{
				response.setError(errors.size() + " invalid parameter(s) were found in the request", errors);
				return;
			}
			
			storeValue(job, request, ri.getParameter(SOURCE_DERIVEDBIOASSAY));
			storeValue(job, request, ri.getParameter(CHILD_NAME));
			storeValue(job, request, ri.getParameter(CHILD_DESCRIPTION));
			storeValue(job, request, ri.getParameter(COPY_ANNOTATIONS));
			
			// Files
			List<DataFileType> fileTypes = (List<DataFileType>)configuration.getValues("fileTypes");
			if (fileTypes != null)
			{
				for (DataFileType ft : fileTypes)
				{
					storeValues(job, request, ri.getParameter("file." + ft.getExternalId()));
				}
			}
			storeValues(job, request, ri.getParameter("morefiles"));

			// Parameters
			List<PluginParameter<?>> tp = getToolParameters((List<String>)configuration.getValues("parameters"));
			if (tp != null)
			{
				for (PluginParameter<?> pp : tp)
				{
					storeValue(job, request, pp);
				}
			}
			
			DerivedBioAssay bioAssay = (DerivedBioAssay)job.getValue(SOURCE_DERIVEDBIOASSAY);
			PluginConfiguration config = configuration.getPluginConfiguration();
			response.setSuggestedJobName("Manual analysis '" + config.getName() +
				"' on bioassay '" + bioAssay.getName() + "'");

			response.setExecuteImmediately("Job configuration complete", Job.ExecutionTime.SHORTEST, false);
		}
	}


	@Override
	public RequestInformation getRequestInformation(GuiContext context, String command) 
		throws BaseException 
	{
		RequestInformation requestInformation = null;
		if (command.equals(Request.COMMAND_CONFIGURE_JOB))
		{
			requestInformation = getConfigureJobParameters(context);
		}
		else if (command.equals(Request.COMMAND_CONFIGURE_PLUGIN))
		{
			requestInformation = getConfigurePluginParameters(context);
		}
		return requestInformation;
	}
	// ----------------------------------------------
	/*
		From the AnnotationSetterPlugin interface
		-------------------------------------------
	*/
	/**
		Return a query that finds annotation types for bioassay sets.
	*/
	public ItemQuery<AnnotationType> getAnnotationTypes()
	{
		return AnnotationType.getQuery(Item.BIOASSAYSET);
	}
	public String getAnnotationMessage()
	{
		return "The selected annotations will be copied to the child bioassay set that " +
				"is created by the manual transform.";
	}
	// ----------------------------------------------

	@SuppressWarnings("unchecked")
	private RequestInformation getConfigureJobParameters(GuiContext context)
	{
		if (configureJob != null) return configureJob;
		DbControl dc = null;
		List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();
		try
		{
			dc = sc.newDbControl();
			// Get the configuration options
			PluginConfiguration config = getCurrentConfiguration(dc);
			boolean allowMoreFiles = Boolean.TRUE.equals(configuration.getValue("allowMoreFiles"));
			List<DataFileType> fileTypes = (List<DataFileType>)configuration.getValues("fileTypes");
			boolean configIsAnnotated = config.isAnnotated();
			DerivedBioAssay source = getCurrentDerivedBioAssay(dc);
					
			// Source bioassay set
			parameters.add(getSourceDerivedBioAssayParameter(null, null));
			
			// Names of child bioassay etc.
			parameters.add(getChildNameParameter(null, null, "After " + config.getName()));
			parameters.add(getChildDescriptionParameter(null, null, null));

			// Annotations section
			if (configIsAnnotated)
			{
				parameters.add(annotationSection);
				parameters.add(getCopyAnnotationsParmeter(null, 
					"If selected, annotations are copied from the configuration to the " +
					"child bioassay set that is created by the manual transform.", true));
			}
			
			// Parameters section
			List<PluginParameter<?>> toolParameters = 
				getToolParameters((List<String>)configuration.getValues("parameters"));
			if (toolParameters != null && toolParameters.size() > 0)
			{
				parameters.add(new PluginParameter<String>(
						"parameterSection", 
						"Parameters", 
						"Enter parameter values relevent to running the external tool.", 
						null));
				parameters.addAll(toolParameters);
			}
			
			parameters.add(new PluginParameter<String>(
				"fileSection", 
				"Result files", 
				"Select the files that was generated by the external tool.", 
				null));
			// reset that current file, otherwise all parameters get the same 
			// file as the default value
			sc.getCurrentContext(Item.FILE).setId(0);
			if (fileTypes != null)
			{
				ItemSubtype childSubtype = (ItemSubtype)configuration.getValue("childSubtype");
				if (childSubtype != null) childSubtype = ItemSubtype.getById(dc, childSubtype.getId());
				
				for (DataFileType ft : fileTypes)
				{
					int multiplicty = 0;
					boolean required = false;
					if (childSubtype != null)
					{
						ItemSubtypeFileType info = childSubtype.getAssociatedDataFileType(ft, false);
						if (info != null)
						{
							multiplicty = info.getAllowMultiple() ? 0 : 1;
							required = info.isRequired();
						}
					}
					parameters.add(new PluginParameter<File>(
							"file." + ft.getExternalId(),
							ft.getName(),
							ft.getDescription(),
							new FileParameterType(null, required, multiplicty)
					));
				}
			}
			
			if (allowMoreFiles)
			{
				parameters.add(new PluginParameter<File>(
					"morefiles",
					"More files...",
					"Select any additional files that was generated by the external tool " +
					"that are not present as pre-defined file types.",
					new FileParameterType(null, false, 0)
				));
			}
			
		}
		finally
		{
			if (dc != null) dc.close();
		}
		
		configureJob = new RequestInformation(
			Request.COMMAND_CONFIGURE_JOB,
			"Files and parameters",
			"Enter information about the child bioassay set, the configuration and " +
			"parameters for the external tool and select/upload files that " +
			"was generated as a result",
			parameters
		);		
		return configureJob;
	}
	
	private RequestInformation getConfigurePluginParameters(GuiContext context)
	{
		if (configurePlugin != null) return configurePlugin;

		List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();
		DbControl dc = null;
		try
		{
			dc = sc.newDbControl();

			ItemQuery<ItemSubtype> subtypesQuery = ItemSubtype.getQuery(Item.DERIVEDBIOASSAY);
			subtypesQuery.order(Orders.asc(Hql.property("name")));
			List<ItemSubtype> subtypes = new ArrayList<ItemSubtype>(subtypesQuery.list(dc));
			parameters.add(new PluginParameter<ItemSubtype>(
				"subtypes",
				"Source bioassay subtypes",
				"Select the subtypes for the source derived bioassay that the external tool can be used with. " +
				"If no subtype is selected, it can be used on any derived bioassay",
				new ItemParameterType<ItemSubtype>(ItemSubtype.class, null, false, 0, subtypes)
			));
			
			parameters.add(new PluginParameter<ItemSubtype>(
				"childSubtype", 
				"Child subtype",
				"Select the subtype that should be assigned to the child derived bioassay " +
				"that is created as a result of the analysis.",
				new ItemParameterType<ItemSubtype>(ItemSubtype.class, null, false, 1, subtypes)
			));
			
			parameters.add(new PluginParameter<String>(
					"parameters",
					"Parameters",
					"Add parameters that are relevant for the external tool. All parameters " +
					"are assumed to be string parameters. Examples:\n" +
					"name\n" +
					"name<newline>helptext\n" +
					"name=option1|option2|option3\n",
					new StringParameterType(65536, null, false, 0, 0, 0)
				));
			
			ItemQuery<DataFileType> fileTypesQuery = DataFileType.getQuery();
			fileTypesQuery.restrict(Restrictions.eq(
					Hql.property("itemType"), 
					Expressions.integer(Item.DERIVEDBIOASSAY.getValue())
				));
			fileTypesQuery.order(Orders.asc(Hql.property("name")));
			List<DataFileType> fileTypes = new ArrayList<DataFileType>(fileTypesQuery.list(dc));
			parameters.add(new PluginParameter<DataFileType>(
				"fileTypes",
				"File types",
				"Select all registered file types that the external tool generates as " +
				"output.",
				new ItemParameterType<DataFileType>(DataFileType.class, null, false, 0, fileTypes)
			));
			
			parameters.add(new PluginParameter<Boolean>(
				"allowMoreFiles",
				"Allow more files",
				"Enable this option if the tool generates files that has not been " +
				"registered as file types with BASE. This allows a user to select " +
				"any number of files and attach those to the resulting bioassay.",
				new BooleanParameterType(null, false)
			));

			configurePlugin = new RequestInformation(
				Request.COMMAND_CONFIGURE_JOB,
				"Files and parameters",
				"This page lets you select which types of output files the external tool " +
				"generates and what parameters that can be given to external tool.",
				parameters);
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return configurePlugin;
	}
	
	private List<PluginParameter<?>> getToolParameters(List<String> values)
	{
		if (values == null) return null;
		if (toolParameters != null) return toolParameters;
		toolParameters = new ArrayList<PluginParameter<?>>(values.size());
		
		for (String value : values)
		{
			// Split on first newline
			if (value == null)
			{
				/* #### CONTINUE-STATEMENT #### */
				continue;
			}
			String[] temp = value.split("(\\n|\\r)+", 2);
			
			String name = temp[0];
			String description = temp.length > 1 ? temp[1] : null;
			int i = name.indexOf('=');
			List<String> options = null;
			if (i >= 1)
			{
				options = Arrays.asList(name.substring(i+1).split("\\|"));
				name = name.substring(0, i);
			}
			
			toolParameters.add(new PluginParameter<String>(
					"parameter." + name, 
					name, 
					description, 
					new StringParameterType(options != null ? 255 : 65536, null, false, 1, 0, 0 , options)
				));
		}
		return toolParameters;
	}
	
}
