/**
	$Id: BioSourceImporter.java 5646 2011-05-27 11:50:50Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins.batchimport;

import java.util.Collections;
import java.util.List;
import java.util.Set;

import net.sf.basedb.core.BioSource;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.util.parser.FlatFileParser;
import net.sf.basedb.util.parser.Mapper;

/**
	Plug-in for importing biosource items in a batch. The plug-in can create new 
	items and updated existing items.

	@author nicklas
	@version 2.8
	@base.modified $Date: 2011-05-27 13:50:50 +0200 (Fri, 27 May 2011) $
*/
public class BioSourceImporter
	extends AbstractItemImporter<BioSource>
{

	private static final Set<GuiContext> guiContexts =  
		Collections.singleton(new GuiContext(Item.BIOSOURCE, GuiContext.Type.LIST));
	
	public BioSourceImporter()
	{}

	/*
		From the InteractivePlugin interface
		------------------------------------
	*/
	@Override
	public Set<GuiContext> getGuiContexts()
	{
		return guiContexts;
	}
	// ------------------------------------

	/*
		From the AbstractItemImporter class
		------------------------------------
	*/
	
	private Mapper nameMapper;
	private Mapper descriptionMapper;
	private Mapper externalIdMapper;
	
	/**
		Useable methods are:
		<ul>
		<li>{@link PropertyIdMethod#NAME}
		<li>{@link PropertyIdMethod#EXTERNAL_ID}
		<li>{@link InternalIdMethod#INTERNAL_ID}
		<li>{@link FallbackIdMethod#NAME_OR_ID}
		</ul>
	*/
	@Override
	protected IdMethod[] getIdMethods()
	{
		return new IdMethod[] { PropertyIdMethod.NAME, PropertyIdMethod.EXTERNAL_ID, 
				InternalIdMethod.INTERNAL_ID, FallbackIdMethod.NAME_OR_ID};
	}
	
	@Override
	protected Item getItemForSubtypes()
	{
		return Item.BIOSOURCE;
	}
	
	/**
		Adds column mappings for name, externalId and description.
	*/
	@Override
	protected void addMoreColumnMappingParameters(List<PluginParameter<?>> parameters)
	{
		parameters.add(internalIdColumnMapping);
		parameters.add(nameColumnMapping);
		parameters.add(subtypeColumnMapping);
		parameters.add(externalIdColumnMapping);
		parameters.add(descriptionColumnMapping);
		parameters.add(permissionTemplateColumnMapping);
	}
	/**
		Calls {@link BioSource#getQuery()}.
	*/
	@Override
	protected ItemQuery<BioSource> createItemQuery()
	{
		return BioSource.getQuery();
	}
	@Override
	protected void createColumnMappers(FlatFileParser ffp, boolean cropStrings)
	{
		nameMapper = getMapper(ffp, (String)job.getValue("nameColumnMapping"), 
				cropStrings ? BioSource.MAX_NAME_LENGTH : null, null);
		descriptionMapper = getMapper(ffp, (String)job.getValue("descriptionColumnMapping"), 
				cropStrings ? BioSource.MAX_DESCRIPTION_LENGTH : null, null);
		externalIdMapper = getMapper(ffp, (String)job.getValue("externalIdColumnMapping"), 
				cropStrings ? BioSource.MAX_EXTERNAL_ID_LENGTH : null, null);
	}
	@Override
	protected BioSource createItem(DbControl dc, FlatFileParser.Data data)
	{
		BioSource bs = BioSource.getNew(dc);
		updateItem(dc, bs, data);
		return bs;
	}
	@Override
	protected void updateItem(DbControl dc, BioSource biosource, FlatFileParser.Data data)
	{
		if (nameMapper != null) biosource.setName(nameMapper.getValue(data));
		if (descriptionMapper != null) biosource.setDescription(descriptionMapper.getValue(data));
		if (externalIdMapper != null) biosource.setExternalId(externalIdMapper.getValue(data));
		updateItemSubtype(dc, biosource, data);
	}
	// ------------------------------------
}
