/*
	$Id $

	Copyright (C) 2010 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins.executor;

import java.io.IOException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.filter.ElementFilter;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

import net.sf.basedb.core.AnnotationType;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BioAssay;
import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.BooleanParameterType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.FloatParameterType;
import net.sf.basedb.core.Formula;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.IntegerParameterType;
import net.sf.basedb.core.ItemParameterType;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ParameterType;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.StringParameterType;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Orders;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.XMLUtil;

/**
	Represents a parameter definition for the external program.
	A name and type is required. All other settings are optional. 
	Supported values for type is:
	
	<ul>
	<li>string: The parameter is a free-text string parameter with a maximum length
		of 255 characters.
	<li>text: The parameter is a free-text string parameter with a maximum length
		of 64K characters.
	<li>integer, float: The parameter is a numeric parameter. Options 'min' and 'max' can
		be set to restrict the range of valid values.
	<li>boolean: The parameter is a true/false parameter.
	<li>enumeration: The parameter is an enumeration that must take it's value from a predefined
		list of values. The predefined list should be set in the 'enum' option with a line break
		as a delimiter between values.
	<li>factor: The parameter is an enumeration among the experimental factors that has been
		defined for the experiment. The default value is matched against the name/external id
		of the annotation types.
	<li>spotformula: The parameter is an enumeration among the formulas that can be used with
		the current source bioassay set. The default value is matched against the name
		of the formulas.
	<li>assays: The parameter is an enumeration among the bioassays in the current source
		bioassay set. The default value is matched against the name of the bioassays.
	</ul>
	
	@author Nicklas
	@since 2.15
*/
public class ParameterDefinition 
{

	// Url to the dtd-file that define the xml-file for external parameters
	private static final URL dtdFile = ParameterDefinition.class.getResource("/net/sf/basedb/core/dtd/external-plugin-parameters.dtd");

	/**
		Parse an XML document containing parameter definitions.
		@param xml The xml as a string (with or without a DOCTYPE declaration) 
		@param validate TRUE to validate the XML against the dtd (external-plugin-parameters.dtd)
		@return A list with parameter definition
	*/
	@SuppressWarnings("unchecked")
	public static List<ParameterDefinition> parseXml(String xml, boolean validate)
	{
		if (xml == null) return Collections.emptyList();
		List<ParameterDefinition> parameters = null;

		if (!xml.startsWith("<!DOCTYPE"))
		{
			xml = "<!DOCTYPE parameter-definition SYSTEM \"external-plugin-parameters.dtd\">" + xml;
		}
		try
		{
			Document doc = XMLUtil.getXml(xml, dtdFile, validate);
			parameters = new ArrayList<ParameterDefinition>(doc.getContentSize());
			Iterator<Element> it = doc.getDescendants(new ElementFilter("parameter"));
			Set<String> names = new HashSet<String>();
			while (it.hasNext())
			{
				Element e = it.next();
				ParameterDefinition pd = new ParameterDefinition();
				pd.setType(e.getAttributeValue("type"));
				pd.setName(e.getAttributeValue("name"));
				pd.setRequired(XMLUtil.getBooleanAttribute(e, "required", false));
				pd.setMultiple(XMLUtil.getBooleanAttribute(e, "multiple", false));
				pd.setDefaultValue(Values.getStringOrNull(e.getChildText("default")));
				pd.setDescription(Values.getStringOrNull(e.getChildText("description")));
				
				if (validate && !names.add(pd.getName()))
				{
					throw new BaseException("Duplicate parameter name: " + pd.getName());
				}
				
				Iterator<Element> options = e.getDescendants(new ElementFilter("option"));
				while (options.hasNext())
				{
					Element o = options.next();
					String key = o.getAttributeValue("key");
					String value = o.getText();
					pd.setOption(key, value);
				}
				parameters.add(pd);
			}
		}
		catch (JDOMException e)
		{
			throw new BaseException(e);
		}
		catch (IOException e)
		{
			throw new BaseException(e);
		}
		return parameters;
	}
	
	private String type;
	private String name;
	private boolean required;
	private boolean multiple;
	private String defaultValue;
	private String description;
	private Map<String, String> options;
	
	/**
		Create a new empty parameter definition. Before it can
		be used a name and type must be specified.
	*/
	public ParameterDefinition()
	{
		this.options = new HashMap<String, String>();
	}

	/**
		Get the type of the parameter.
	*/
	public String getType() 
	{
		return type;
	}
	/**
		Set the type of the parameter. For supported values see the
		class documentation.
	*/
	public void setType(String type) 
	{
		this.type = type;
	}

	/**
		Get the name of the parameter. This value is required.
	*/
	public String getName() 
	{
		return name;
	}

	/**
		Set the name of the parameter.
	*/
	public void setName(String name) 
	{
		this.name = name;
	}

	/**
		Is this parameter a required parameter or not?
	*/
	public boolean isRequired() 
	{
		return required;
	}

	/**
		Set if this parameter should be a required or optional parameter.
	*/
	public void setRequired(boolean required) 
	{
		this.required = required;
	}

	/**
		Can this parameter have multiple values or only one?
	*/
	public boolean isMultiple()
	{
		return multiple;
	}
	
	/**
		Set if this parameter can have multiple values or only one.
	*/
	public void setMultiple(boolean multiple)
	{
		this.multiple = multiple;
	}
	
	/**
		Get the default value for this parameter.
	*/
	public String getDefaultValue() 
	{
		return defaultValue;
	}

	/**
		Set the default value for this parameter.
	*/
	public void setDefaultValue(String defaultValue) 
	{
		this.defaultValue = defaultValue;
	}

	/**
		Get a description/help text about this parameter.
	*/
	public String getDescription() 
	{
		return description;
	}

	/**
		Set a description/help text for this parmeter.
	*/
	public void setDescription(String description) 
	{
		this.description = description;
	}
	
	/**
		Get the names of other options that are defined for this parmeter.
		Useful options depend on the parameter type and are described in
		the class documentation.
	*/
	public Set<String> getOptions()
	{
		return options.keySet();
	}
	
	/**
		Get the value for an option.
	*/
	public String getOption(String key)
	{
		return options.get(key);
	}
	
	/**
		Set the value for an option.
	*/
	public void setOption(String key, String value)
	{
		options.put(key, value);
	}
	
	/**
		Render this parameter definition as an XML fragment. Multiple fragments can be 
		concatenated and then wrapped with:
		<pre class="code>
&lt;!DOCTYPE parameter-definition SYSTEM "external-plugin-parameters.dtd"&gt;
&lt;parameter-definition&gt;
...
&lt;/parameter-definition&gt;
</pre>
		@param compact If TRUE, the generated XML is compacted and without line breaks.
		@return An XML representation of this parameter definition
	*/
	public String toXmlFragment(boolean compact)
	{
		Element parameter = new Element("parameter");
		parameter.setAttribute("type", getType());
		parameter.setAttribute("name", getName());
		parameter.setAttribute("required", isRequired() ? "true" : "false");
		parameter.setAttribute("mutliple", isMultiple() ? "true" : "false");
		
		if (getDescription() != null)
		{
			Element description = new Element("description");
			description.setText(getDescription());
			parameter.addContent(description);
		}
		if (getDefaultValue() != null)
		{
			Element defaultValue = new Element("default");
			defaultValue.setText(getDefaultValue());
			parameter.addContent(defaultValue);
		}
		
		if (options.size() > 0)
		{
			Element options = new Element("options");
			for (Map.Entry<String, String> option : this.options.entrySet())
			{
				Element opt = new Element("option");
				opt.setAttribute("key", option.getKey());
				opt.setText(option.getValue());
				options.addContent(opt);
			}
			parameter.addContent(options);
		}
		Format format = compact ? Format.getCompactFormat() : Format.getPrettyFormat();
		format.setEncoding("UTF-8").setOmitDeclaration(true);
		XMLOutputter out = new XMLOutputter(format);
		return out.outputString(parameter);
	}
	
	/**
		Convert this parameter definition to a PluginParameter object.
		@param dc A DbControl to use for database access
		@param source The source bioassay set
	*/
	@SuppressWarnings("unchecked")
	public PluginParameter<?> toPluginParameter(DbControl dc, BioAssaySet source)
	{
		String type = getType();
		ParameterType pType = null;
		String defaultValue = getDefaultValue();
		int multiplicity = isMultiple() ? 0 : 1;
		if ("string".equals(type))
		{
			pType = new StringParameterType(255, defaultValue, isRequired(), multiplicity, 0, 0);
		}
		else if ("text".equals(type))
		{
			pType = new StringParameterType(65535, defaultValue, isRequired(), multiplicity, 0, 0);
		}
		else if ("integer".equals(type))
		{
			Integer min = Values.getInteger(getOption("min"), null);
			Integer max = Values.getInteger(getOption("max"), null);
			pType = new IntegerParameterType(min, max, Values.getInteger(defaultValue, null), 
					isRequired(), multiplicity, 0, 0, null);
		}
		else if ("float".equals(type))
		{
			Float min = Values.getFloat(getOption("min"), null);
			Float max = Values.getFloat(getOption("max"), null);
			pType = new FloatParameterType(min, max, Values.getFloat(defaultValue, null), 
					isRequired(), multiplicity, 0, 0, null);
		}
		else if ("boolean".equals(type))
		{
			pType = new BooleanParameterType(defaultValue == null ? null : Values.getBoolean(defaultValue), isRequired());
		}
		else if ("enumeration".equals(type))
		{
			String[] options = getOption("enum").split("\n");
			pType = new StringParameterType(255, defaultValue, isRequired(), multiplicity, 0, 0, Arrays.asList(options));
		}
		else if ("factor".equals(type))
		{
			// Select among the experimental factors defined in the experiment
			ItemQuery<AnnotationType> query = source.getExperiment().getExperimentalFactors();
			query.include(Include.MINE, Include.SHARED, Include.IN_PROJECT, Include.OTHERS);
			query.order(Orders.asc(Hql.property("name")));
			
			AnnotationType defaultFactor = null;
			List<AnnotationType> allFactors = query.list(dc);
			if (defaultValue != null)
			{
				for (AnnotationType at : allFactors)
				{
					if (defaultValue.equalsIgnoreCase(at.getName()) || 
						defaultValue.equalsIgnoreCase(at.getExternalId())) 
					{
						defaultFactor = at;
						break;
					}
				}
			}
			pType = new ItemParameterType<AnnotationType>(AnnotationType.class, 
					defaultFactor, isRequired(), multiplicity, allFactors);
		}
		else if ("spotformula".equals(type))
		{
			// Select a formula to export as a spot field
			ItemQuery<Formula> query = Formula.getQuery(Formula.Type.COLUMN_EXPRESSION, 
					source.getRawDataType(), source.getIntensityTransform());
			query.include(Include.MINE, Include.SHARED, Include.IN_PROJECT, Include.OTHERS);
			query.order(Orders.asc(Hql.property("name")));
			Formula defaultFormula = null;
			List<Formula> allFormulas = query.list(dc);
			if (defaultValue != null)
			{
				for (Formula f : allFormulas)
				{
					if (defaultValue.equalsIgnoreCase(f.getName())) 
					{
						defaultFormula = f;
						break;
					}
				}
			}
			pType = new ItemParameterType<Formula>(Formula.class, 
					defaultFormula, isRequired(), multiplicity, allFormulas);

		}
		else if ("assays".equals(type))
		{
			// Select special assays
			ItemQuery<BioAssay> query = source.getBioAssays();
			query.order(Orders.asc(Hql.property("name")));
			BioAssay defaultAssay = null;
			List<BioAssay> allAssays = query.list(dc);
			if (defaultValue != null)
			{
				for (BioAssay ba : allAssays)
				{
					if (defaultValue.equalsIgnoreCase(ba.getName())) 
					{
						defaultAssay = ba;
						break;
					}
				}
			}
			pType = new ItemParameterType<BioAssay>(BioAssay.class, 
					defaultAssay, isRequired(), multiplicity, allAssays);			
		}
		return new PluginParameter(getName(), getName(), getDescription(), pType);
		
	}
	
	
}
