/**
$Id: TestPackedFileExporter.java 5319 2010-04-20 08:02:16Z nicklas $

Copyright (C) 2007 Nicklas Nordborg, Martin Svensson

This file is part of BASE - BioArray Software Environment.
Available at http://base.thep.lu.se/

BASE is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License, or
(at your option) any later version.

BASE is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Directory;
import net.sf.basedb.core.File;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemContext;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.ParameterType;
import net.sf.basedb.core.Path;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.PluginConfiguration;
import net.sf.basedb.core.PluginConfigurationRequest;
import net.sf.basedb.core.PluginDefinition;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.PluginResponse;
import net.sf.basedb.core.RequestInformation;
import net.sf.basedb.core.SessionControl;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.plugin.Response;

import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.zip.GZIPInputStream;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

import org.apache.tools.bzip2.CBZip2InputStream;

import com.ice.tar.TarEntry;
import com.ice.tar.TarInputStream;

public class TestPackedFileExporter
{
	static boolean ok = true;

	/**	
	 @param args
	 */
	public static void main(String[] args)
	{
		TestUtil.checkArgs(args);
		TestUtil.begin();
		ok = test_all();
		TestUtil.stop();
	}

	static boolean test_all()
	{
		write("++Testing TestPackedFileExporter");
		String packedDir = "packedFiles";
		// Directory where packed files are created
		int packedDirId = TestDirectory.test_create(true, packedDir);
		
		// Upload zip file to '/unpackzip'
		int dirId = TestDirectory.test_create(true, "unpackzip");
		TestFile.test_unpack_zip("data/test.unpack.zip", dirId);
		
		// Files and directories to pack
		List<String> filePaths = new ArrayList<String>();
		List<String> directoryPaths = new ArrayList<String>();
		filePaths.add("/unpackzip/index.html");
		directoryPaths.add("/unpackzip/images");
		set_selected(filePaths, directoryPaths);
		
		int pluginId = TestPluginDefinition.test_get("net.sf.basedb.plugins.PackedFileExporter");

		// Create job to zip files
		int zipConfig = test_create_config(pluginId, "Zip", "net.sf.basedb.util.zip.ZipFilePacker");
		Path zip = new Path("/" + packedDir + "/test.compressed.file.zip", Path.Type.FILE);
		int zipJobId = test_create_job(zipConfig, dirId, zip.toString());
		ok = TestJob.test_execute(zipJobId, false) && ok;
		test_read_zip(zip);

		// Create job to pack files in tar
		int tarConfig = test_create_config(pluginId, "Tar", "net.sf.basedb.util.zip.TarFilePacker");
		Path tar = new Path("/" + packedDir + "/test.compressed.file.tar", Path.Type.FILE);
		int tarJobId = test_create_job(tarConfig, dirId, tar.toString());
		ok = TestJob.test_execute(tarJobId, false) && ok;
		test_read_tar(tar, false, false);
		
		// Create job to pack files in tar.gz
		int tarGzConfig = test_create_config(pluginId, "Tar.gz", "net.sf.basedb.util.zip.GzipFilePacker");
		Path tarGz = new Path("/" + packedDir + "/test.compressed.file.tar.gz", Path.Type.FILE);
		int tarGzJobId = test_create_job(tarGzConfig, dirId, tarGz.toString());
		ok = TestJob.test_execute(tarGzJobId, false) && ok;
		test_read_tar(tarGz, true, false);
		
		// Create job to pack files in tar.bz2
		int tarBz2Config = test_create_config(pluginId, "Tar.bz2", "net.sf.basedb.util.zip.Bzip2FilePacker");
		Path tarBz2 = new Path("/" + packedDir + "/test.compressed.file.tar.bz2", Path.Type.FILE);
		int tarBz2JobId = test_create_job(tarBz2Config, dirId, tarBz2.toString());
		ok = TestJob.test_execute(tarBz2JobId, false) && ok;
		test_read_tar(tarBz2, false, true);
		
		// Clean up
		if (TestUtil.waitBeforeDelete()) TestUtil.waitForEnter();
		TestJob.test_delete(zipJobId);
		TestJob.test_delete(tarJobId);
		TestJob.test_delete(tarGzJobId);
		TestJob.test_delete(tarBz2JobId);
		TestPluginConfiguration.test_delete(zipConfig);
		TestPluginConfiguration.test_delete(tarConfig);
		TestPluginConfiguration.test_delete(tarGzConfig);
		TestPluginConfiguration.test_delete(tarBz2Config);
		TestDirectory.test_delete(packedDirId, true);
		TestDirectory.test_delete(dirId, true);		
				
		write("++Testing PackedFileExporter "+(ok ? "OK" : "Failed")+"\n");
		return ok;
	}

	static int test_create_config(int pluginId, String name, String packerClass)
	{
		if (pluginId == 0 || !TestUtil.hasPermission(Permission.CREATE, Item.PLUGINCONFIGURATION)) return 0;
		int id = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();

			PluginDefinition pd = PluginDefinition.getById(dc, pluginId);
			PluginConfiguration pc = pd.newPluginConfiguration();
			pc.setName(name);
			
			PluginConfigurationRequest request = pc.configure();
			write_request_information(request.getRequestInformation());
			request.setParameterValue("packer", packerClass);
			
			PluginResponse response = request.invoke();
			if (response.getStatus() == Response.Status.DONE)
			{
				response.saveParameters(dc);
				dc.saveItem(pc);
				dc.commit();
			}
			else
			{
				throw new BaseException(response.getMessage(), response.getErrorList().get(0));
			}
			id = pc.getId();
			write("--Create configuration for " + name + " file export OK");
		}
		catch (Throwable ex)
		{
			write("--Create configuration for " + name  + " file export FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return id;
	}
	
	static void set_selected(List<String> filePaths, List<String> dirPaths)
	{
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			SessionControl sc = TestUtil.getSessionControl();
			ItemContext fileCC = sc.getCurrentContext(Item.FILE);
			ItemContext dirCC = sc.getCurrentContext(Item.DIRECTORY);
			
			for (String path : filePaths)
			{
				File f = File.getByPath(dc, new Path(path, Path.Type.FILE), false);
				fileCC.getSelected().add(f.getId());
			}
			for (String path : dirPaths)
			{
				Directory d = Directory.getByPath(dc, new Path(path, Path.Type.DIRECTORY));
				dirCC.getSelected().add(d.getId());
			}
			
			write("--Set selected files and directories OK");
		}
		catch (Throwable ex)
		{
			write("--Set selected files and directories FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}		
	}
	
	static int test_create_job(int configId, int rootDirId, String saveAs)
	{
		if (!TestUtil.hasPermission(Permission.CREATE, Item.JOB) || configId == 0) return 0;
		int id = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			PluginDefinition pd = PluginDefinition.getByClassName(dc, "net.sf.basedb.plugins.PackedFileExporter");
			
			PluginConfiguration config = PluginConfiguration.getById(dc, configId);
			
			Job job = Job.getNew(dc, pd, config, null);
			job.setName("Test " + config.getName());
			job.setSendMessage(false);
			PluginConfigurationRequest request = job.configure(new GuiContext(Item.FILE, GuiContext.Type.LIST));
			
			request.setParameterValue("root", Directory.getById(dc, rootDirId));
			request.setParameterValue("overwrite", true);
			request.setParameterValue("saveAs", saveAs);
			request.setParameterValue("removeItems", false);
		
			write_request_information(request.getRequestInformation());
			PluginResponse response = request.invoke();
			
			if (response.getStatus() == Response.Status.DONE)
			{
				response.saveParameters(dc);
				dc.saveItem(job);
				dc.commit();
			}
			else
			{
				throw new BaseException("Status: " + response.getStatus() +"\n" + response.getMessage(), response.getErrorList().get(0));
			}
			id = job.getId();
			write("--Create job for PackedFileExporter OK");
		}
		catch (Throwable ex)
		{
			write("--Create job for PackedFileExporter FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		 	
		return id;
	}

	static void write(String message)
	{
		System.out.println(message);
	}
	
	static void write_request_information(RequestInformation ri)
	{
		if (TestUtil.getSilent()) return;

		write("Request information for command: "+ri.getCommand());
		write("\tTitle:\t"+ri.getTitle());
		write("\tDescription:\t"+ri.getDescription());
		for (PluginParameter<?> pp : ri.getParameters())
		{
			ParameterType pt = pp.getParameterType();
			System.out.println("\tParameter:\t"+pp.getName() + "\t" + pp.getLabel() + "\t" + pt );
		}
	}
	
	/**
		Test to read a zip file located in BASE's file system
		The content of the zip file is listed if not silent. 
	  	@param zipPath The path to the zip-file that should read
	*/
	static void test_read_zip(Path zipPath)
	{
		if (!TestUtil.hasPermission(Permission.READ, Item.FILE)) return;
		DbControl dc = null;		
		ZipEntry entry = null;
		try
		{
			dc = TestUtil.getDbControl();
			File zipFile = File.getByPath(dc, zipPath, false);
			ZipInputStream zios = new ZipInputStream(zipFile.getDownloadStream(0));
			if (!TestUtil.getSilent()) write ("--Contents in: " + zipFile.getName());
			while ((entry = zios.getNextEntry()) != null)
			{
				if (!TestUtil.getSilent())
				{
					write("--Entry in zipfile: " + entry.getName());
				}
			}
			write ("--Read zip-file OK: " + zipPath.toString());
		}
		catch (Throwable t)
		{
			write ("--Read zip-file FAILED: " + zipPath.toString());
			t.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	/**
		Test to read a tar-file and write the files to prompt.
	  	@param tarPath A Path to the file that should be read.  
	*/
	static void test_read_tar(Path tarPath, boolean isGzip, boolean isBzip)
	{
		if (!TestUtil.hasPermission(Permission.READ, Item.FILE)) return;
		DbControl dc = null;
		TarEntry entry = null;
		try
		{
			dc = TestUtil.getDbControl();
			File tarFile = File.getByPath(dc, tarPath, false);
			InputStream download = tarFile.getDownloadStream(0);
			if (isGzip) download = new GZIPInputStream(download);
			if (isBzip) 
			{
				download.read(); // 'B'
				download.read(); // 'Z'
				download = new CBZip2InputStream(download);
			}
			TarInputStream tos = new TarInputStream(download);
			if (!TestUtil.getSilent()) write ("--Contents in: " + tarFile.getName());
			while ((entry = tos.getNextEntry()) != null)
			{
				if (!TestUtil.getSilent())
				{
					write ("--Entry in tarfile: " +  entry.getName());
				}
			}
			tos.close();
			write ("--Read tar-file OK: " + tarPath.toString());
		}		
		catch (Throwable t)
		{
			write ("--Read tar-file FAILED: " + tarPath.toString());
			t.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}	

}
