/*
	$Id: TestSample.java 5696 2011-08-12 12:51:53Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
import net.sf.basedb.core.*;
import net.sf.basedb.core.query.ResultList;
import net.sf.basedb.util.biomaterial.BioSourceToSampleTransformer;
import net.sf.basedb.util.biomaterial.ChildrenTransformer;
import net.sf.basedb.util.biomaterial.ParentsTransformer;
import net.sf.basedb.util.biomaterial.SampleToBioSourceTransformer;

import java.util.Collections;
import java.util.Date;
import java.util.LinkedHashSet;
import java.util.Set;

public class TestSample
{

	static boolean ok = true;
	public static void main(String[] args)
	{
		TestUtil.checkArgs(args);
		TestUtil.begin();
		ok = test_all();
		TestUtil.stop();
	}

	static boolean test_all()
	{
		write("++Testing samples");
		write_header();
		// Standard tests: create, load, list
		int bioSourceId = TestBioSource.test_create(null, false);
		int id = test_create(0, "Sample #1 (standalone)", true);
		int id2 = test_create(bioSourceId, "Sample #2 (biosource parent)", false); 
		int id3 = test_create(id, "Sample #3 (sample parent)", false);
		int id4 = test_create_pooled("Sample #4 (pooled)", id, id2, id3);

		test_load(id);
		test_list(-1);
		
		write_events_header();
		test_add_event(id, 100.0f);
		test_list_events(id, 4);  // creation event, create event for id3+id4, and extra event
		
		write_parents_header();
		test_list_sources(id4, 3); // The three id, id2 and id3 samples
		test_list_children(id, 2); // id3 and id4

		// Transformers
		test_childrenTransformer(id, 2);
		test_parentsTransformer(id4, 3);
		test_sampleToBioSourceTransformer(id2, 1);
		test_bioSourceToSampleTransformer(bioSourceId, 1);

		// Standard test: Delete
		if (TestUtil.waitBeforeDelete()) TestUtil.waitForEnter();
		test_delete(id4);
		test_delete(id3);
		test_delete(id2);
		test_delete(id);
		TestBioSource.test_delete(bioSourceId);
	
		write("++Testing samples "+(ok ? "OK" : "Failed")+"\n");
		return ok;
	}

	static int test_create(int parentId, String name, boolean setAll)
	{
		if (!TestUtil.hasPermission(Permission.CREATE, Item.SAMPLE)) return 0;
		int id = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Sample s = Sample.getNew(dc);
			if (setAll)
			{
				s.setName("Test sample");
				s.setDescription("Added at "+new Date());
				s.setExternalId("XCV-455-SAMPLE");
				s.setOriginalQuantity(500.0f);
			}
			if (parentId != 0)
			{
				BioMaterial parent = BioMaterial.getById(dc, parentId);
				s.getCreationEvent().setSource(parent);
				s.setName(parent.getName() + ".s");
			}
			if (name != null) s.setName(name);
			dc.saveItem(s);
			dc.commit();
			id = s.getId();
			dc = TestUtil.getDbControl();
			dc.reattachItem(s, false);
			write_item(0, s);
			write("--Create sample OK");
		}
		catch (Throwable ex)
		{
			write("--Create sample FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return id;
	}

	static int test_create_pooled(String name, int... sourceIds)
	{
		if (!TestUtil.hasPermission(Permission.CREATE, Item.SAMPLE)) return 0;
		int id = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Sample s = Sample.getNew(dc);
			s.setName(name);
			s.setDescription("Added at "+new Date());
			s.setExternalId("POOLED-455-SAMPLE");
			s.setOriginalQuantity(500.0f);
			BioMaterialEvent evt = s.getCreationEvent();
			for (int sourceId : sourceIds)
			{
				Sample src = Sample.getById(dc, sourceId);
				Float original = src.getOriginalQuantity();
				BioMaterialEventSource eventSource = evt.addSource(src);
				if (original != null)
				{
					eventSource.setUsedQuantity(original/2);
				}
			}
			dc.saveItem(s);
			dc.commit();
			id = s.getId();
			dc = TestUtil.getDbControl();
			dc.reattachItem(s, false);
			write_item(0, s);
			write("--Create pooled sample OK");
		}
		catch (Throwable ex)
		{
			write("--Create pooled sample FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return id;
	}

	static void test_load(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Sample s = Sample.getById(dc, id);
			write_item(0, s);
			write("--Load sample OK");
		}
		catch (Throwable ex)
		{
			write("--Load sample FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static boolean test_list(int expectedResults)
	{
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			ItemResultList<Sample> l = Sample.getQuery().list(dc);
			for (int i = 0; i<l.size(); i++)
			{
				write_item(i, l.get(i));
			}
			if (expectedResults >= 0 && expectedResults != l.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+l.size());
			}
			write("--List samples OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List samples FAILED");
			ex.printStackTrace();
			ok = false;
			return false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return true;
	}

	static void test_delete(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Sample s = Sample.getById(dc, id);
			dc.deleteItem(s);
			Set<ItemProxy> using = s.getUsingItems();
			if (using.size() > 0) 
			{
			   throw new BaseException(s + " is used by " + using);
			}
			dc.commit();
			write("--Delete sample OK");
		}
		catch (Throwable ex)
		{
			write("--Delete sample FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_delete_all()
	{
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			ItemResultList<Sample> l = Sample.getQuery().list(dc);
			int numDeleted = Trashcan.delete(dc.getSessionControl(), l, true, null);
			if (numDeleted != l.size()) throw new BaseException("Could not delete all items: " + numDeleted + " of " + l.size());
			dc.commit();
			write("--Delete all samples OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--Delete all samples FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	static void write_header()
	{
		if (!TestUtil.getSilent())
		{
			write("   \tID \tName      \tDescription\tExternal id\tParent\tQuantity\tRemain");
			write("-- \t-- \t--------- \t-----------\t-----------\t------\t--------\t------");
		}
	}
	static void write_item(int i, Sample s)
		throws BaseException
	{
		if (!TestUtil.getSilent()) System.out.println(i+":\t"+s.getId()+"\t"+s.getName()+"\t"+s.getDescription()+
			"\t"+s.getExternalId()+"\t"+s.getParent()+"\t"+s.getOriginalQuantity()+"\t"+s.getRemainingQuantity());
	}
	static void write_events_header()
	{
		if (!TestUtil.getSilent())
		{
			write("   \tID \tBiomaterial\tEvent type\tEvent date\tEntry date\tUsed quantity\tComment");
			write("-- \t-- \t-----------\t----------\t----------\t----------\t-------------\t-------");
		}
	}
	static void write_item(int i, BioMaterialEvent evt, Sample s)
		throws BaseException
	{
		if (!TestUtil.getSilent()) System.out.println(i+":\t"+evt.getId()+"\t"+evt.getBioMaterial()+"\t"+evt.getEventType()+
			"\t"+evt.getEventDate()+"\t"+evt.getEntryDate()+"\t"+evt.getUsedQuantity(s)+"\t"+evt.getComment());
	}

	static void write_parents_header()
	{
		if (!TestUtil.getSilent())
		{
			write("   \tID \tName      \tDescription\tExternal id\tQuantity\tRemain\tUsed");
			write("-- \t-- \t--------- \t-----------\t-----------\t--------\t------\t----");
		}
	}

	static void write_item(int i, BioMaterialEventSource evtSrc)
		throws BaseException
	{
		if (!TestUtil.getSilent()) 
		{
			MeasuredBioMaterial bm = (MeasuredBioMaterial)evtSrc.getBioMaterial();
			System.out.println(i+":\t"+bm.getId()+"\t"+bm.getName()+"\t"+bm.getDescription()+
					"\t"+bm.getExternalId()+"\t"+"\t"+bm.getOriginalQuantity()+"\t"+bm.getRemainingQuantity()+
					"\t"+evtSrc.getUsedQuantity());
		}
	}

	static void write(String message)
	{
		System.out.println(message);
	}
	
	static int test_add_event(int sampleId, float usedQuantity)
	{
		if (sampleId == 0) return 0;
		int eventId = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Sample s = Sample.getById(dc, sampleId);
			BioMaterialEvent bme = s.newEvent();
			bme.setUsedQuantity(usedQuantity);
			bme.setComment("Added a new event");
			dc.saveItem(bme);
			dc.commit();
			eventId = bme.getId();
			dc = TestUtil.getDbControl();
			dc.reattachItem(bme, false);
			write_item(0, bme, s);
			write("--Add event for sample OK");
		}
		catch (Throwable ex)
		{
			write("--Add event for sample FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return eventId;
	}
	
	static void test_list_events(int sampleId, int expectedResults)
	{
		if (sampleId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Sample s = Sample.getById(dc, sampleId);
			ItemResultList<BioMaterialEvent> l = s.getEvents().list(dc);
			for (int i = 0; i<l.size(); i++)
			{
				write_item(i, l.get(i), s);
			}
			if (expectedResults >= 0 && expectedResults != l.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+l.size());
			}
			write("--List events for sample OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List events for sample FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_list_sources(int sampleId, int expectedResults)
	{
		if (sampleId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Sample s = Sample.getById(dc, sampleId);
			BioMaterialEvent evt = s.getCreationEvent();
			ResultList<BioMaterialEventSource> l = evt.getEventSources().list(dc);
			for (int i = 0; i<l.size(); i++)
			{
				write_item(i, l.get(i));
			}
			if (expectedResults >= 0 && expectedResults != l.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+l.size());
			}
			write("--List sources for sample OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List sources for sample FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_list_children(int sampleId, int expectedResults)
	{
		if (sampleId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Sample s = Sample.getById(dc, sampleId);
			ItemResultList<Sample> ev = s.getChildSamples().list(dc);
			
			for (int i=0; i<ev.size(); i++)
			{
				write_item(i, ev.get(i));
			}
			if (expectedResults >= 0 && expectedResults != ev.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+ev.size());
			}
			write("--List children for sample OK ("+ev.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List children for sample FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	static void test_childrenTransformer(int sampleId, int expectedResults)
	{
		if (sampleId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Sample s = Sample.getById(dc, sampleId);
			ChildrenTransformer<Sample> pct = new ChildrenTransformer<Sample>(dc, false, Sample.getQuery());
			Set<Sample> children = new LinkedHashSet<Sample>();
			pct.transform(Collections.singleton(s), children);

			int i = 0;
			for (Sample c : children)
			{
				write_item(i, c);
				i++;
			}
			if (expectedResults >= 0 && expectedResults != children.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+children.size());
			}
			write("--List children for sample using transformer OK ("+children.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List children for sample using transformer FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_parentsTransformer(int sampleId, int expectedResults)
	{
		if (sampleId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Sample s = Sample.getById(dc, sampleId);
			ParentsTransformer<Sample> pct = new ParentsTransformer<Sample>(dc, false, Sample.getQuery());
			Set<Sample> parents = new LinkedHashSet<Sample>();
			pct.transform(Collections.singleton(s), parents);

			int i = 0;
			for (Sample c : parents)
			{
				write_item(i, c);
				i++;
			}
			if (expectedResults >= 0 && expectedResults != parents.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+parents.size());
			}
			write("--List parents for sample using transformer OK ("+parents.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List parents for sample using transformer FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_sampleToBioSourceTransformer(int sampleId, int expectedResults)
	{
		if (sampleId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Sample s = Sample.getById(dc, sampleId);
			SampleToBioSourceTransformer pct = new SampleToBioSourceTransformer(dc);
			Set<BioSource> parents = new LinkedHashSet<BioSource>();
			pct.transform(Collections.singleton(s), parents);

			int i = 0;
			for (BioSource c : parents)
			{
				TestBioSource.write_item(i, c);
				i++;
			}
			if (expectedResults >= 0 && expectedResults != parents.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+parents.size());
			}
			write("--List biosource for sample using transformer OK ("+parents.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List biosource for sample using transformer FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_bioSourceToSampleTransformer(int bioSourceId, int expectedResults)
	{
		if (bioSourceId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			BioSource bs = BioSource.getById(dc, bioSourceId);
			BioSourceToSampleTransformer pct = new BioSourceToSampleTransformer(dc);
			Set<Sample> children = new LinkedHashSet<Sample>();
			pct.transform(Collections.singleton(bs), children);

			int i = 0;
			for (Sample c : children)
			{
				write_item(i, c);
				i++;
			}
			if (expectedResults >= 0 && expectedResults != children.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+children.size());
			}
			write("--List sample for biosource using transformer OK ("+children.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List sample for biosource using transformer FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
}
