/* $Id $
	------------------------------------------------------------------
	Copyright (C) 2010 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
	------------------------------------------------------------------

	JavaScript functions for the bioplates and well mapping

	@author Nicklas
	@version 2.16
*/

/**
	A list of biomaterial items. The prefix is used 
	to find DOM elements for the items in the list. Each
	item should have a prefix.item-id element and may
	also have a prefix.item-id.info and prefix.item-id.label 
	elements.
	
	Use addItem(id, name) to add items to the list.

*/
function ItemList(prefix)
{
	this.prefix = prefix;
	this.items = new Array();
	this.selectedItemClass = 'selected';
	this.linkedItemClass = 'linked';
	this.mappedItemClass = 'mapped';
	
	/**
		Get the main DOM tag for the item list. It's ID should be
		the same as the list prefix.
	*/
	ItemList.prototype.getTag = function()
	{
		if (!this.tag)
		{
			this.tag = document.getElementById(this.prefix);
		}
		return this.tag;
	}

	
	/**
		Add an item to the list. The DOM should contain an element
		with id=prefix.item-id
	*/
	ItemList.prototype.addItem = function(id, name)
	{
		var item = new Item(this, id, name);
		this.items[this.items.length] = item;
		this.items['item.' + id] = item;
		return item;
	}

	/**
		Get the item with the given id.
	*/
	ItemList.prototype.getItem = function(id)
	{
		return this.items['item.' + id];
	}
	
	/**
		Get the next unmapped item after the given item in the list.
	*/
	ItemList.prototype.nextUnmappedItem = function(item)
	{
		var i = 0;
		// Find the given item
		while (i < this.items.length && this.items[i] != item)
		{
			i++;
		}
		// Move to the next until an unmapped item is found
		i++;
		while (i < this.items.length)
		{
			if (!this.items[i].mappedWell) return this.items[i];
			i++;
		}
		return null;
	}
	
	ItemList.prototype.unmapAll = function(graphics)
	{
		for (var i = 0; i < this.items.length; i++)
		{
			var item = this.items[i];
			item.hideLink(graphics);
			item.unmapWell();
		}
	}
}

/**
	A single item in a list of biomaterial. Use ItemList.addItem() to 
	add items to the list.
	@param list The list object
	@param id The id of the item
	@param name The name (label) of the item
*/
function Item(list, id, name)
{
	this.list = list;
	this.id = id;
	this.name = name;
	this.selected = false;
	this.linked = false;
	this.mappedWell = null;
	
	/**
		Get the main DOM tag for the item. It should have and ID combined
		of 'list-prefix.item-id'
	*/
	Item.prototype.getTag = function()
	{
		if (!this.tag)
		{
			this.tag = document.getElementById(this.list.prefix+'.'+this.id);
		}
		return this.tag;
	}
	
	/**
		Add a CSS class to the item's DOM tag.
	*/
	Item.prototype.addClass = function(cls)
	{
		Main.addClass(this.getTag(), cls);
	}

	/**
		Remove a CSS class from the item's DOM tag.
	*/
	Item.prototype.removeClass = function(cls)
	{
		Main.removeClass(this.getTag(), cls);
	}
	
	/**
		Set the contents of the info element. The tag should have an ID combined
		of 'list-prefix.item-id.info'.
	*/
	Item.prototype.setInfo = function(info)
	{
		var infoTag = document.getElementById(this.list.prefix+'.'+this.id + '.info');
		if (!infoTag) return;
		infoTag.innerHTML = info;
	}
	
	/**
		Mark the item as selected/deselected.
	*/
	Item.prototype.setSelected = function(selected)
	{
		this.selected = selected;
		Main.addOrRemoveClass(this.getTag(), this.list.selectedItemClass, selected);
	}
	
	/**
		Mark the item as linked. Do not call this method. Use drawLink/hideLink 
		instead.
	*/
	Item.prototype.setLinked = function(linked)
	{
		this.linked = linked;
		Main.addOrRemoveClass(this.getTag(), this.list.linkedItemClass, linked);
	}
	
	/**
		Draw a link from the item to the well it is mapped to. If
		the item is not mapped, no link is drawn.
		
		@param graphics A jsGraphics object
		@param pen A jsGraphics pen object
		@param overwrite TRUE to overwrite existing link, FALSE to not draw if there already is a link
	*/
	Item.prototype.drawLink = function(graphics, pen, overwrite)
	{
		if (!this.mappedWell) return;
		if (this.linkDiv) 
		{
			if (!overwrite) return;
			this.hideLink(graphics);
		}
		
		var itemPos = Main.getElementPosition(this.getTag(), true);
		var wellPos = Main.getElementPosition(this.mappedWell.getTag());
		var listPos = Main.getElementPosition(this.list.getTag(), true);
		// Do not draw the link if the item is not visible
		if (itemPos.top < listPos.top || (itemPos.bottom) > (listPos.bottom)) return;
		
		var points = new Array();
		points[0] = new jsPoint(itemPos.right, itemPos.top+itemPos.height/2);
		points[1] = new jsPoint(wellPos.left, wellPos.top+wellPos.height/2);
		this.linkDiv = graphics.drawLine(pen, points[0], points[1]);
		this.setLinked(true);
		this.mappedWell.setLinked(true);
	}
	
	/**
		Hide the link from the item to the well it is mapped to.
	*/
	Item.prototype.hideLink = function(graphics)
	{
		if (!this.linkDiv) return;
		graphics.clearDrawing(this.linkDiv);
		this.linkDiv = null;
		this.setLinked(false);
		if (this.mappedWell)
		{
			this.mappedWell.setLinked(false);
		}
	}
	
	/**
		Map the item to the given well. Existing mapping for the
		current item and well are automatically removed. The
		item's info tag is updated with the coordinate of the well.
	*/
	Item.prototype.mapToWell = function(well)
	{
		this.unmapWell();
		well.unmapItem();
		this.mappedWell = well;
		well.mappedItem = this;
		this.setInfo(well.getCoordinate());
		this.addClass(this.list.mappedItemClass);
		well.addClass(well.plate.mappedWellClass);
	}
	
	/**
		Remove the current mapping.
	*/
	Item.prototype.unmapWell = function()
	{
		if (!this.mappedWell) return;
		this.setInfo('');
		this.removeClass(this.list.mappedItemClass);
		this.mappedWell.removeClass(this.mappedWell.plate.mappedWellClass);
		this.mappedWell.mappedItem = null;
		this.mappedWell = null;
	}
	
	/**
		Scroll the list so that this item comes into view. If the item is already visible
		nothing is changed
	*/
	Item.prototype.scrollIntoView = function()
	{
		var itemTag = this.getTag();
		var itemPos = Main.getElementPosition(itemTag, true);
		var listTag = this.list.getTag();
		var listPos = Main.getElementPosition(listTag, true);
		
		if (itemPos.top < listPos.top)
		{
			// Scroll up
			listTag.scrollTop = listTag.scrollTop - (listPos.top - itemPos.top);
		}
		else if (itemPos.bottom > listPos.bottom)
		{
			// Scroll down
			listTag.scrollTop = listTag.scrollTop + (itemPos.bottom - listPos.bottom);
		}
	}
}

/**
	Represents a plate for storing biomaterial.
	
 	@param prefix Prefix used to find DOM tags for the wells
	@param id The id of the plate
	@param name The name of the plate
	@param rows The number of rows on the plate
	@param columns The number of columns on the plate
	@param isSourcePlate If the given plate is a source plate or destination plate
*/
function Plate(prefix, id, name, rows, columns, isSourcePlate)
{
	this.prefix = prefix;
	this.id = id;
	this.name = name;
	this.rows = rows;
	this.columns = columns;
	this.isSourcePlate = isSourcePlate;
	this.selectedWellClass = 'selected';
	this.linkedWellClass = 'linked';
	this.mappedWellClass = 'mapped';
	this.wells = new Array();
	
	/**
		Get the well on the given coordinate. If the coordinate it outside the
		plate boundaries, null is returned, otherwise a Well object.
	*/
	Plate.prototype.getWell = function(row, column)
	{
		if (row < 0 || row >= this.rows) return null;
		if (column < 0 || column >= this.columns) return null;
		var well = this.wells[row+':'+column];
		if (!well)
		{
			well = new Well(this, row, column);
			this.wells[this.wells.length] = well;
			this.wells[row + ':' + column] = well;
		}
		return well;
	}
	
	Plate.prototype.unmapAll = function(graphics)
	{
		for (var i = 0; i < this.wells.length; i++)
		{
			var well = this.wells[i];
			well.hideLink(graphics);
			well.unmapWell();
		}
	}

}

/**
	A well on a plate.
	@param plate The plate
	@param row The well's row coordinate
	@param column The well's column coordinate
	@returns
*/
function Well(plate, row, column)
{
	this.plate = plate;
	this.row = row;
	this.column = column;
	this.selected = false;
	this.linked = false;
	this.mappedItem = null;
	
	/**
		Get the DOM tag that represents the well. The element should have an
		ID combined of 'plate-prefix.row.column'.
	 */
	Well.prototype.getTag = function()
	{
		if (!this.tag)
		{
			this.tag = document.getElementById(this.plate.prefix+'.'+this.row+'.'+this.column);
		}
		return this.tag;
	}
	
	/**
		Add a CSS class to the well DOM tag.
	*/
	Well.prototype.addClass = function(cls)
	{
		Main.addClass(this.getTag(), cls);
	}
	
	/**
		Remove a CSS class from the item's DOM tag.
	*/
	Well.prototype.removeClass = function(cls)
	{
		Main.removeClass(this.getTag(), cls);
	}
	
	/**
		Get the coordinate of the well as a string. Rows are alphabetically
		numbered.
	 */
	Well.prototype.getCoordinate = function()
	{
		return Plates.toAlphaCoordinate[this.row] + (this.column+1);
	}

	/**
		Set the contents of the well element. The tag should have an ID combined
		of 'plate-prefix.row.column.info'.
	*/
	Well.prototype.setInfo = function(info)
	{
		var infoTag = document.getElementById(this.plate.prefix+'.'+this.row+'.'+this.column + '.info');
		if (!infoTag) return;
		infoTag.innerHTML = info;
	}
	
	/**
		Mark the well as selected/deselected.
	*/
	Well.prototype.setSelected = function(selected)
	{
		this.selected = selected;
		Main.addOrRemoveClass(this.getTag(), this.plate.selectedWellClass, selected);
	}

	/**
		Mark the well as linked. Do not call this method. Use drawLink/hideLink 
		instead.
	*/
	Well.prototype.setLinked = function(linked)
	{
		this.linked = linked;
		Main.addOrRemoveClass(this.getTag(), this.plate.linkedWellClass, linked);
	}

	
	/**
		Draw a link from the well to the item it is mapped to. If
		the well is not mapped, no link is drawn.
		
		@param graphics A jsGraphics object
		@param pen A jsGraphics pen object
		@param overwrite TRUE to overwrite existing link, FALSE to not draw if there already is a link
	*/
	Well.prototype.drawLink = function(graphics, pen, overwrite)
	{
		if (this.mappedItem)
		{
			this.mappedItem.drawLink(graphics, pen, overwrite);
		}
		else if (this.mappedWell)
		{
			if (this.mappedWell.plate.isSourcePlate)
			{
				this.mappedWell.drawLink(graphics, pen, overwrite);
			}
			else
			{
				if (this.linkDiv) 
				{
					if (!overwrite) return;
					this.hideLink(graphics);
				}
				
				var srcPos = Main.getElementPosition(this.getTag());
				var destPos = Main.getElementPosition(this.mappedWell.getTag());
				var points = new Array();
				points[0] = new jsPoint(srcPos.left + srcPos.width, srcPos.top+srcPos.height/2);
				points[1] = new jsPoint(destPos.left, destPos.top+destPos.height/2);
				this.linkDiv = graphics.drawLine(pen, points[0], points[1]);
				this.setLinked(true);
				this.mappedWell.setLinked(true);
			}
		}
	}
	
	/**
		Hide the link from the well to the item/well it is mapped to.
	*/
	Well.prototype.hideLink = function(graphics)
	{
		if (this.mappedItem)
		{
			this.mappedItem.hideLink(graphics);
		}
		if (this.linkDiv)
		{
			graphics.clearDrawing(this.linkDiv);
			this.linkDiv = null;
			this.setLinked(false);
			if (this.mappedWell)
			{
				this.mappedWell.setLinked(false);
			}
		}
		else if (this.mappedWell && this.mappedWell.plate.isSourcePlate)
		{
			this.mappedWell.hideLink(graphics);
		}
	}
	
	/**
		Remove the current mapping.
	*/
	Well.prototype.unmapItem = function()
	{
		if (!this.mappedItem) return;
		this.mappedItem.unmapWell(this);
	}
	
	/**
		Map this *source* well to the other *destination* well.
		An existing mapping for the current well is automatically removed. 
	*/
	Well.prototype.mapToWell = function(well)
	{
		this.unmapWell();
		well.unmapWell();
		this.mappedWell = well;
		well.mappedWell = this;
		this.addClass(this.plate.mappedWellClass);
		well.addClass(well.plate.mappedWellClass);
		well.setInfo(this.getCoordinate());
	}

	/**
		Remove the current mapping to another well.
	*/
	Well.prototype.unmapWell = function()
	{
		if (!this.mappedWell) return;
		this.mappedWell.setInfo('');
		this.removeClass(this.plate.mappedWellClass);
		this.mappedWell.removeClass(this.mappedWell.plate.mappedWellClass);
		this.mappedWell.mappedWell = null;
		this.mappedWell = null;
	}
	
}

// Namespace for utility functions
var Plates = {};

Plates.toAlphaCoordinate = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 
                            'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z'];



