
package com.dawidweiss.ant.filters;

import java.io.IOException;
import java.io.Reader;
import java.util.Iterator;
import java.util.Hashtable;

import org.apache.tools.ant.types.*;
import org.apache.tools.ant.filters.*;
import org.apache.tools.ant.Project;
import org.apache.tools.ant.util.*;
import java.io.*;
import java.util.Properties;



/**
 * Replaces tokens in the original input with file contents read from
 * properties supplied by the user.
 */
public final class ReplaceTokensWithFileContents
    extends BaseParamFilterReader
    implements ChainableReader {

    /** Default "begin token" character. */
    private static final char DEFAULT_BEGIN_TOKEN = '@';

    /** Default "end token" character. */
    private static final char DEFAULT_END_TOKEN = '@';

    /** Data that must be read from, if not null. */
    private StringBuffer queuedData = new StringBuffer();

    /** Hashtable to hold the replacee-replacer pairs (String to String). */
    private Hashtable hash = new Hashtable();

    /** Character marking the beginning of a token. */
    private char beginToken = DEFAULT_BEGIN_TOKEN;

    /** Character marking the end of a token. */
    private char endToken = DEFAULT_END_TOKEN;

    /**
     * Constructor for "dummy" instances.
     * 
     * @see BaseFilterReader#BaseFilterReader()
     */
    public ReplaceTokensWithFileContents() {
        super();
    }

    /**
     * Creates a new filtered reader.
     *
     * @param in A Reader object providing the underlying stream.
     *           Must not be <code>null</code>.
     */
    public ReplaceTokensWithFileContents(final Reader in) {
        super(in);
    }


    /**
     * Returns the next character in the filtered stream, replacing tokens
     * from the original stream.
     * 
     * @return the next character in the resulting stream, or -1
     * if the end of the resulting stream has been reached
     * 
     * @exception IOException if the underlying stream throws an IOException
     * during reading     
     */
    private final int internalRead() throws IOException {
        final int ch;
        if (queuedData.length() > 0) {
            ch = queuedData.charAt(0);
            if (queuedData.length() > 0) {
                queuedData.deleteCharAt(0);
            }
        } else {
            ch = in.read();
        }
        return ch;
    }

    public final int read() throws IOException {
        if (!getInitialized()) {
            initialize();
            setInitialized(true);
        }
        
        int ch = internalRead();
        if (ch == beginToken) {
            final StringBuffer key = new StringBuffer();
            do  {
                ch = internalRead();
                if (ch == endToken || ch == -1)
                    break;
                if (ch == beginToken) {
                    queuedData.insert( 0, key.toString() + beginToken );
                    return beginToken;
                }
                key.append( (char) ch );
            } while (true);

            if (ch == -1) {
                queuedData.insert(0, key.toString());
                return beginToken;
            } else {
                ReplaceableToken token = (ReplaceableToken) hash.get(key.toString());

                if (token != null) {
                    getProject().log("Recognized token: " + (key.length() > 60 ? key.substring(0,60) + "..." : key.toString()), Project.MSG_VERBOSE );

                    queuedData.insert(0,token.getReplacement());

                    return read();
                } else {
                    getProject().log("Unrecognized token: " + (key.length() > 60 ? key.substring(0,60) + "..." : key.toString()), Project.MSG_DEBUG );
                    queuedData.insert( 0, key.toString() + endToken );
                    return beginToken;
                }
            }
        }
        return ch;
    }

    /**
     * Sets the "begin token" character.
     * 
     * @param beginToken the character used to denote the beginning of a token
     */
    public final void setBeginToken(final char beginToken) {
        this.beginToken = beginToken;
    }

    /**
     * Returns the "begin token" character.
     * 
     * @return the character used to denote the beginning of a token
     */
    private final char getBeginToken() {
        return beginToken;
    }

    /**
     * Sets the "end token" character.
     * 
     * @param endToken the character used to denote the end of a token
     */
    public final void setEndToken(final char endToken) {
        this.endToken = endToken;
    }

    /**
     * Returns the "end token" character.
     * 
     * @return the character used to denote the end of a token
     */
    private final char getEndToken() {
        return endToken;
    }

    /**
     * Adds a token element to the map of tokens to replace.
     * 
     * @param token The token to add to the map of replacements.
     *              Must not be <code>null</code>.
     */
    public final void addConfiguredToken(final Token token) {
        hash.put(token.getKey(), token.getValue());
    }

    /**
     * Sets the map of tokens to replace.
     * 
     * @param hash A map (String->String) of token keys to replacement
     * values. Must not be <code>null</code>.
     */
    private void setTokens(final Hashtable hash) {
        this.hash = hash;
    }

    /**
     * Returns the map of tokens which will be replaced.
     * 
     * @return a map (String->String) of token keys to replacement
     * values
     */
    private final Hashtable getTokens() {
        return hash;
    }

    /**
     * Creates a new ReplaceTokens using the passed in
     * Reader for instantiation.
     * 
     * @param rdr A Reader object providing the underlying stream.
     *            Must not be <code>null</code>.
     * 
     * @return a new filter based on this configuration, but filtering
     *         the specified reader
     */
    public final Reader chain(final Reader rdr) {
        ReplaceTokensWithFileContents newFilter = new ReplaceTokensWithFileContents(rdr);
        newFilter.setBeginToken(getBeginToken());
        newFilter.setEndToken(getEndToken());
        newFilter.setTokens(getTokens());
        newFilter.setInitialized(true);
        return newFilter;
    }

    /**
     * Initializes tokens and loads the replacee-replacer hashtable.
     */
    private final void initialize() {
        Parameter[] params = getParameters();
        if (params != null) {
            for (int i = 0; i < params.length; i++) {
                if (params[i] != null) {
                    final String type = params[i].getType();
                    if ("tokenchar".equals(type)) {
                        final String name = params[i].getName();
                        if ("begintoken".equals(name)) {
                            beginToken = params[i].getValue().charAt(0);
                        } else if ("endtoken".equals(name)) {
                            endToken = params[i].getValue().charAt(0);
                        }
                    } else if ("token".equals(type)) {
                        final String name = params[i].getName();
                        final String value = params[i].getValue();
                        hash.put(name, value);
                    } else if ("file".equals(type)) {
                        File props = getProject().resolveFile( params[i].getValue() );
                        if (props.canRead()) {
                            InputStream inputs = null;
                            try
                            {
                                inputs = new FileInputStream( props );
                                Properties p = new Properties();
                                p.load(inputs);

                                // resolve files pointed to by the token properties relative
                                for (Iterator keyIterator = p.keySet().iterator(); keyIterator.hasNext(); )
                                {
                                    String key = (String) keyIterator.next();
                                    hash.put( key, new ReplaceableToken( props, key, p.getProperty(key)));
                                }
                            }
                            catch (IOException e)
                            {
                                throw new org.apache.tools.ant.BuildException("Error when reading tokens file " 
                                    + props.getAbsolutePath() + ": " + e.toString());
                            }
                            finally {
                                try {
                                    if (inputs != null) inputs.close();
                                }
                                catch (IOException e) {}
                            }
                            getProject().log("Loaded tokens from file: " + props.getAbsolutePath(), Project.MSG_VERBOSE);
                        }
                        else {
                            getProject().log("Cannot read tokens file: " + props.getAbsolutePath(), Project.MSG_ERR ); 
                            throw new org.apache.tools.ant.BuildException("Cannot read tokens file: " + props.getAbsolutePath());
                        }
                    } else {
                        getProject().log("Unknown parameter: " + type, Project.MSG_VERBOSE);
                        throw new org.apache.tools.ant.BuildException("Unknown parameter: " + type);
                    }
                }
            }
        }
    }
    
    
    /** Holds a resolved token */
    private class ReplaceableToken {
        private String encoding;
        private File   replacement;
        
        public ReplaceableToken(File base, String key, String value) throws IOException {
            encoding = "UTF-8";
            String path = value;
            
            if (base.isFile())
                base = base.getParentFile();

            if (path.startsWith("encoding:")) {
                int i = path.indexOf(":");
                int j = path.indexOf(":", i+1);
                encoding = path.substring(i+1, j);
                path = path.substring( j+1 );
            }
            
            path = getProject().replaceProperties(path);
            File replacement = FileUtils.newFileUtils().resolveFile(base, path);

            getProject().log("Token: " + key + " value: " + value 
                + " resolved as: " + replacement.getAbsolutePath()
                + " relative to: " + base.getAbsolutePath(), Project.MSG_VERBOSE);

            if (replacement.exists() && replacement.canRead()) {
                this.replacement = replacement;
            }
            else {
                throw new IOException("Cannot find token replacement file: "
                    + replacement.getAbsolutePath());
            }
        }
        
        public String getReplacement() throws UnsupportedEncodingException, IOException {
            FileInputStream is = new FileInputStream( replacement );
            try {
                byte [] contents = new byte [ (int) replacement.length() ];
                
                if (is.read( contents ) != replacement.length())
                    throw new IOException("Could not read entire file: " + replacement.getAbsolutePath());
                
                return new String( contents, encoding );
            }
            finally {
                try {
                    if (is != null) is.close();
                }
                catch (IOException e) {}
            }
        }
    }

    /**
     * Holds a token
     */
    public static class Token {

        /** Token key */
        private String key;

        /** Token value */
        private String value;

        /**
         * Sets the token key
         * 
         * @param key The key for this token. Must not be <code>null</code>.
         */
        public final void setKey(String key) {
            this.key = key;
        }

        /**
         * Sets the token value
         * 
         * @param value The value for this token. Must not be <code>null</code>.
         */
        public final void setValue(String value) {
            this.value = value;
        }

        /**
         * Returns the key for this token.
         * 
         * @return the key for this token
         */
        public final String getKey() {
            return key;
        }

        /**
         * Returns the value for this token.
         * 
         * @return the value for this token
         */
        public final String getValue() {
            return value;
        }
    }
}
